/*
 * Copyright (c) 2024, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef HSS_HANDLER_TASK_H_
#define HSS_HANDLER_TASK_H_

#include <stdint.h>
#include "ti_hssc.h"
#include "tps2hcs10.h"

/* Packet Definitions */
#define HSS_DEV_PAYLOAD_SIZE 32
#define HSS_DEV_QUEUE_LENGTH 32
#define HSS_SPI_FREQUENCY 2000000
#define NUM_INPUT_REPORTS 16

/* Header used for write commands */
typedef __packed struct sWriteHeader
{
    unsigned addr : 7;
    unsigned rw : 1;
} tWriteHeader;

typedef __packed struct sSPIFrameWrite
{
    tWriteHeader header;
    uint8_t payload[2];
} tSPIFrameWrite;

typedef __packed struct sHSSHostCommand
{
    uint8_t opcode;
    uint8_t addr;
    uint8_t payloadSize;
    uint8_t payload[HSS_DEV_PAYLOAD_SIZE];
} tHSSHostCommand;

typedef __packed struct sOutputFeatureReport
{
    uint8_t dirty;
    uint8_t addr;
    uint8_t payloadSize;
    uint8_t payload[HSS_DEV_PAYLOAD_SIZE];
} tOutputFeatureReport;

/* Command OpCodes */
#define HSS_DATA_WRITE_CMD 0x01
#define HSS_DATA_READ_CMD 0x02
#define HSS_DIGIPOT_SET_CMD 0x03
#define HSS_UPDATE_DEV_CONFIG_CMD 0x04
#define HSS_UPDATE_CH_CONFIG_CMD 0x05
#define HSS_UPDATE_FLT_CONFIG_CMD 0x06
#define HSS_UPDATE_SNS_CMD 0x07
#define HSS_INITIATE_FIRMWARE_UPDATE 0x08
#define HSS_IOCTL 0x09
#define HSS_SAVE_CONFIG_FLASH 0x0A
#define HSS_LOAD_CONFIG_FLASH 0x0B
#define HSS_ERASE_CONFIG_FLASH 0x0C

/* IOCTL Codes */
#define HSS_IOCTL_SETLHI 0x01
#define HSS_IOCTL_GETFW_VER 0x02

/* Channel Mappings */
#define HSS_CH_1 1
#define HSS_CH_0 0


typedef __packed struct sHSSRawReadResponse
{
    uint8_t dirtyByte;
    uint8_t addr;
    uint16_t data;
} tHSSRawReadResponse;

typedef __packed struct sReadResponse
{
    uint8_t fltStatus;
    uint16_t payload;
} tReadResponse;

typedef __packed struct sIOCTLResponse
{
    uint8_t dirtyByte;
    uint8_t ioctl;
    uint8_t payload[16];
} tIOCTLResponse;

/* Structure used for the configuration block of TPS2HCS10-Q1 */
typedef __packed struct sSmartFuseChanConfig
{
    uint8_t dirtyByte;
    uint8_t channelNum;
    TPS2HC10S_PWM_CH1 pwmConfig;
    TPS2HC10S_ILIM_CONFIG_CH1 iLimConfig;
    TPS2HC10S_DIAG_CONFIG_CH1 diagConfig;
    TPS2HC10S_I2T_CONFIG_CH1 i2tConfig;
    uint8_t padding[2];
} tSmartFuseChanConfig;

typedef __packed struct sSmartFuseDevConfig
{
    uint8_t dirtyByte;
    TPS2HC10S_CRC_CONFIG crcConf;
    TPS2HC10S_LPM lpm;
    TPS2HC10S_FAULT_MASK devFaultMask;
    TPS2HC10S_ABIST_RESULT abistRes;
    TPS2HC10S_SW_STATE swState;
    TPS2HC10S_DEVICE_SAF devStatConfig;
    TPS2HC10S_DEV_CONFIG devConfig;
    TPS2HC10S_ADC_CONFIG adcConfig;
    TPS2HC10S_SLEEP sleepConfig;
    uint8_t lhiStatus;
} tSmartFuseDevConfig;

/* Structure used for the fault block of TPS2HCSxx */
typedef __packed struct sHSSDiagnosticData
{
    uint8_t dirtyByte;
    TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE globalFaultType;
    TPS2HC10S_FLT_STAT_CH1 channel1Fault;
    TPS2HC10S_FLT_STAT_CH2 channel2Fault;
    uint8_t faultPinStatus;
} tHSSFaultData;

#define HSS_BYTE_SWAP(f) ((f << 8) | (f >> 8))

void hssc_handleGenericWrite(uint8_t addr, uint8_t* payload, uint8_t size);
void hssc_handleGenericWriteReg(uint8_t addr, uint16_t payload);
void hssc_handleGenericRead(uint8_t addr, uint8_t* respData, uint8_t size);
void hssc_handleGenericReadReg(uint8_t addr, uint8_t* respData);

#define HSSC_CH_OFFSET 8

#endif
