/*
 * Copyright (c) 2024, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
 /* Standard includes */
 #include <stdbool.h>
 #include <stdint.h>
 #include <string.h>

/* Device/Driver Includes */
#include "inc/hw_gpio.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_sysctl.h"
#include "inc/hw_ssi.h"
#include "inc/hw_nvic.h"
#include "driverlib/sysctl.h"
#include "driverlib/flash.h"
#include "driverlib/gpio.h"
#include "driverlib/ssi.h"
#include "driverlib/i2c.h"
#include "driverlib/pin_map.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"

/* Program includes */
#include "hss_handler_task.h"
#include "sense_handler_task.h"
#include "hss_definitions.h"
#include "usb_spi_dev.h"

/* RTOS header files */
#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>

/* RTOS Elements */
QueueHandle_t xHSSCommandQueue;
void hsscCommandTask(void *pvParameters);

/* Statics Functions */
static void spiSendData(uint8_t *data, uint8_t len, uint8_t* respData);
static void updateConfigurationData(tSmartFuseDevConfig* hssConfig);
static void updateChanConfigurationData(tSmartFuseChanConfig* hssConfig,
                                            uint8_t chOffset);
static void updateFaultData(tHSSFaultData* faultData);
static void handleIOCTL(uint8_t ioctl, uint8_t* data);
static void loadConfigurationFromFlash(void);

/* Static Variables */
static tHSSHostCommand curHssCommand;
static tSmartFuseDevConfig devConfig;
static tSmartFuseChanConfig chan1Config;
static tSmartFuseChanConfig chan2Config;
static SemaphoreHandle_t mutexSPISem;
static tReadResponse readResp;
static tHSSFaultData faultData;
const uint32_t magicWord = HSS_MAGIC_WORD;
uint32_t* magicWordPtr = (uint32_t*)HSS_FLASH_CONFIG_ADDR;

void hsscCommandTask(void *pvParameters)
{
    /* Create the queue used by the queue send and queue receive tasks. */
    xHSSCommandQueue = xQueueCreate(HSS_DEV_QUEUE_LENGTH, 
                                       sizeof(curHssCommand));

    /* Initializing the semaphore */
    mutexSPISem = xSemaphoreCreateMutex();

    /* Configuring the SPI peripheral IOs */
    MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOA);
    MAP_GPIOPinConfigure(GPIO_PA2_SSI0CLK);
    MAP_GPIOPinConfigure(GPIO_PA4_SSI0RX);
    MAP_GPIOPinConfigure(GPIO_PA5_SSI0TX);
    GPIODirModeSet(GPIO_PORTA_BASE, GPIO_PIN_5 | GPIO_PIN_4 |
                   GPIO_PIN_2, GPIO_DIR_MODE_HW);
    GPIOPadConfigSet(GPIO_PORTA_BASE, GPIO_PIN_5 | GPIO_PIN_4 |
                   GPIO_PIN_2, GPIO_STRENGTH_8MA, GPIO_PIN_TYPE_STD);
    GPIOPinTypeGPIOOutput(GPIO_PORTA_BASE, GPIO_PIN_3);
    GPIOPinWrite(GPIO_PORTA_BASE, GPIO_PIN_3, GPIO_PIN_3);

    /* Configuring the SPI module */
    MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_SSI0);
    MAP_SSIClockSourceSet(SSI0_BASE, SSI_CLOCK_SYSTEM);
    MAP_SSIConfigSetExpClk(SSI0_BASE, SysCtlClockGet(), SSI_FRF_MOTO_MODE_1,
                       SSI_MODE_MASTER, HSS_SPI_FREQUENCY, 8);
    MAP_SSIEnable(SSI0_BASE);
    MAP_SSIIntEnable(SSI0_BASE, SSI_RXFF | SSI_RXTO);

    /* Pin configuration for FAULT */
    GPIOPinTypeGPIOInput(GPIO_PORTF_BASE, GPIO_PIN_2);

    /* Pin configuration for LHI */
    GPIOPinWrite(GPIO_PORTF_BASE, GPIO_PIN_3, 0);
    GPIOPinTypeGPIOOutput(GPIO_PORTF_BASE, GPIO_PIN_3);

    /* Zeroing out our input reports */
    memset((void*)&faultData, 0x00, sizeof(faultData));
    memset((void*)&chan1Config, 0x00, sizeof(chan1Config));
    memset((void*)&chan2Config, 0x00, sizeof(chan2Config));
    chan1Config.channelNum = 0;
    chan2Config.channelNum = 1;
    memset((void*)&devConfig, 0x00, sizeof(devConfig));

    /* If the magic word is at the configuration memory space, it means we have
        a saved configuration */
    loadConfigurationFromFlash();

    while(1)
    {
        /* Wait for the next command */
        xQueueReceive(xHSSCommandQueue, &curHssCommand, portMAX_DELAY);
        
        /* Command handler */
        switch(curHssCommand.opcode)
        {
            case HSS_DATA_WRITE_CMD:
                hssc_handleGenericWrite(curHssCommand.addr,
                                  (uint8_t*)curHssCommand.payload,
                                  curHssCommand.payloadSize);
                break;
            case HSS_DATA_READ_CMD:
                /* Doing the read twice */
                hssc_handleGenericRead(curHssCommand.addr,
                                  (uint8_t*)curHssCommand.payload + 1,
                                   curHssCommand.payloadSize);
                hssc_handleGenericRead(curHssCommand.addr,
                                  (uint8_t*)curHssCommand.payload + 1,
                                   curHssCommand.payloadSize);

                /* Making a dirty byte and sending it */
                curHssCommand.payload[0] = 0x01;
                USBHID_HSS_sendReport(HSS_USB_FEATURE_RAWREAD_ID,
                                        (uint8_t*)curHssCommand.payload,
                                            portTICK_PERIOD_S);
                break;
            case HSS_UPDATE_FLT_CONFIG_CMD:
                updateFaultData(&faultData);
                if(GPIOPinRead(GPIO_PORTF_BASE, GPIO_PIN_2) & GPIO_PIN_2)
                {
                    faultData.faultPinStatus = 0;
                }
                else
                {
                    faultData.faultPinStatus = 1;
                }
                USBHID_HSS_sendReport(HSS_USB_FEATURE_FAULT_ID,
                                        (uint8_t*)&faultData,
                                            portTICK_PERIOD_S);
                break;
            case HSS_UPDATE_DEV_CONFIG_CMD:
                updateConfigurationData(&devConfig);
                USBHID_HSS_sendReport(HSS_USB_FEATURE_DEV_CONF_ID,
                                        (uint8_t*)&devConfig,
                                            portTICK_PERIOD_S);
                break;
            case HSS_UPDATE_CH_CONFIG_CMD:
                if(curHssCommand.addr == HSS_CH_1)
                {
                    updateChanConfigurationData(&chan2Config, HSSC_CH_OFFSET);
                    USBHID_HSS_sendReport(HSS_USB_FEATURE_CH_CONF_ID,
                        (uint8_t*)&chan2Config, portTICK_PERIOD_S);
                }
                else if(curHssCommand.addr == HSS_CH_0)
                {
                    updateChanConfigurationData(&chan1Config, 0);
                    USBHID_HSS_sendReport(HSS_USB_FEATURE_CH_CONF_ID,
                        (uint8_t*)&chan1Config, portTICK_PERIOD_S);
                }
                break;
            case HSS_UPDATE_SNS_CMD:
                setCurrentSenseReporting((SENSE_READ_MODE)curHssCommand.payload[0],
                                            curHssCommand.payload[1]);
                break;
            case HSS_INITIATE_FIRMWARE_UPDATE:
                /* Firmware update. From here, we have to have to disable USB,
                    and invoke the ROM bootloader */
                ROM_USBDevDisconnect(USB0_BASE);
                ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOD);
                ROM_GPIOPinTypeUSBAnalog(GPIO_PORTD_BASE, GPIO_PIN_5
                                            | GPIO_PIN_4);

                ROM_IntMasterDisable();
                ROM_SysTickIntDisable();
                ROM_SysTickDisable();

                HWREG(NVIC_DIS0) = 0xffffffff;
                HWREG(NVIC_DIS1) = 0xffffffff;

                ROM_SysCtlPeripheralEnable(SYSCTL_PERIPH_USB0);
                ROM_SysCtlPeripheralReset(SYSCTL_PERIPH_USB0);
                ROM_SysCtlUSBPLLEnable();

                ROM_SysCtlDelay(ROM_SysCtlClockGet() / 3);

                ROM_IntMasterEnable();
                ROM_UpdateUSB(0);
                break;
            case HSS_SAVE_CONFIG_FLASH:
                ROM_FlashErase(HSS_FLASH_CONFIG_ADDR);
                ROM_FlashProgram((void*)&magicWord, HSS_FLASH_CONFIG_ADDR, 
                                    sizeof(uint32_t));
                ROM_FlashProgram((void*)&devConfig, HSS_CONFIG_DEV_ADDR, 
                                    sizeof(tSmartFuseDevConfig));
                ROM_FlashProgram((void*)&chan1Config, HSS_CONFIG_CHAN1_ADDR, 
                                    sizeof(tSmartFuseChanConfig));
                ROM_FlashProgram((void*)&chan2Config, HSS_CONFIG_CHAN2_ADDR, 
                                    sizeof(tSmartFuseChanConfig));
                break;
            case HSS_LOAD_CONFIG_FLASH:
                loadConfigurationFromFlash();
                break;
            case HSS_ERASE_CONFIG_FLASH:
                ROM_FlashErase(HSS_FLASH_CONFIG_ADDR);
                break;
            case HSS_IOCTL:
                handleIOCTL(curHssCommand.addr, &curHssCommand.payloadSize);
                break;
            default:
                break;
        }
    }
}

static void handleIOCTL(uint8_t ioctl, uint8_t* data)
{
    tIOCTLResponse resp;

    switch(ioctl)
    {
        case HSS_IOCTL_SETLHI:
            if(data[0] == 0x00)
            {
                GPIOPinWrite(GPIO_PORTF_BASE, GPIO_PIN_3, 0);
            }
            else
            {
                GPIOPinWrite(GPIO_PORTF_BASE, GPIO_PIN_3, GPIO_PIN_3);
            }
            break;
        case HSS_IOCTL_GETFW_VER:
            resp.dirtyByte = 0x01;
            resp.ioctl = HSS_IOCTL_GETFW_VER;
            resp.payload[0] = HSS_FIRMWARE_VER_ID & 0xFF;
            resp.payload[1] = HSS_FIRMWARE_VER_ID >> 8;
            USBHID_HSS_sendReport(HSS_USB_FEATURE_IOCTL_ID,
                        (uint8_t*)&resp, portTICK_PERIOD_S);
            break;
        default:
            break;
    }
}

static void spiSendData(uint8_t *data, uint8_t len, uint8_t* respData)
{
    uint8_t ii;
    uint32_t tempData;
    xSemaphoreTake(mutexSPISem, portMAX_DELAY);

    GPIOPinWrite(GPIO_PORTA_BASE, GPIO_PIN_3, 0);

    for(ii=0;ii<len;ii++)
    {
        MAP_SSIDataPut(SSI0_BASE, data[ii]);
        MAP_SSIDataGet(SSI0_BASE, &tempData);

        if(respData != NULL)
        {
            respData[ii] = tempData & 0xFF;
        }
    }

    /* Waiting for the SPI transaction to finish before de-asserting CS */
    while(HWREG(SSI0_BASE + SSI_O_SR) & SSI_SR_BSY);
    GPIOPinWrite(GPIO_PORTA_BASE, GPIO_PIN_3, GPIO_PIN_3);

    xSemaphoreGive(mutexSPISem);
}

void hssc_handleGenericWriteReg(uint8_t addr, uint16_t payload)
{
    /* Constructing the packet */
    tSPIFrameWrite writeCmd = 
    {
        .header = 
        {
            .addr = addr,
            .rw = 1,
        },
    };

    /* Copying the payload. Since this is just 2-bytes,
        we are doing manually. */
    writeCmd.payload[1] = (uint8_t)(payload & 0xFF);
    writeCmd.payload[0] = (uint8_t)((payload >> 8) & 0xFF);

    /* Sending the write command */
    spiSendData((uint8_t*)&writeCmd, 3, NULL);
}

void hssc_handleGenericWrite(uint8_t addr, uint8_t* payload, uint8_t size)
{
    /* Constructing the packet */
    tSPIFrameWrite writeCmd = 
    {
        .header = 
        {
            .addr = addr,
            .rw = 1,
        },
    };

    /* Copying the payload. Since this is just 2-bytes,
        we are doing manually. */
    writeCmd.payload[0] = payload[1];
    writeCmd.payload[1] = payload[0];

    /* Sending the write command */
    spiSendData((uint8_t*)&writeCmd, size+1, NULL);
}

void hssc_handleGenericRead(uint8_t addr, uint8_t* respData, uint8_t size)
{
    /* Constructing the packet */
    tSPIFrameWrite readCmd = 
    {
        .header = 
        {
            .addr = addr,
            .rw = 0,
        },
    };
    
    /* Sending the read command */
    spiSendData((uint8_t*)&readCmd, size+1, respData);
}

void hssc_handleGenericReadReg(uint8_t addr, uint8_t* respData)
{
    uint8_t tempData;

    /* Constructing the packet */
    tSPIFrameWrite readCmd = 
    {
        .header = 
        {
            .addr = addr,
            .rw = 0,
        },
    };

    /* Sending the read command */
    spiSendData((uint8_t*)&readCmd, 3, respData);

    tempData = respData[0];
    respData[0] = respData[1];
    respData[1] = tempData;
}

/* Polls the HSS for updated channel configuration information for provided
    channel number */
static void updateChanConfigurationData(tSmartFuseChanConfig* hssConfig,
                                            uint8_t chOffset)
{
    /* Populating all of the data */
    hssc_handleGenericRead(TPS2HC10S_PWM_CH1_REG + chOffset, (uint8_t*)&readResp, 2);
    hssc_handleGenericRead(TPS2HC10S_ILIM_CONFIG_CH1_REG + chOffset,
                        (uint8_t*)&readResp, 2);
    hssConfig->pwmConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_DIAG_CONFIG_CH1_REG + chOffset,
                        (uint8_t*)&readResp, 2);
    hssConfig->iLimConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_I2T_CONFIG_CH1_REG + chOffset,
                        (uint8_t*)&readResp, 2);
    hssConfig->diagConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_I2T_CONFIG_CH1_REG + chOffset,
                        (uint8_t*)&readResp, 2);
    hssConfig->i2tConfig.word = readResp.payload;

    /* Setting the dirty bit so the host knows we have new data */
    hssConfig->dirtyByte = 1;
}

/* Polls the HSS for device data */
static void updateConfigurationData(tSmartFuseDevConfig* hssConfig)
{
    /* Populating all of the data */
    hssc_handleGenericRead(TPS2HC10S_CRC_CONFIG_REG, (uint8_t*)&readResp, 2);
    hssc_handleGenericRead(TPS2HC10S_LPM_REG, (uint8_t*)&readResp, 2);
    hssConfig->crcConf.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_FAULT_MASK_REG, (uint8_t*)&readResp, 2);
    hssConfig->lpm.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_ABIST_RESULT_REG, (uint8_t*)&readResp, 2);
    hssConfig->devFaultMask.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_SW_STATE_REG, (uint8_t*)&readResp, 2);
    hssConfig->abistRes.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_DEVICE_SAF_REG, (uint8_t*)&readResp, 2);
    hssConfig->swState.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_DEV_CONFIG_REG, (uint8_t*)&readResp, 2);
    hssConfig->devStatConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_ADC_CONFIG_REG, (uint8_t*)&readResp, 2);
    hssConfig->devConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_SLEEP_REG, (uint8_t*)&readResp, 2);
    hssConfig->adcConfig.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_SLEEP_REG, (uint8_t*)&readResp, 2);
    hssConfig->sleepConfig.word = readResp.payload;

    /* Setting the fault bit */
    if(GPIOPinRead(GPIO_PORTF_BASE, GPIO_PIN_3))
    {
        hssConfig->lhiStatus = 1;
    }
    else
    {
        hssConfig->lhiStatus = 0;
    }

    /* Setting the dirty bit so the host knows we have new data */
    hssConfig->dirtyByte = 1;
}

/* Polls the HSS for fault data */
static void updateFaultData(tHSSFaultData* faultData)
{
    /* Populating all of the data */
    hssc_handleGenericRead(TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_REG,
                            (uint8_t*)&readResp, 2);
    hssc_handleGenericRead(TPS2HC10S_FLT_STAT_CH1_REG, (uint8_t*)&readResp, 2);
    faultData->globalFaultType.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_FLT_STAT_CH2_REG, (uint8_t*)&readResp, 2);
    faultData->channel1Fault.word = readResp.payload;

    hssc_handleGenericRead(TPS2HC10S_FLT_STAT_CH2_REG, (uint8_t*)&readResp, 2);
    faultData->channel2Fault.word = readResp.payload;

    /* Setting the dirty bit so the host knows we have new data */
    faultData->dirtyByte = 1;
}

static void loadConfigurationFromFlash(void)
{
    /* Checking to see if the magic word is there. If it is, it means that there
        is a configuration data waiting to be loaded to the device */
    if(*(magicWordPtr) == HSS_MAGIC_WORD)
    {
        memcpy((void*)&devConfig, (void*)HSS_CONFIG_DEV_ADDR,
                        sizeof(tSmartFuseDevConfig));
        memcpy((void*)&chan1Config, (void*)HSS_CONFIG_CHAN1_ADDR,
                        sizeof(tSmartFuseChanConfig));
        memcpy((void*)&chan2Config, (void*)HSS_CONFIG_CHAN2_ADDR,
                        sizeof(tSmartFuseChanConfig));

        /* Taking the buffered configuration data and pushing it to the high-
            side switch */
        hssc_handleGenericWriteReg(TPS2HC10S_CRC_CONFIG_REG, 
                                    HSS_BYTE_SWAP(devConfig.crcConf.word));
        hssc_handleGenericWriteReg(TPS2HC10S_LPM_REG, 
                                    HSS_BYTE_SWAP(devConfig.lpm.word));
        hssc_handleGenericWriteReg(TPS2HC10S_FAULT_MASK_REG, 
                                   HSS_BYTE_SWAP(devConfig.devFaultMask.word));
        hssc_handleGenericWriteReg(TPS2HC10S_SW_STATE_REG, 
                                    HSS_BYTE_SWAP(devConfig.swState.word));
        hssc_handleGenericWriteReg(TPS2HC10S_DEVICE_SAF_REG, 
                                  HSS_BYTE_SWAP(devConfig.devStatConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_DEV_CONFIG_REG,
                                      HSS_BYTE_SWAP(devConfig.devConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_ADC_CONFIG_REG,
                                      HSS_BYTE_SWAP(devConfig.adcConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_SLEEP_REG,
                                    HSS_BYTE_SWAP(devConfig.sleepConfig.word));
                                    
        hssc_handleGenericWriteReg(TPS2HC10S_ILIM_CONFIG_CH1_REG,
                                   HSS_BYTE_SWAP(chan1Config.iLimConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_DIAG_CONFIG_CH1_REG,
                                   HSS_BYTE_SWAP(chan1Config.diagConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_I2T_CONFIG_CH1_REG,
                                    HSS_BYTE_SWAP(chan1Config.i2tConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_PWM_CH1_REG,
                                    HSS_BYTE_SWAP(chan1Config.pwmConfig.word));
                                    
        hssc_handleGenericWriteReg(TPS2HC10S_ILIM_CONFIG_CH2_REG,
                                   HSS_BYTE_SWAP(chan2Config.iLimConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_DIAG_CONFIG_CH2_REG,
                                   HSS_BYTE_SWAP(chan2Config.diagConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_I2T_CONFIG_CH2_REG,
                                    HSS_BYTE_SWAP(chan2Config.i2tConfig.word));
        hssc_handleGenericWriteReg(TPS2HC10S_PWM_CH2_REG,
                                    HSS_BYTE_SWAP(chan2Config.pwmConfig.word));
    }
}
