/*
 * Copyright (c) 2022, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,

 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <efuse_definitions.h>

/* Standard Includes */
#include <FreeRTOS.h>
#include <semphr.h>
#include <task.h>
#include <queue.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "i2c_algorithm_task.h"

/* Variable and state definitions */
void intADCConversionComplete(uintptr_t arg);
static SemaphoreHandle_t adcSem;
static SemaphoreHandle_t tickSem;
QueueHandle_t stateChangeQueue;
volatile bool persistValuesFlag = false;
volatile bool resetValuesFlag = false;
volatile bool stateChange = false;

/* Prototypes */
static void intTimerTickComplete(uintptr_t arg);
static void persistConfiguredValues(void);

/* Master Configuration Structure. This is used to keep track of state variables
 * during the algorithm run */
tFuseStateParameters curStates[EFUSE_NUM_OF_CHANNELS] =
{
     /* ALPHA2HC (Channel 1) */
    {
         .state = EFUSE_STATE_NO_SHUTDOWN,
         .softShutCounter = 0,
         .cooldownCounter = 0,
         .runningSum = 0,
         .skip = false
    },
    /* ALPHA2HC (Channel 2) */
    {
         .state = EFUSE_STATE_NO_SHUTDOWN,
         .softShutCounter = 0,
         .cooldownCounter = 0,
         .runningSum = 0,
         .skip = false
    }
};

/* Non-Volatile flash defaults. This can be saved via a SPI command  */
#pragma DATA_SECTION(savedConfigurations, ".fuseConfig")
#pragma RETAIN(savedConfigurations)
const tFuseChannel savedConfigurations[EFUSE_NUM_OF_CHANNELS];

tFuseChannel loadedConfigurations[EFUSE_NUM_OF_CHANNELS];

/* Default parameters - The user can reset to these values via a SPI command */
const tFuseChannel defaultConfigurations[EFUSE_NUM_OF_CHANNELS] =
{
     /* ALPHA2HC (Channel 1) */
     {
          .triggerThreshold = EFUSE_A2HC1_TRIGGER_THRESHOLD,
          .shutdownThreshold = EFUSE_A2HC1_SHUTDOWN_THRESHOLD,
          .softShutThreshold = EFUSE_A2HC1_SOFT_SHUT_THRESHOLD,
          .softCountThreshold = EFUSE_A2HC1_SOFT_SHUT_COUNT_THRESHOLD,
          .cooldownCountThreshold = EFUSE_A2HC1_COOLDOWN_COUNT_THRESHOLD,
     },
     /* ALPHA2HC (Channel 2) */
     {
          .triggerThreshold = EFUSE_A2HC1_TRIGGER_THRESHOLD,
          .shutdownThreshold = EFUSE_A2HC1_SHUTDOWN_THRESHOLD,
          .softShutThreshold = EFUSE_A2HC1_SOFT_SHUT_THRESHOLD,
          .softCountThreshold = EFUSE_A2HC1_SOFT_SHUT_COUNT_THRESHOLD,
          .cooldownCountThreshold = EFUSE_A2HC1_COOLDOWN_COUNT_THRESHOLD,
     },
};

void *adcSampleThread(void *arg0)
{
    uint32_t curResult;
    uint32_t ii;
    tFuseChannel* curFuse;
    tFuseStateParameters* curParam;
    tFuseStateParameters* curStateChange;

    /* Configuring our sampling timer to guarantee a delta T (100uS) */
    TIMER32_1->CONTROL = TIMER32_CONTROL_MODE | TIMER32_CONTROL_IE;
    TIMER32_1->LOAD = 4800;

    /* Create RTOS Queue */
    attr.mq_flags = 0;
    attr.mq_maxmsg = 64;
    attr.mq_msgsize = 2;
    attr.mq_curmsgs = 0;

    stateChangeQueue = mq_open("StateChange", O_RDWR | O_CREAT, 0664, &attr);

    /* Setting up timer tick interrupt */
    hwiParams.priority = configMAX_SYSCALL_INTERRUPT_PRIORITY;
    HwiP_create(INT_T32_INT1, intTimerTickComplete, &hwiParams);
    HwiP_enableInterrupt(INT_T32_INT1);

    /* Checking to see if there are values loaded in the flash location or
     * if we need to just keep the default values
     */
    if(savedConfigurations[0].triggerThreshold != 0xFFFFFFFF)
    {
        memcpy(loadedConfigurations, savedConfigurations,
               sizeof(loadedConfigurations));
    }
    else
    {
        memcpy(loadedConfigurations, defaultConfigurations,
                       sizeof(loadedConfigurations));
    }

    /* Configuring GPIOs for Analog In */
    MAP_GPIO_setAsPeripheralModuleFunctionInputPin(
            GPIO_PORT_P4,
            GPIO_PIN0 | GPIO_PIN4 | GPIO_PIN7,
            GPIO_TERTIARY_MODULE_FUNCTION);
    MAP_GPIO_setAsPeripheralModuleFunctionInputPin(
            GPIO_PORT_P5,
            GPIO_PIN1 | GPIO_PIN4 | GPIO_PIN5,
            GPIO_TERTIARY_MODULE_FUNCTION);
    MAP_GPIO_setAsPeripheralModuleFunctionInputPin(
            GPIO_PORT_P6,
            GPIO_PIN1,
            GPIO_TERTIARY_MODULE_FUNCTION);
    MAP_GPIO_setAsPeripheralModuleFunctionInputPin(
            GPIO_PORT_P8,
            GPIO_PIN3 | GPIO_PIN4 | GPIO_PIN6 | GPIO_PIN7,
            GPIO_TERTIARY_MODULE_FUNCTION);

    /* Initializing DRV device */
    drv8106_initializeDevice(DRV_SPI);

    /* Setting ADC in 12-bit mode sourced from 48MHz MCLK */
    MAP_ADC14_enableModule();
    MAP_ADC14_initModule(ADC_CLOCKSOURCE_MCLK, ADC_PREDIVIDER_1, ADC_DIVIDER_1,
                         0);
    MAP_ADC14_setResolution(ADC_12BIT);

    /* Configuring ADC Memory (Reduced Channels)
             * with 3.3v reference
             * 10A_DV2 -> P5.5 ->  A0   -> MEM0
             * 3A_DV1  -> P8.6 ->  A19  -> MEM1
             * 30A_CH2 -> P4.4 ->  A9   -> MEM2
     */
    MAP_ADC14_configureConversionMemory(ADC_MEM0, ADC_VREFPOS_AVCC_VREFNEG_VSS,
                                        ADC_INPUT_A0, false);
    MAP_ADC14_configureConversionMemory(ADC_MEM1, ADC_VREFPOS_AVCC_VREFNEG_VSS,
                                        ADC_INPUT_A19, false);
    MAP_ADC14_configureConversionMemory(ADC_MEM2, ADC_VREFPOS_AVCC_VREFNEG_VSS,
                                        ADC_INPUT_A9, false);

    /* Initializing sampling and timer semaphore */
    sem_init(&adcSem, 0, 0);
    sem_init(&tickSem, 0, 0);

    /* Enabling interrupts with the module, NVIC, and peripheral */
    hwiParams.priority = ~0;
    HwiP_create(INT_ADC14, intADCConversionComplete, &hwiParams);
    MAP_ADC14_enableInterrupt(ADC_INT2 | ADC_INT13);
    HwiP_enableInterrupt(INT_ADC14);

    /* Setting up the ADC sample timer to automatically step through the
     * sequence convert.
     */
    MAP_ADC14_enableSampleTimer(ADC_AUTOMATIC_ITERATION);
    ADC14->CTL0 = (ADC14->CTL0 & ~(ADC14_CTL0_CONSEQ_MASK))
                        | (ADC14_CTL0_CONSEQ_1) | (ADC14_CTL0_SHT0__192);

    /* Default everything as on */
    for(ii=0;ii<EFUSE_NUM_OF_CHANNELS;ii++)
    {
        GPIO_write(curStates[ii].gpioHandle,     1);
        GPIO_write(curStates[ii].diagnosticHandle, 1);
    }

    /* Starting the sample timer */
    TIMER32_1->CONTROL |= TIMER32_CONTROL_ENABLE;

    while (1)
    {
        /* Persist the values to info flash */
        if (persistValuesFlag == true)
        {
            persistValuesFlag = false;
            persistConfiguredValues();
        }

        /* Reset values with defaults from flash */
        if (resetValuesFlag == true)
        {
            resetValuesFlag = false;
            memcpy(loadedConfigurations, defaultConfigurations,
                   sizeof(loadedConfigurations));
        }

        /* The ON/OFF state changed */
        if (stateChange == true)
        {
            stateChange = false;
            mq_receive(stateChangeQueue, (void*) &curMessage, 2,
                       (unsigned int*) &prio);
            curStateChange = &curStates[curMessage[0]];

            GPIO_write(curStateChange->gpioHandle, curMessage[1]);
            GPIO_write(curStateChange->diagnosticHandle, curMessage[1]);

            /* If we are turning it off we have to reset the state */
            if(curMessage[1] == 0)
            {
                curStateChange->cooldownCounter = 0;
                curStateChange->runningSum = 0;
                curStateChange->softShutCounter = 0;
                curStateChange->state = EFUSE_STATE_NO_SHUTDOWN;
                curStateChange->skip = true;
            }
        }

        /* Wait for the timer tick to expire */
        sem_wait(&tickSem);

        /* Enabling diagnostics  and starting the conversion*/
        MAP_ADC14_enableConversion();
        MAP_ADC14_toggleConversionTrigger();

        /* Waiting for the ADC sample to complete */
        sem_wait(&adcSem);

        /* State Handling */
        for(ii=0;ii<EFUSE_NUM_OF_CHANNELS;ii++)
        {
            curResult = ADC14->MEM[ii];
            curFuse = &loadedConfigurations[ii];
            curParam = &curStates[ii];

            /* If it is the DRV device we have to subtract an offset */
            if(ii == EFUSE_DRV_ID)
            {
                curResult = (curResult - EFUSE_DRV_OFFSET);
            }

            if(curParam->skip == true)
            {
                curParam->skip = false;
                continue;
            }

            if ((curResult >= curFuse->triggerThreshold)
                    && (curResult < curFuse->softShutThreshold))
            {
                /* We are going from no shutdown to shutdown state */
                if (curParam->state == EFUSE_STATE_NO_SHUTDOWN)
                {
                    curParam->runningSum = (curResult*curResult);
                    curParam->state = EFUSE_STATE_SHUTDOWN;
                }
                else
                {
                    /* If we were in cooldown before */
                    if(curParam->state != EFUSE_STATE_SHUTDOWN)
                    {
                        curParam->state = EFUSE_STATE_SHUTDOWN;
                    }

                    /* Add the running sum */
                    curParam->runningSum = (curParam->runningSum
                            + (curResult * curResult));

                    /* Check to see if we tripped a shutdown */
                    if(curParam->runningSum > curFuse->shutdownThreshold)
                    {
                        /* Turn off the channel */
                        GPIO_write(curParam->gpioHandle, 0);
                        GPIO_write(curParam->diagnosticHandle, 0);

                        /* Enable Cooldown State */
                        curParam->cooldownCounter = 0;
                        curParam->state = EFUSE_STATE_COOLDOWN;
                    }
                }
            }
            /* We were in shutdown, but are now in cooldown */
            else if ((curResult < curFuse->triggerThreshold)
                    && (curParam->state == EFUSE_STATE_SHUTDOWN))
            {
                curParam->state = EFUSE_STATE_COOLDOWN;
                curParam->cooldownCounter = 0;
            }
            /* We are in active cooldown */
            else if ((curResult < curFuse->triggerThreshold)
                    && (curParam->state == EFUSE_STATE_COOLDOWN))
            {
                curParam->cooldownCounter = curParam->cooldownCounter + 1;

                if (curParam->cooldownCounter
                        == curFuse->cooldownCountThreshold)
                {
                    curParam->state = EFUSE_STATE_NO_SHUTDOWN;

                    /* Enabling diagnostics and the channel to retry */
                    GPIO_write(curParam->diagnosticHandle, 1);
                    GPIO_write(curParam->gpioHandle, 1);
                }
            }
            /* Soft fuse shutdown (fixed delay) */
            else if(curResult >= curFuse->softShutThreshold)
            {
                if(curParam->state != EFUSE_STATE_FIXED_DELAY_SHUTDOWN)
                {
                    curParam->state = EFUSE_STATE_FIXED_DELAY_SHUTDOWN;
                    curParam->softShutCounter = 0;
                }
                else
                {
                    curParam->softShutCounter = curParam->softShutCounter + 1;

                    if (curParam->softShutCounter
                            == curFuse->softCountThreshold)
                    {
                        GPIO_write(curParam->gpioHandle, 0);
                        curParam->state = EFUSE_STATE_COOLDOWN;
                        curParam->cooldownCounter = 0;
                    }
                }
            }
        }
    }
}

void persistConfiguredValues(void)
{
    // TODO: Implement saving values to flash
}

void intADCConversionComplete(uintptr_t arg)
{
    /* Sampling has completed, post semaphore */
    sem_post(&adcSem);
}

void intTimerTickComplete(uintptr_t arg)
{
    /* Clearing the IFG and posting the semaphore */
    sem_post(&tickSem);
}
