/*
 * Copyright (c) 2024, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
#include <stdint.h>
#include <stdbool.h>

/* Drivers/Device Includes */
#include "inc/hw_gpio.h"
#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "inc/hw_sysctl.h"
#include "driverlib/sysctl.h"
#include "driverlib/gpio.h"
#include "driverlib/pin_map.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "usblib/usblib.h"
#include "usblib/usbhid.h"
#include "usblib/device/usbdevice.h"
#include "usblib/device/usbdhid.h"
#include "utils/ustdlib.h"
#include "usb_spi_dev.h"
#include "usb_task_handler.h"

/* RTOS header files */
#include <FreeRTOS.h>
#include <task.h>
#include <queue.h>
#include <string.h>

/* RTOS Elements */
QueueHandle_t xCommandQueue;
void mainHandlerTask(void *pvParameters);
extern QueueHandle_t xHSSCommandQueue;

/* Global connect/suspended variables */
volatile bool g_bConnected = false;
volatile bool g_bSuspended = false;

//*****************************************************************************
//
// This enumeration holds the various states that the keyboard can be in during
// normal operation.
//
//*****************************************************************************
volatile enum
{
    //
    // Unconfigured.
    //
    STATE_UNCONFIGURED,

    //
    // No keys to send and not waiting on data.
    //
    STATE_IDLE,

    //
    // Waiting on data to be sent out.
    //
    STATE_SENDING
}
g_eSwitchState = STATE_UNCONFIGURED;

/* Static Variables */
volatile bool hidDataToWrite = false;

//*****************************************************************************
//
// Handles asynchronous events from the HID keyboard driver.
//
// \param pvCBData is the event callback pointer
// \param ui32Event identifies the event we are being called back for.
// \param ui32MsgData is an event-specific value.
// \param pvMsgData is an event-specific pointer.
//
// This function is called by the HID keyboard driver to inform the application
// of particular asynchronous events related to operation of the keyboard HID
// device.
//
// \return Returns 0 in all cases.
//
//*****************************************************************************
uint32_t
HighSideSwitchHandler(void *pvCBData, uint32_t ui32Event, uint32_t ui32MsgData,
                void *pvMsgData)
{
    /* Command Handler */
    tUSBHostCommand tempCommand;

    switch (ui32Event)
    {
        //
        // The host has connected to us and configured the device.
        //
        case USB_EVENT_CONNECTED:
        {
            tempCommand.opcode = USB_HSS_CONNECT_EVENT;
            tempCommand.payload[0] = USB_HSS_CONNECT_CON;
            xQueueSendFromISR(xCommandQueue, &tempCommand, NULL);
            break;
        }

        //
        // The host has disconnected from us.
        //
        case USB_EVENT_DISCONNECTED:
        {
            tempCommand.opcode = USB_HSS_CONNECT_EVENT;
            tempCommand.payload[0] = USB_HSS_CONNECT_DISCON;
            xQueueSendFromISR(xCommandQueue, &tempCommand, NULL);
            break;
        }

        //
        // We receive this event every time the host acknowledges transmission
        // of a report.  It is used here purely as a way of determining whether
        // the host is still talking to us or not.
        //
        case USB_EVENT_TX_COMPLETE:
        {
            //
            // Enter the idle state since we finished sending something.
            //
            g_eSwitchState = STATE_IDLE;
            break;
        }

        //
        // This event indicates that the host has suspended the USB bus.
        //
        case USB_EVENT_SUSPEND:
        {
            tempCommand.opcode = USB_HSS_SUS_EVENT;
            tempCommand.payload[0] = USB_HSS_SUS_SUSPEND;
            xQueueSendFromISR(xCommandQueue, &tempCommand, NULL);
            break;
        }

        //
        // This event signals that the host has resumed signalling on the bus.
        //
        case USB_EVENT_RESUME:
        {
            tempCommand.opcode = USB_HSS_SUS_EVENT;
            tempCommand.payload[0] = USB_HSS_SUS_RESUME;
            xQueueSendFromISR(xCommandQueue, &tempCommand, NULL);
            break;
        }
        
        /* Send HSS commands directly over to the HSS task handler */
        case HSS_COMMAND_RECEIVED:
        {
            xQueueSendFromISR(xHSSCommandQueue, ((uint8_t*)pvMsgData)+4, NULL);
            break;
        }

        //
        // We ignore all other events.
        //
        default:
        {
            break;
        }
    }

    return(0);
}


void mainHandlerTask(void *pvParameters)
{
    volatile uint32_t dataSent;
    tUSBHostCommand curCommand;

    //
    // Configure the required pins for USB operation.
    //
    MAP_SysCtlPeripheralEnable(SYSCTL_PERIPH_GPIOD);
    MAP_GPIOPinTypeUSBAnalog(GPIO_PORTD_BASE, GPIO_PIN_4 | GPIO_PIN_5);

    /* Create the queue used by the queue send and queue receive tasks. */
    xCommandQueue = xQueueCreate(USB_HSS_QUEUE_LENGTH, sizeof(curCommand));

    //
    // Not configured initially.
    //
    g_bConnected = false;
    g_bSuspended = false;

    //
    // Initialize the USB stack for device mode.  We do not operate in USB
    // device mode with active monitoring of VBUS and therefore, we will
    // specify eUSBModeForceDevice as the operating mode instead of
    // eUSBModeDevice.  To use eUSBModeDevice, the EK-TM4C123GXL LaunchPad
    // must have the R28 and R29 populated with zero ohm resistors.
    //
    USBStackModeSet(0, eUSBModeForceDevice, 0);

    //
    // Pass our device information to the USB HID device class driver,
    // initialize the USB controller, and connect the device to the bus.
    //
    USBHID_HSS_DeviceInit();

    //
    // The main loop starts here. 
    //
    while(1)
    {
        curCommand.opcode = USB_HSS_NO_EVENT;

        /* Wait here until we get a connect event and disregard to
            anything that is not a connect event */
        while(curCommand.opcode != USB_HSS_CONNECT_EVENT &&
            curCommand.payload[0] != USB_HSS_CONNECT_CON)
        {
            xQueueReceive(xCommandQueue, &curCommand, portMAX_DELAY);
            g_bConnected = true;
        }

        /* Enter idle state */
        g_eSwitchState = STATE_IDLE;

        /* While we are connected... */
        while(g_bConnected)
        {
            /* Wait for the next command */
            xQueueReceive(xCommandQueue, &curCommand, portMAX_DELAY);

            /* Command handler */
            switch(curCommand.opcode)
            {
                case USB_HSS_CONNECT_EVENT:
                    if(curCommand.payload[0] == USB_HSS_CONNECT_DISCON)
                    {
                        g_bConnected = false;
                        break;
                    }
                    break;
                case USB_HSS_SUS_EVENT:
                    if(curCommand.payload[0] == USB_HSS_SUS_SUSPEND)
                    {
                        g_bSuspended = true;
                    }
                    else
                    {
                        g_bSuspended = false;
                    }
                    break;
                default:
                    break;
                
            }
        }
    }
}
