/*
 * Copyright (c) 2024, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "AutoConfigurationGui.h"
#include <Dbt.h>
#include <initguid.h>
#include <usbiodef.h>
#include <vcclr.h>
#include <iostream>
#include <sstream>
#include <bitset>
#include <cliext/queue>
#include <msclr\marshal.h>
#include "hssc_defines.h"
#include "language_localizations.h"

using namespace System;
using namespace System::Runtime::InteropServices;

/* Generic function that logs provided message to the "log" view in the GUI with a timestamped
    prefix. */
void autohssconfigurator::AutoConfigurationGui::logMessageToWindow(String^ str, ... array<Object^>^ args)
{
    if (demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        return;
    }

    String^ tempString = String::Format(str, args);
    String^ dateString = DateTime::Now.ToString("[hh:mm:ss] ");

    this->logTextBox->AppendText(dateString + tempString + Environment::NewLine);

    if ((manualPersistEnableCheckBox->CheckState == CheckState::Unchecked) && (dirtyPersist == true))
    {
        if (!commandStatusLabel->Text->Contains("*Persist Pending*"))
        {
            commandStatusLabel->Text = commandStatusLabel->Text + "*Persist Pending*";
        }
    }
    else
    {

        commandStatusLabel->Text = tempString;
    }
}

/* Callback for handling Windows messages. This is mainly done to detect connect and
    disconnect events*/
INT_PTR WINAPI WinProcCallback(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    LRESULT lRet = 1;

    switch (message)
    {
    case WM_CREATE:
        break;
    case WM_DESTROY:
        break;
    case WM_DEVICECHANGE:
        switch (wParam)
        {
        case DBT_DEVICEREMOVECOMPLETE:
        case DBT_DEVICEARRIVAL:
        {
            PDEV_BROADCAST_DEVICEINTERFACE p = (PDEV_BROADCAST_DEVICEINTERFACE)lParam;
            if (p->dbcc_devicetype == DBT_DEVTYP_DEVICEINTERFACE &&
                p->dbcc_classguid == GUID_DEVINTERFACE_USB_DEVICE)
            {
                if (wcsstr(p->dbcc_name, L"#VID_1CBE&PID_000D#"))
                {
                    if (wParam == DBT_DEVICEARRIVAL)
                    {
                        autohssconfigurator::AutoConfigurationGui::connQueue.Enqueue(0x01);
                    }
                    else
                    {
                        autohssconfigurator::AutoConfigurationGui::connQueue.Enqueue(0x00);
                    }
                }
            }
        }
        break;
        }
        break;
    default:
        lRet = DefWindowProc(hWnd, message, wParam, lParam);
        break;
    }
    return lRet;
}

void autohssconfigurator::AutoConfigurationGui::updateMainGUIFromSelection(void)
{
    tHSSReceiveTransaction^ trans;
    uint8_t payload[3];

    /* Writing a dummy write command to take the device out of sleep mode if it is in there */
    payload[0] = HSSC_RAW_WRITE_SIZE;
    payload[1] = 0xFF;
    payload[2] = 0xFF;
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, 0xFF, payload, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);
    Sleep(1);

    /* Telling the HSS to update the configuration data for the selected address */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    /* Getting the initial device data */
    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the initial channel 1 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the initial channel 2 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 1, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 1;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Pinging for the firmware version */
    payload[0] = HSSCommand::HSS_IOCTL_GETVERSION;
    cmd = new HSSCommand(HSSCommand::HSS_HCS_IOCTL, payload, 1);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_IOCTL_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_IOCTL_RESP_SIZE;
    this->commManager->getPacketData(trans);
}

void autohssconfigurator::AutoConfigurationGui::updateConnectionStatus()
{
    if (this->InvokeRequired)
    {
        UpdateConnnectionCallback^ updateCallback = gcnew UpdateConnnectionCallback(this,
            &AutoConfigurationGui::AutoConfigurationGui::updateConnectionStatus);
        this->Invoke(updateCallback);
    }
    else
    {
        if (commManager->isHidDeviceConnected())
        {
            connectionLabel->ForeColor = System::Drawing::Color::Green;
            connectionLabel->Text = curConnectWord;
            mainTabControl->Enabled = true;
            logMessageToWindow("High-side switch device found and connected.");
            updateMainGUIFromSelection();
            settingsToolStripMenuItem->Enabled = true;
            toolsToolStripMenuItem->Enabled = true;
            saveSettingsToolStripMenuItem->Enabled = true;
            loaToolStripMenuItem->Enabled = true;
            exportToolStripMenuItem->Enabled = true;
            demoModeToolStripMenuItem->Enabled = false;
        }
        else
        {
            connectionLabel->ForeColor = System::Drawing::Color::Red;
            connectionLabel->Text = curdisConnectWord;
            mainTabControl->Enabled = false;
            this->commManager->openDevice();
            logMessageToWindow("High-side switch device disconnected.");
            settingsToolStripMenuItem->Enabled = false;
            toolsToolStripMenuItem->Enabled = false;
            saveSettingsToolStripMenuItem->Enabled = false;
            loaToolStripMenuItem->Enabled = false;
            exportToolStripMenuItem->Enabled = false;
            demoModeToolStripMenuItem->Enabled = true;
        }
    }
}

/* Thread that polls for connection events */
void autohssconfigurator::AutoConfigurationGui::connectionThreadFunction(Object^ obj)
{
    AutoConfigurationGui^ guiObj = (AutoConfigurationGui^)obj;
    uint8_t status;

    /* Queuing a single connect event on startup */
    connQueue.Enqueue(0x01);

    while (1)
    {
        Sleep(200);
        if (connQueue.Count != 0)
        {
            status = connQueue.Dequeue();
            /* Disconnect Event */
            if (status == 0)
            {
                guiObj->commManager->closeDevice();
            }
            /* Connect Event */
            else
            {
                guiObj->commManager->openDevice();
            }

            /* Update the GUI for the updated connection status */
            guiObj->updateConnectionStatus();
        }

        /* See if the form is closing */
        if (guiObj->isShuttingDown)
        {
            return;
        }
    }
}

/* Thread function that polls for faults */
void autohssconfigurator::AutoConfigurationGui::faultReadFunctionThread(Object^ obj)
{
    HSSCommand* faultStatus;
    tHSSReceiveTransaction^ faultTrans;
    AutoConfigurationGui^ guiObj = (AutoConfigurationGui^)obj;

    while (1)
    {
        Sleep(FLT_DATA_PERIOD_MS);

        if (guiObj->commManager->isHidDeviceConnected() && (manualUpdateEnabled == false))
        {

            /* Queuing a fault update */
            faultStatus = new HSSCommand(HSSCommand::HSS_HCS_FAULT_CONFIG, 0, NULL, 0);
            guiObj->commManager->sendPacketData(faultStatus->getBufferData(), faultStatus->getCommandSize());

            /* Getting the updated fault status */
            faultTrans = gcnew tHSSReceiveTransaction;
            faultTrans->addr = 0;
            faultTrans->recData = (uint8_t*)malloc(128);
            faultTrans->reportID = HSSCommunicationManager::HSSC_FAULT_ID;
            faultTrans->receiveSize = HSSCommand::HSS_TPS274C65_FAULT_STATE_CONFIG_RESP_SIZE;
            guiObj->commManager->getPacketData(faultTrans);

            free(faultStatus);
        }

        /* See if the form is closing */
        if (guiObj->isShuttingDown)
        {
            return;
        }
    }
}

/* Thread Function that polls for sense data */
void autohssconfigurator::AutoConfigurationGui::senseReadFunctionThread(Object^ obj)
{
    AutoConfigurationGui^ guiObj = (AutoConfigurationGui^)obj;
    tHSSReceiveTransaction^ senseTrans;

    while (1)
    {
        Sleep(SNS_DATA_PERIOD_MS);

        if (guiObj->commManager->isHidDeviceConnected())
        {
            if (guiObj->isCurrentSenseEnabled)
            {
                /* Getting the sense data */
                senseTrans = gcnew tHSSReceiveTransaction;
                senseTrans->addr = 0;
                senseTrans->recData = (uint8_t*)malloc(128);
                senseTrans->reportID = HSSCommunicationManager::HSSC_SNS_ID;
                senseTrans->receiveSize = HSSCommand::HSS_HCS_SNS_DATA_RESP_SIZE;
                guiObj->commManager->getPacketData(senseTrans);
            }
        }

        /* See if the form is closing */
        if (guiObj->isShuttingDown)
        {
            return;
        }
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::commandBox_KeyDown(System::Object^ sender, System::Windows::Forms::KeyEventArgs^ e)
{
    if (e->KeyCode == Keys::Enter)
    {
        sendCommandButton_Click(sender, e);
    }
}

int autohssconfigurator::AutoConfigurationGui::char2int(char input)
{
    if (input >= '0' && input <= '9')
        return input - '0';
    if (input >= 'A' && input <= 'F')
        return input - 'A' + 10;
    if (input >= 'a' && input <= 'f')
        return input - 'a' + 10;
    return 0;
}

/* This function assumes src to be a zero terminated sanitized string with
  an even number of [0-9a-f] characters, and target to be sufficiently large */
void autohssconfigurator::AutoConfigurationGui::hex2bin(const char* src, uint8_t* target)
{
    while (*src && src[1])
    {
        *(target++) = char2int(*src) * 16 + char2int(src[1]);
        src += 2;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::sendCommandButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    String^ curCommand = commandBox->Text;
    array<String^>^ splitString = curCommand->Split(' ');
    uint8_t curOpCode;
    uint8_t readRegAddr;
    uint16_t writeBuffer;
    String^ tempString;
    String^ tempString2;
    char* convertedString;
    uint8_t payload[3] = { 0,0 };
    HSSCommand* cmd;
    tHSSReceiveTransaction^ readTrans;
    String^ dateString = DateTime::Now.ToString("[hh:mm:ss] ");

    if (splitString->Length == 0)
    {
        commandLogBox->AppendText("INVALID COMMAND: " + curCommand + Environment::NewLine);
        return;
    }

    if (splitString[0]->ToUpper() == "READ")
    {
        if (splitString->Length != 2)
        {
            commandLogBox->AppendText("INVALID FORMAT! Read Command:  READ <REG_NUM_HEX>" + Environment::NewLine);
            return;
        }
        curOpCode = HSSCommand::HSS_HCS_RAW_READ;

        tempString = splitString[1]->Replace("0x", "");

        if (tempString->Length == 1)
        {
            tempString = "0" + tempString;
        }
        else if (tempString->Length > 2)
        {
            tempString = tempString->Substring(tempString->Length - 2);
        }

        convertedString = (char*)(void*)Marshal::StringToHGlobalAnsi(tempString);
        hex2bin(convertedString, &readRegAddr);
        
        /* Currently, the device only supports 16-bit reads */
        payload[0] = 2;

        cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_READ, readRegAddr, payload, 2);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
        commandLogBox->AppendText(dateString + "READ REGISTER: 0x" + tempString + "\r\n");

        readTrans = gcnew tHSSReceiveTransaction;
        readTrans->recData = (uint8_t*)malloc(128);
        readTrans->reportID = HSSCommunicationManager::HSSC_RAWREAD_ID;
        readTrans->receiveSize = HSSCommand::HSS_HCS_RAW_READ_RESP_SIZE;
        commManager->getPacketData(readTrans);

        if (cmdHistoryBox->Items->Count == 0 || (cmdHistoryBox->SelectedItem->ToString() != commandBox->Text))
        {
            cmdHistoryBox->Items->Add(commandBox->Text);
            cmdHistoryBox->SelectedIndex = cmdHistoryBox->Items->Count - 1;
        }

    }
    else if (splitString[0]->ToUpper() == "WRITE")
    {
        if (splitString->Length != 3)
        {
            commandLogBox->AppendText("INVALID FORMAT! Read Command:  WRITE <REG_NUM_HEX> <REG_WRITE_DATA_HEX>" + Environment::NewLine);
            return;
        }

        curOpCode = HSSCommand::HSS_HCS_RAW_WRITE;

        /* Currently, the device only supports 16-bit writes so setting the legnth byte to 2 */
        payload[0] = 2;

        tempString = splitString[1]->Replace("0x", "");

        if (tempString->Length == 1)
        {
            tempString = "0" + tempString;
        }
        else if (tempString->Length > 2)
        {
            tempString = tempString->Substring(tempString->Length - 2);
        }

        convertedString = (char*)(void*)Marshal::StringToHGlobalAnsi(tempString);
        hex2bin(convertedString, &readRegAddr);
        
        tempString2 = splitString[2]->Replace("0x", "");

        while(tempString2->Length < 4)
        {
            tempString2 = "0" + tempString2;
        }
        
        if (tempString2->Length > 4)
        {
            tempString2 = tempString2->Substring(tempString2->Length - 4);
        }

        convertedString = (char*)(void*)Marshal::StringToHGlobalAnsi(tempString2);
        hex2bin(convertedString, (uint8_t*)&writeBuffer);
        payload[2] = writeBuffer & 0xFF;
        payload[1] = writeBuffer >> 8;

        cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, readRegAddr, payload, 3);
        commandLogBox->AppendText(dateString + "WRITE REGISTER: 0x" + tempString + " 0x" + tempString2 + "\r\n");
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        if (cmdHistoryBox->Items->Count == 0 || (cmdHistoryBox->SelectedItem->ToString() != commandBox->Text))
        {
            cmdHistoryBox->Items->Add(commandBox->Text);
            cmdHistoryBox->SelectedIndex = cmdHistoryBox->Items->Count - 1;
        }

    }
    else
    {
        commandLogBox->AppendText("INVALID COMMAND: " + splitString[0] + Environment::NewLine);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::AutoConfigurationGui_Load(System::Object^ sender, System::EventArgs^ e)
{
    /* Registering the program for Windows kernel notifications so that we can know when a device was connected
        and disconnected. */
    WNDCLASSEX wndClass = { 0 };
    bool isClassRegistered = false;
    DEV_BROADCAST_DEVICEINTERFACE NotificationFilter = {
        sizeof(DEV_BROADCAST_DEVICEINTERFACE),
        DBT_DEVTYP_DEVICEINTERFACE,
        0,
        GUID_DEVINTERFACE_USB_DEVICE
    };

    wndClass.cbSize = sizeof(WNDCLASSEX);
    wndClass.hInstance = reinterpret_cast<HINSTANCE>(GetModuleHandle(0));
    wndClass.lpfnWndProc = reinterpret_cast<WNDPROC>(WinProcCallback);
    wndClass.lpszClassName = L"message_class";

    isClassRegistered = RegisterClassEx(&wndClass);

    messageWindow = CreateWindowEx(0, L"message_class", L"message_window", 0, 0, 0, 0, 0,
        HWND_MESSAGE, NULL, GetModuleHandle(0), NULL);

    /* Register for connection events */
    RegisterDeviceNotification(messageWindow, &NotificationFilter, DEVICE_NOTIFY_WINDOW_HANDLE);

    /* Populating the default values */
    populateDefaultValues();

    /* Initializing local variables and starting threads */
    this->commManager = gcnew HSSCommunicationManager();
    this->commManager->setVerboseMode(true);
    connThreadStart = gcnew ParameterizedThreadStart(autohssconfigurator::AutoConfigurationGui::connectionThreadFunction);
    connThread = gcnew Thread(connThreadStart);
    connThread->Start(this);

    /* Setting up the thread for sensing */
    senseThreadStart = gcnew ParameterizedThreadStart(autohssconfigurator::AutoConfigurationGui::senseReadFunctionThread);
    senseThread = gcnew Thread(senseThreadStart);
    senseThread->Start(this);

    /* Setting up the thread to read fault status */
    faultThreadStart = gcnew ParameterizedThreadStart(autohssconfigurator::AutoConfigurationGui::faultReadFunctionThread);
    faultThread = gcnew Thread(faultThreadStart);
    faultThread->Start(this);

    /* Registering the receive Handler */
    recHandler = gcnew HSSDeviceReceiveFinished(this, &AutoConfigurationGui::bufferReceiveEventHandler);
    commManager->hssReceiveBufferDelegate += recHandler;

    this->fuseConfig = gcnew TPS2HCS10Configuration();

    /* Initializing images */
    redButton = Image::FromFile("images/redBut.png");
    greenButton = Image::FromFile("images/greenBut.png");
}

void autohssconfigurator::AutoConfigurationGui::populateDefaultValues(void)
{
    /* Array values for I2T trip values */
    i2tArrayValues = gcnew array<double>{8.8, 13.1, 26.3, 39.4, 52.5, 65.6, 78.8, 91.9, 109.4, 126.9, 144.4, 166.3, 192.5, 218.8, 262.5, 350};
    i2tChart->ChartAreas[0]->AxisX->Minimum = 0;

    /* Default 20-gauge wire values */
    twentyGaugeWireValues = gcnew array<float,2>
    {
        {5.0f *1.3f, 5.5f * 1.3f, 6.5f * 1.3f, 7.5f * 1.3f, 8.5f * 1.3f, 9.5f * 1.3f, 10.5f * 1.3f, 11.5f * 1.3f, 12.5f * 1.3f,
            13.5f * 1.3f, 14.5f * 1.3f, 15.5f * 1.3f, 16.5f * 1.3f, 17.5f * 1.3f, 19.0f * 1.3f},
        {100.0f, 10.0f, 4.5f, 3.2f, 2.1f, 1.8f, 1.5f, 1.2f, 1.1f, 1.0f, 0.98f, 0.90f, 0.8f, 0.78f, 0.5f}
    };

    /* Default 14-gauge wire values */
    sixteenGaugeWireValues = gcnew array<float, 2>
    {
        {5.0f * 2.0f, 5.5f * 2.0f, 6.5f * 2.0f, 7.5f * 2.0f, 8.5f * 2.0f, 9.5f * 2.0f, 10.5f * 2.0f, 11.5f * 2.0f, 
            12.5f * 2.0f, 13.5f * 2.0f, 14.5f * 2.0f, 15.5f * 2.0f, 16.5f * 2.0f, 17.5f * 2.0f, 19.0f * 2.0f},
        { 100.0f, 10.0f, 4.5f, 3.2f, 2.1f, 1.8f, 1.5f, 1.2f, 1.1f, 1.0f, 0.98f, 0.90f, 0.8f, 0.78f, 0.5f }
    };

    avgValsCh1 = gcnew System::Collections::Generic::List<float>();
    avgValsCh2 = gcnew System::Collections::Generic::List<float>();

    dontRunEvents = true;
    chanSettingBox->SelectedIndex = 0;
    i2tChanNumBox->SelectedIndex = 0;
    senseSourceSelBox->SelectedIndex = 0;
    gaugeSelectionBox->SelectedIndex = 0;
    manualUpdateEnabled = false;
    faultUpdateStatus->SelectedIndex = 0;
    i2tChart->ChartAreas[0]->Axes[1]->Crossing = Double::MinValue;
    i2tCompleteChart->ChartAreas[0]->Axes[1]->Crossing = Double::MinValue;
    dontRunEvents = false;
}

void autohssconfigurator::AutoConfigurationGui::redrawI2TGraph(void)
{
    double curI2TValue = 0.0;
    double i2tTripValue = 0.0;
    double nominalValue = 0.0;
    double swLimitValue = 0.0;
    double hwLimitValue = 0.0;
    double swDelayValue = 0.0;
    double incrementValue = 0.0;
    double curCurrentValue = 0.0;
    double curSNSValue = 700.0;
    double curYMax = 5.0;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Clearing out the current values for all series */
    for (int ii = 0; ii < 3; ii++)
    {
        while (i2tChart->Series[ii]->Points->Count != 0)
        {
            i2tChart->Series[ii]->Points->RemoveAt(0);
        }
    }

    for (int ii = 0; ii < i2tCompleteChart->Series->Count; ii++)
    {
        while (i2tCompleteChart->Series[ii]->Points->Count != 0)
        {
            i2tCompleteChart->Series[ii]->Points->RemoveAt(0);
        }
    }

    i2tTripValue = i2tArrayValues[i2tTrack->Value];
    nominalValue = Double::Parse(nomCurrentBox->SelectedItem->ToString());
    swLimitValue = Double::Parse(swLimitBox->SelectedItem->ToString());
    hwLimitValue = Double::Parse(hwLimitBox->SelectedItem->ToString());
    swDelayValue = Double::Parse(swDelayBox->SelectedItem->ToString())/1000.0;

    /* Plotting the I2T Curve */
    incrementValue = (swLimitValue - nominalValue) / 1000.0;

    /* Priming the graph and drawing the maximum for nominal current */
    curCurrentValue = nominalValue;
    curI2TValue = i2tTripValue;

    /* Nominal Current */
    i2tCompleteChart->Series[0]->Points->AddXY(0, 10000);
    i2tCompleteChart->Series[0]->Points->AddXY(curCurrentValue, 10000);

    i2tChart->Series[0]->Points->AddXY(curCurrentValue, curI2TValue);
    i2tChart->ChartAreas[0]->AxisX->Minimum = curCurrentValue;

    for (int ii = 1; ii < 1000; ii++)
    {
        curCurrentValue = nominalValue + (ii*incrementValue);
        curI2TValue = i2tTripValue / ((curCurrentValue * curCurrentValue) - (nominalValue * nominalValue));
        i2tChart->Series[0]->Points->AddXY(curCurrentValue, curI2TValue);
        i2tChart->Series[1]->Points->AddXY(curCurrentValue*(0.94 - (curResTolerance/100.0f)), curI2TValue * 0.92);
        i2tChart->Series[2]->Points->AddXY(curCurrentValue*(1.06 + (curResTolerance/100.0f)), curI2TValue * 1.08);
        i2tCompleteChart->Series[1]->Points->AddXY(curCurrentValue, curI2TValue);
        if (curI2TValue > curYMax)
        {
            curYMax = curI2TValue;
        }
    }

    i2tCompleteChart->Series[1]->Points[0]->Label = nominalValue + "A";
    i2tCompleteChart->Series[1]->Points[500]->Label = i2tTripValue + " A2s";

    if (swLimitValue < hwLimitValue)
    {
        i2tCompleteChart->Series[2]->Points->AddXY(curCurrentValue, swDelayValue);
        i2tCompleteChart->Series[2]->Points[0]->Label = "SW: " + swLimitValue + "A";

        i2tCompleteChart->Series[2]->Points->AddXY(swLimitValue, swDelayValue);
        i2tCompleteChart->Series[2]->Points->AddXY(hwLimitValue, swDelayValue);

        i2tCompleteChart->Series[3]->Points->AddXY(hwLimitValue, 0.0001);
        i2tCompleteChart->Series[3]->Points->AddXY(hwLimitValue * 1.25, 0.0001);
        i2tCompleteChart->Series[3]->Points[0]->Label = "HW: " + hwLimitValue + "A";
    }

    /* Extrapolating the I2T minimums */
    while ((curCurrentValue * 0.94) < swLimitValue)
    {
        curCurrentValue += incrementValue;
        curI2TValue = i2tTripValue / ((curCurrentValue * curCurrentValue) - (nominalValue * nominalValue));
        i2tChart->Series[1]->Points->AddXY(curCurrentValue * 0.94, curI2TValue * 0.92);
    }

    i2tChart->Series[2]->Points[500]->Label = i2tTripValue + " A2s";

    /* Adjusting the maximum/scales */
    i2tChart->ChartAreas[0]->AxisX->Maximum = swLimitValue;

    if (gaugeSelectionBox->SelectedIndex == 0)
    {
        i2tChart->ChartAreas[0]->AxisY->Maximum = curYMax;
    }
    else
    {
        i2tChart->ChartAreas[0]->AxisY->Maximum = 4000;
    }

    i2tCompleteChart->ChartAreas[0]->AxisX->Maximum = hwLimitValue * 1.1;
    i2tCompleteChart->ChartAreas[0]->AxisY->Maximum = curYMax;

}

System::Void autohssconfigurator::AutoConfigurationGui::ConfigurationGui_FormClosing(System::Object^ sender, System::Windows::Forms::FormClosingEventArgs^ e)
{
    isShuttingDown = true;

    /* Waiting for the threads to end */
    while (connThread->IsAlive);

    /* Removing the receive handler delegate */
    commManager->hssReceiveBufferDelegate -= recHandler;

    /* Save the output buffer if it is closing */
    if (fileLogger != nullptr)
    {
        fileLogger->Close();
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::sendRawWriteGUIResponse(HSSCommand* cmd)
{
    if (manualPersistEnabled == false)
    {
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    }
    else
    {
        dirtyPersist = true;
        if (!commandStatusLabel->Text->Contains("*Persist Pending*"))
        {
            commandStatusLabel->Text = commandStatusLabel->Text + "*Persist Pending*";
        }
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tTrack_Scroll(System::Object^ sender, System::EventArgs^ e)
{
    i2tTripValueLabel->Text = i2tArrayValues[i2tTrack->Value].ToString() + " A2s";

    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setI2TConfig((i2tTrack->Value << TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("I2T trip value of channel {0} set to {1}", curChan + 1, i2tTripValueLabel->Text);

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::nomCurrentBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }
    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setI2TConfig((nomCurrentBox->SelectedIndex) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_NOM_CUR_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Nominal current of channel {0} set to {1}", curChan + 1, nomCurrentBox->SelectedItem->ToString());

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::hwLimitBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setILIMConfig((hwLimitBox->SelectedIndex) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Hardware current limit of channel {0} set to {1}", curChan + 1, hwLimitBox->SelectedItem->ToString());

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::nomCurrentBox_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "I2T_CONFIG_CHx:NOM_CUR_CHx";
    helpLabel->Text = "Sets the nomial current value of the current channel for I2T calculations only.";
}

System::Void autohssconfigurator::AutoConfigurationGui::nomCurrentLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tTripValueLabel_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "I2T_CONFIG_CHx:I2T_TRIP_CHx";
    helpLabel->Text = "Sets ISWCL value for I2T functionality for current channel.";
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tTripLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::swCurrentLabel_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "I2T_CONFIG_CHx:ISWCL_CHx";
    helpLabel->Text = "Sets the delayed shutdown threshold current that kicks in after the I2T curve.";
}

System::Void autohssconfigurator::AutoConfigurationGui::swCurrentLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::swDelayLabel_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "I2T_CONFIG_CHx:SWCL_DLY_TMR_CHx";
    helpLabel->Text = "Sets the software current limit delay timer. Sets the time after which the channel shuts down when the current exceeds ISWCL.";
}

System::Void autohssconfigurator::AutoConfigurationGui::swDelayLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::cooldownLabel_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "I2T_CONFIG_CHx:TCLDN_CHx";
    helpLabel->Text = "Sets the cooldown time for I2T functionality for the current channel (time to retry after I2T shutdown).";
}

System::Void autohssconfigurator::AutoConfigurationGui::cooldownLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::hwLimitLabel_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "ILIM_CONFIG_CHx:ILIMIT_SET_CHx";
    helpLabel->Text = "Sets the hardware current limit of the channel. This is independent of the I2T settings.";
}

System::Void autohssconfigurator::AutoConfigurationGui::hwLimitLabel_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::enChanBoxCC_ItemCheck(System::Object^ sender, System::Windows::Forms::ItemCheckEventArgs^ e)
{
    uint16_t selectedItems = 0;
    UInt32 count = enChanBoxCC->Items->Count;
    uint8_t payload[3];

    if (dontRunEvents == true)
    {
        return;
    }

    selectedItems = (1 << e->Index);

    if (enChanBoxCC->GetItemChecked(0) == true)
    {
        selectedItems |= 0b0001;
    }

    if (enChanBoxCC->GetItemChecked(1) == true)
    {
        selectedItems |= 0b0010;
    }

    /* Updating with the new state */
    if (e->NewValue == CheckState::Checked)
    {
        selectedItems |= (1 << e->Index);
    }
    else
    {
        selectedItems &= ~(1 << e->Index);
    }

    /* Sending the enable command */
    payload[0] = HSSCommand::HSS_HCS_RAW_WRITE_PL_SIZE;
    payload[1] = selectedItems & 0xFF;
    payload[2] = selectedItems >> 8;
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_SW_STATE_REG, payload, 3);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
    fuseConfig->setSwitchState(selectedItems);

    // TODO: Add code here that enables/disables setting controls based on which EN button was clicked

    logMessageToWindow("Switch enable configuration changed to: 0b{0} ", gcnew String(std::bitset<8>(selectedItems).to_string().data()));
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdHistoryBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    commandBox->Text = cmdHistoryBox->SelectedItem->ToString();
}

System::Void autohssconfigurator::AutoConfigurationGui::parallelBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (parallelBox->Checked)
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() | TPS2HC10S_DEV_CONFIG_PARALLEL_12_MASK);
        logMessageToWindow("Parallel channel option enabled.");
    }
    else
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() & ~TPS2HC10S_DEV_CONFIG_PARALLEL_12_MASK);
        logMessageToWindow("Parallel channel option disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = {2, (uint8_t)(fuseConfig->getDevConfig() & 0xFF), (uint8_t)(fuseConfig->getDevConfig() >> 8)};

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::wdEnableBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (wdEnableBox->Checked)
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() | TPS2HC10S_DEV_CONFIG_WD_EN_MASK);
        logMessageToWindow("Watchdog timer enabled.");
    }
    else
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() & ~TPS2HC10S_DEV_CONFIG_WD_EN_MASK);
        logMessageToWindow("Watchdog timer disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevConfig() & 0xFF), (uint8_t)(fuseConfig->getDevConfig() >> 8)};

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::wdtDurationBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setDevConfig((wdtDurationBox->SelectedIndex << 1) | (fuseConfig->getDevConfig() & ~TPS2HC10S_DEV_CONFIG_WD_TO_MASK));
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevConfig() & 0xFF), (uint8_t)(fuseConfig->getDevConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Watchdog duration changed to {0}", wdtDurationBox->SelectedIndex);
}

System::Void autohssconfigurator::AutoConfigurationGui::pwmDelayBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (pwmDelayBox->Checked)
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() & ~TPS2HC10S_DEV_CONFIG_PWM_SHIFT_DIS_MASK);
        logMessageToWindow("PWM delay enabled.");
    }
    else
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() | TPS2HC10S_DEV_CONFIG_PWM_SHIFT_DIS_MASK);
        logMessageToWindow("PWM delay disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevConfig() & 0xFF), (uint8_t)(fuseConfig->getDevConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::abistSNSEnable_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (abistSNSEnable->Checked)
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_ISNS_ABIST_RUN_MASK);
        logMessageToWindow("Current sense ABIST enabled.");
    }
    else
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_ISNS_ABIST_RUN_MASK);
        logMessageToWindow("Current sense ABIST disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::abistBGEnable_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (abistBGEnable->Checked)
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_BG_ABIST_RUN_MASK);
        logMessageToWindow("Bandgap ABIST enabled.");
    }
    else
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_BG_ABIST_RUN_MASK);
        logMessageToWindow("Bandgap ABIST disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::abistVDDEnable_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    /*if (abistVDDEnable->Checked)
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_VDDUVP_ABIST_RUN_MASK);
        logMessageToWindow("VDD ABIST enabled.");
    }
    else
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_VDDUVP_ABIST_RUN_MASK);
        logMessageToWindow("VDD ABIST disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };


    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd); */
}

System::Void autohssconfigurator::AutoConfigurationGui::abistADCEnable_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (abistADCEnable->Checked)
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_ADC_ABIST_RUN_MASK);
        logMessageToWindow("ADC ABIST enabled.");
    }
    else
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_ADC_ABIST_RUN_MASK);
        logMessageToWindow("ADC ABIST disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcEnableBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (adcEnableBox->Checked)
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_ADC_EN_MASK);
        logMessageToWindow("Device-level ADC enabled.");
    }
    else
    {
        fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_ADC_EN_MASK);
        logMessageToWindow("Device-level ADC disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcSouceInVBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (adcSouceInVBox->Checked)
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() & ~TPS2HC10S_ADC_CONFIG_ADC_VBB_DIS_MASK);
        logMessageToWindow("Device-level VBB ADC sampling enabled.");
    }
    else
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() | TPS2HC10S_ADC_CONFIG_ADC_VBB_DIS_MASK);
        logMessageToWindow("Device-level VBB ADC sampling disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getADCConfig() & 0xFF), (uint8_t)(fuseConfig->getADCConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcSourceOutVBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (adcSourceOutVBox->Checked)
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() & ~TPS2HC10S_ADC_CONFIG_ADC_VSNS_DIS_MASK);
        logMessageToWindow("Device-level VOUT ADC sampling enabled.");
    }
    else
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() | TPS2HC10S_ADC_CONFIG_ADC_VSNS_DIS_MASK);
        logMessageToWindow("Device-level VOUT ADC sampling disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getADCConfig() & 0xFF), (uint8_t)(fuseConfig->getADCConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcSourceJunctionBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (adcSourceJunctionBox->Checked)
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() & ~TPS2HC10S_ADC_CONFIG_ADC_TSNS_DIS_MASK);
        logMessageToWindow("Device-level temperature ADC sampling enabled.");
    }
    else
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig()  | TPS2HC10S_ADC_CONFIG_ADC_TSNS_DIS_MASK);
        logMessageToWindow("Device-level temperature ADC sampling disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getADCConfig() & 0xFF), (uint8_t)(fuseConfig->getADCConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcSourceCurrentBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (adcSourceCurrentBox->Checked)
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() & ~TPS2HC10S_ADC_CONFIG_ADC_ISNS_DIS_MASK);
        logMessageToWindow("Device-level current ADC sampling enabled.");
    }
    else
    {
        fuseConfig->setADCConfig(fuseConfig->getADCConfig() | TPS2HC10S_ADC_CONFIG_ADC_ISNS_DIS_MASK);
        logMessageToWindow("Device-level current ADC sampling disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getADCConfig() & 0xFF), (uint8_t)(fuseConfig->getADCConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskCLBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskCLBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_ILIMIT_MASK);
        logMessageToWindow("Current limit fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_ILIMIT_MASK);
        logMessageToWindow("Current limit fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskShortVBBBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskShortVBBBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_SHRT_VBB_MASK);
        logMessageToWindow("Short-to-VBB fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_SHRT_VBB_MASK);
        logMessageToWindow("Short-to-VBB fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskOpenLoadBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskOpenLoadBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_OL_OFF_MASK);
        logMessageToWindow("Open load OFF fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_OL_OFF_MASK);
        logMessageToWindow("Open load OFF fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskOpenLoadOnBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskOpenLoadOnBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_OL_ON_MASK);
        logMessageToWindow("Open load ON fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_OL_ON_MASK);
        logMessageToWindow("Open load ON fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskWDTBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskWDTBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_WD_ERR_MASK);
        logMessageToWindow("WDT fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_WD_ERR_MASK);
        logMessageToWindow("WDT fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskVBBUVLO_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskVBBUVLO->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_VBB_UVLO_MASK);
        logMessageToWindow("VBB UVLO fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_VBB_UVLO_MASK);
        logMessageToWindow("VBB UVLO fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultMaskSPIBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskSPIBox->Checked)
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() | TPS2HC10S_FAULT_MASK_MASK_SPI_ERR_MASK);
        logMessageToWindow("SPI fault mask enabled.");
    }
    else
    {
        fuseConfig->setDeviceFaultMask(fuseConfig->getDeviceFaultMask() & ~TPS2HC10S_FAULT_MASK_MASK_SPI_ERR_MASK);
        logMessageToWindow("SPI fault mask disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDeviceFaultMask() & 0xFF), (uint8_t)(fuseConfig->getDeviceFaultMask() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::latchFaultBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultMaskCLBox->Checked)
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() & ~TPS2HC10S_DEV_CONFIG_FLT_LTCH_DIS_MASK);
        logMessageToWindow("Fault latch enabled.");
    }
    else
    {
        fuseConfig->setDevConfig(fuseConfig->getDevConfig() | TPS2HC10S_DEV_CONFIG_FLT_LTCH_DIS_MASK);
        logMessageToWindow("Fault latch disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevConfig() & 0xFF), (uint8_t)(fuseConfig->getDevConfig() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::lpmCh1ThreshBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setLowPowerMode((lpmCh1ThreshBox->SelectedIndex << TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_OFS) | (fuseConfig->getLowPowerMode() & ~TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_MASK));
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getLowPowerMode() & 0xFF), (uint8_t)(fuseConfig->getLowPowerMode() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("LPM threshold for channel 1 changed to {0}", lpmCh1ThreshBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::lpmCh2ThreshBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setLowPowerMode((lpmCh2ThreshBox->SelectedIndex << TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_OFS) | (fuseConfig->getLowPowerMode() & ~TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_MASK));
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getLowPowerMode() & 0xFF), (uint8_t)(fuseConfig->getLowPowerMode() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("LPM threshold for channel 2 changed to {0}", lpmCh2ThreshBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::lpmEnableBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (lpmEnableBox->Checked)
    {
        fuseConfig->setLowPowerMode(fuseConfig->getLowPowerMode() | TPS2HC10S_LPM_LPM_MASK);
        logMessageToWindow("Low-power mode enabled.");
    }
    else
    {
        fuseConfig->setLowPowerMode(fuseConfig->getLowPowerMode() & ~TPS2HC10S_LPM_LPM_MASK);
        logMessageToWindow("Low-power mode disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getLowPowerMode() & 0xFF), (uint8_t)(fuseConfig->getLowPowerMode() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::capChargeEnableBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChannel = chanSettingBox->SelectedIndex;

    if (capChargeEnableBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setILIMConfig((((capChargeModeBox->SelectedIndex) + 1) << TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_OFS) |
            (fuseConfig->getILIMConfig(curChannel) & ~TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK), curChannel);
        logMessageToWindow("Capacitive charging mode enabled on channel {0}", curChannel);
    }
    else
    {
        fuseConfig->setILIMConfig(fuseConfig->getILIMConfig(curChannel) & ~TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK, curChannel);
        logMessageToWindow("Capacitive charging mode disabled on channel {0}", curChannel);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChannel) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChannel) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8*curChannel), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    dontRunEvents = true;
    updateCurrentLimitSettings(curChannel);
    dontRunEvents = false;
}

System::Void autohssconfigurator::AutoConfigurationGui::capChargeModeBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChannel = chanSettingBox->SelectedIndex;

    fuseConfig->setILIMConfig((((capChargeModeBox->SelectedIndex) + 1) << TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_OFS) |
        (fuseConfig->getILIMConfig(curChannel) & ~TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK), curChannel);

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChannel) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChannel) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChannel), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Capacitive charge mode changed on channel {0} to {1}", curChannel, capChargeModeBox->SelectedItem->ToString());
    
    dontRunEvents = true;
    updateCurrentLimitSettings(curChannel);
    dontRunEvents = false;
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingENBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);  

    if (curChan == 0)
    {
        if (chanSettingENBox->CheckState == CheckState::Checked)
        {
            fuseConfig->setSwitchState(fuseConfig->getSwitchState() | TPS2HC10S_SW_STATE_CH1_ON_MASK);
            logMessageToWindow("Channel 1 switch state enabled.");
        }
        else
        {
            fuseConfig->setSwitchState(fuseConfig->getSwitchState() & ~TPS2HC10S_SW_STATE_CH1_ON_MASK);
            logMessageToWindow("Channel 1 switch state disabled.");
        }
    }
    else
    {
        if (chanSettingENBox->CheckState == CheckState::Checked)
        {
            fuseConfig->setSwitchState(fuseConfig->getSwitchState() | TPS2HC10S_SW_STATE_CH2_ON_MASK);
            logMessageToWindow("Channel 2 switch state enabled.");
        }
        else
        {
            fuseConfig->setSwitchState(fuseConfig->getSwitchState() & ~TPS2HC10S_SW_STATE_CH2_ON_MASK);
            logMessageToWindow("Channel 2 switch state disabled.");
        }
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getSwitchState() & 0xFF), (uint8_t)(fuseConfig->getSwitchState() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_SW_STATE_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingPwmEnBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (chanSettingPwmEnBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(curChan) | TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, curChan);
        logMessageToWindow("PWM on channel 1 enabled.");
    }
    else
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(curChan) & ~TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, curChan);
        logMessageToWindow("PWM on channel 1 disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG + (8*curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingPWMFreqBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setPWMConfig((chanSettingPWMFreqBox->SelectedIndex << TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_OFS) |
                (fuseConfig->getPWMConfig(curChan) & ~TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM frequency on channel {0} set to {1}", curChan + 1, chanSettingPWMFreqBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingPWMDutyBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setPWMConfig((chanSettingPWMDutyBox->SelectedIndex << TPS2HC10S_PWM_CH1_PWM_DTY_CH1_OFS) |
                (fuseConfig->getPWMConfig(curChan) & ~TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM duty cycle on channel {0} set to {1}", curChan + 1, chanSettingPWMDutyBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        updateGuiFromConfig(true);
        return;
    }

    uint8_t curChannel = chanSettingBox->SelectedIndex;
    tHSSReceiveTransaction^ trans;

    /* Getting the selected channel data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, curChannel, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = curChannel;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);
}

System::Void autohssconfigurator::AutoConfigurationGui::lhiBehaviorBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (dontRunEvents == true)
    {
        return;
    }

    if (curChan == 0)
    {
        fuseConfig->setDevSAFConfig((lhiBehaviorBox->SelectedIndex << TPS2HC10S_DEVICE_SAF_CH1_LH_IN_OFS) |
            (fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_CH1_LH_IN_MASK));
        uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };
        /* Sending the updated data */
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
        sendRawWriteGUIResponse(cmd);
        free(cmd);
    }
    else
    {
        fuseConfig->setDevSAFConfig((lhiBehaviorBox->SelectedIndex << TPS2HC10S_DEVICE_SAF_CH2_LH_IN_OFS) |
            (fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_CH2_LH_IN_MASK));
        uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF), (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };
        /* Sending the updated data */
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
        sendRawWriteGUIResponse(cmd);
        free(cmd);
    }

    logMessageToWindow("LHI behavior on channel {0} set to {1}", curChan + 1, lhiBehaviorBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingLPMCurrentBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (dontRunEvents == true)
    {
        return;
    }

    if (curChan == 0)
    {
        fuseConfig->setLowPowerMode((chanSettingLPMCurrentBox->SelectedIndex << TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_OFS) |
            (fuseConfig->getLowPowerMode() & ~TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_MASK));
        uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getLowPowerMode() & 0xFF), (uint8_t)(fuseConfig->getLowPowerMode() >> 8) };
        /* Sending the updated data */
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
        sendRawWriteGUIResponse(cmd);
        free(cmd);
    }
    else
    {
        fuseConfig->setLowPowerMode((chanSettingLPMCurrentBox->SelectedIndex << TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_OFS) |
            (fuseConfig->getLowPowerMode() & ~TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_MASK));
        uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getLowPowerMode() & 0xFF), (uint8_t)(fuseConfig->getLowPowerMode() >> 8) };
        /* Sending the updated data */
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
        sendRawWriteGUIResponse(cmd);
        free(cmd);
    }

    logMessageToWindow("LPM exit current on channel {0} set to {1}", curChan + 1, chanSettingLPMCurrentBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanVBBADCBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (chanVBBADCBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_VSNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("VOUT ADC sampling on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_VSNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("VOUT ADC sampling on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::chanJunctionADCBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (chanJunctionADCBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_TSNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("Temperature ADC sampling on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_TSNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("Temperature ADC sampling on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::chanCurrentADCBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (chanCurrentADCBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_ISNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("Load current ADC sampling on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_ISNS_DIS_CH1_MASK, curChan);
        logMessageToWindow("Load current ADC sampling on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::adcScaleBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setChanDiagConfig((adcScaleBox->SelectedIndex << TPS2HC10S_DIAG_CONFIG_CH1_ISNS_SCALE_CH1_OFS) |
        (fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_ISNS_SCALE_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("ADC scaling on channel {0} changed to {1}.", curChan + 1, adcScaleBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::wbOnThreshBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setChanDiagConfig((wbOnThreshBox->SelectedIndex << TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_THLD_CH1_OFS) |
        (fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_THLD_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Open-load threshold on channel {0} changed to {1}.", curChan + 1, wbOnThreshBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::olScalingBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (olScalingBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_EN_CH1_MASK, curChan);
        logMessageToWindow("Open-load scaling on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_EN_CH1_MASK, curChan);
        logMessageToWindow("Open-load scaling on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::olOffPullBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (olOffPullBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_OL_OFF_EN_CH1_MASK, curChan);
        logMessageToWindow("Open-load OFF pull-up on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_OL_OFF_EN_CH1_MASK, curChan);
        logMessageToWindow("Open-load OFF pull-up on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::wbOffCurrrentBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setChanDiagConfig((wbOffCurrrentBox->SelectedIndex << TPS2HC10S_DIAG_CONFIG_CH1_OL_PU_STR_CH1_OFS) |
        (fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_OL_PU_STR_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Open-load OFF pull-up on channel {0} changed to {1}.", curChan + 1, wbOffCurrrentBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::wbBlankBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setChanDiagConfig((wbBlankBox->SelectedIndex << TPS2HC10S_DIAG_CONFIG_CH1_OL_SVBB_BLANK_CH1_OFS) |
        (fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_OL_SVBB_BLANK_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Open-load blanking on channel {0} changed to {1}.", curChan + 1, wbBlankBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::vbbPulldownBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (vbbPulldownBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_SVBB_EN_CH1_MASK, curChan);
        logMessageToWindow("VBB short pull-down on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_SVBB_EN_CH1_MASK, curChan);
        logMessageToWindow("VBB short pull-down on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::shutdownLatchBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (shutdownLatchBox->Checked == true)
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) | TPS2HC10S_DIAG_CONFIG_CH1_LATCH_CH1_MASK, curChan);
        logMessageToWindow("Shutdown retry on channel {0} set to latch.", curChan + 1);
    }
    else
    {
        fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_LATCH_CH1_MASK, curChan);
        logMessageToWindow("Shutdown retry on channel {0} set to auto-retry.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::shutdownRetryBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    return System::Void();
}

System::Void autohssconfigurator::AutoConfigurationGui::slewRateBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setChanDiagConfig((slewRateBox->SelectedIndex) |
        (fuseConfig->getChanDiagConfig(curChan) & ~TPS2HC10S_DIAG_CONFIG_CH1_SLRT_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getChanDiagConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getChanDiagConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Slew rate on channel {0} changed to {1}.", curChan + 1, slewRateBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingI2TEnBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    if (chanSettingI2TEnBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setILIMConfig(fuseConfig->getILIMConfig(curChan) | TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK, curChan);
        logMessageToWindow("I2T on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setILIMConfig(fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK, curChan);
        logMessageToWindow("I2T on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingNomBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setI2TConfig((chanSettingNomBox->SelectedIndex) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_NOM_CUR_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Nominal current on channel {0} changed to {1}.", curChan + 1, chanSettingNomBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingI2TBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setI2TConfig((chanSettingI2TBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("I2T trip value on channel {0} changed to {1}.", curChan + 1, chanSettingI2TBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingSWCLBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setI2TConfig((chanSettingSWCLBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Software current limit on channel {0} changed to {1}.", curChan + 1, chanSettingSWCLBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingSWDelayBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setI2TConfig((chanSettingSWDelayBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Software delay on channel {0} changed to {1}.", curChan + 1, chanSettingSWDelayBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingCLTimer_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setI2TConfig((chanSettingCLTimer->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Cooldown timer on channel {0} changed to {1}.", curChan + 1, chanSettingCLTimer->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMEn1_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (cmdPWMEn1->CheckState == CheckState::Checked)
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(0) | TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, 0);
        logMessageToWindow("PWN on channel 1 enabled.");
    }
    else
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(0) & ~TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, 0);
        logMessageToWindow("PWN on channel 1 disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(0) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(0) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMEn2_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (cmdPWMEn2->CheckState == CheckState::Checked)
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(1) | TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, 1);
        logMessageToWindow("PWN on channel 2 enabled.");
    }
    else
    {
        fuseConfig->setPWMConfig(fuseConfig->getPWMConfig(1) & ~TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK, 1);
        logMessageToWindow("PWN on channel 1 disabled.");
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(1) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(1) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMFreqBox1_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setPWMConfig((cmdPWMFreqBox1->SelectedIndex << TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_OFS) |
        (fuseConfig->getPWMConfig(0) & ~TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK), 0);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(0) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(0) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM frequency on channel 1 changed to {0}.", cmdPWMFreqBox1->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMFreqBox2_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setPWMConfig((cmdPWMFreqBox2->SelectedIndex << TPS2HC10S_PWM_CH2_PWM_FREQ_CH2_OFS) |
        (fuseConfig->getPWMConfig(1) & ~TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK), 1);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(1) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(1) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM frequency on channel 2 changed to {0}.", cmdPWMFreqBox2->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMDutyBox1_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setPWMConfig((cmdPWMDutyBox1->SelectedIndex << TPS2HC10S_PWM_CH1_PWM_DTY_CH1_OFS) |
        (fuseConfig->getPWMConfig(0) & ~TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK), 0);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(0) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(0) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM duty on channel 1 changed to {0}.", cmdPWMDutyBox1->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::cmdPWMDutyBox2_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    fuseConfig->setPWMConfig((cmdPWMDutyBox2->SelectedIndex << TPS2HC10S_PWM_CH2_PWM_DTY_CH2_OFS) |
        (fuseConfig->getPWMConfig(1) & ~TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK), 1);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getPWMConfig(1) & 0xFF), (uint8_t)(fuseConfig->getPWMConfig(1) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("PWM duty on channel 2 changed to {0}.", cmdPWMDutyBox2->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::mainTabControl_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t selTab = mainTabControl->SelectedIndex;
    tHSSReceiveTransaction^ trans;
    HSSCommand* cmd;

    if (manualRefreshEnabled == true)
    {
        return;
    }

    /* Updating the device config and both channels */
    if (selTab == 0)
    {
        /* Telling the HSS to update the configuration data for the selected address */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        /* Getting the initial device data */
        trans = gcnew tHSSReceiveTransaction;
        trans->addr = 0;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);

        /* Getting the initial channel 1 data */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 0, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        trans = gcnew tHSSReceiveTransaction;
        trans->addr = 0;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);

        /* Getting the initial channel 2 data */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 1, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        trans = gcnew tHSSReceiveTransaction;
        trans->addr = 1;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);
    }
    else if (selTab == 1)
    {
        uint8_t curChannel = i2tChanNumBox->SelectedIndex;

        /* Getting the initial channel 2 data */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, curChannel, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        trans = gcnew tHSSReceiveTransaction;
        trans->addr = curChannel;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);

    }
    else if (selTab == 2)
    {
        /* Telling the HSS to update the configuration data */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        /* Getting the initial device data */
        trans = gcnew tHSSReceiveTransaction;
        trans->addr = 0;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);
    }
    else if (selTab == 3)
    {
        uint8_t curChannel = chanSettingBox->SelectedIndex;

        /* Telling the HSS to update the configuration data for the selected address */
        cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        /* Getting the initial device data */
        trans = gcnew tHSSReceiveTransaction;
        trans->addr = 0;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);

        /* Getting the initial channel data */

        cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, curChannel, NULL, 0);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        trans = gcnew tHSSReceiveTransaction;
        trans->addr = curChannel;
        trans->recData = (uint8_t*)malloc(128);
        trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
        trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
        this->commManager->getPacketData(trans);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::dvdtPeriodBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint16_t curIndex = dvdtPeriodBox->SelectedIndex;
    uint16_t curPeriod = 0;
    uint8_t curChan = chanSettingBox->SelectedIndex;

    switch (curIndex)
    {
    case 0:
        curPeriod = DVDT_P33;
        break;
    case 1:
        curPeriod = DVDT_P50;
        break;
    case 2:
        curPeriod = DVDT_P67;
        break;
    case 3:
        curPeriod = DVDT_P90;
        break;
    case 4:
        curPeriod = DVDT_1P0;
        break;
    case 5:
        curPeriod = DVDT_1P1;
        break;
    case 6:
        curPeriod = DVDT_1P33;
        break;
    case 7:
        curPeriod = DVDT_1P66;
        break;
    case 8:
        curPeriod = DVDT_2P22;
        break;
    default:
        curPeriod = DVDT_P33;
        break;
    }

    fuseConfig->setILIMConfig((curPeriod << TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_OFS) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), curChan);


    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("dv/dt on channel {0} changed to {1}.", curChan + 1, dvdtPeriodBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chargeLevelBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((chargeLevelBox->SelectedIndex << 4) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Charge level on channel {0} changed to {1}.", curChan + 1, chargeLevelBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::capChargeDurationBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((capChargeDurationBox->SelectedIndex << 8) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Charge duration on channel {0} changed to {1}.", curChan + 1, capChargeDurationBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::pulsePeriodBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((pulsePeriodBox->SelectedIndex << 4) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Pulse period on channel {0} changed to {1}.", curChan + 1, pulsePeriodBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::capTurnOffBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((capTurnOffBox->SelectedIndex << 14) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Overcurrent mode on channel {0} changed to {1}.", curChan + 1, capTurnOffBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingCurrentLimitBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((chanSettingCurrentLimitBox->SelectedIndex) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Hardware current limit on channel {0} changed to {1}.", curChan + 1, chanSettingCurrentLimitBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::inRushValueBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((inRushValueBox->SelectedIndex << 4) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Inrush value on channel {0} changed to {1}.", curChan + 1, inRushValueBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::chanSettingInrushDurationBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((chanSettingInrushDurationBox->SelectedIndex << 8) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Inrush duration on channel {0} changed to {1}.", curChan + 1, chanSettingInrushDurationBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::threshReductionBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (chanSettingBox->SelectedIndex);

    fuseConfig->setILIMConfig((threshReductionBox->SelectedIndex << 14) |
        (fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Threshold reduction on channel {0} changed to {1}.", curChan + 1, threshReductionBox->SelectedItem->ToString());
}

System::Void autohssconfigurator::AutoConfigurationGui::faultBox1_Click(System::Object^ sender, System::EventArgs^ e)
{
    mainTabControl->SelectedIndex = 2;
}

System::Void autohssconfigurator::AutoConfigurationGui::faultBox2_Click(System::Object^ sender, System::EventArgs^ e)
{
    mainTabControl->SelectedIndex = 2;
}

System::Void autohssconfigurator::AutoConfigurationGui::faultBoxGlobal_Click(System::Object^ sender, System::EventArgs^ e)
{
    mainTabControl->SelectedIndex = 2;
}

System::Void autohssconfigurator::AutoConfigurationGui::thermalWarnCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = thWarnCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::thermalWarnCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = thWarnCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::olOnCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = openLoadOnFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::olOnCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = openLoadOnFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::olOffCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = openLoadOffFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::olOffCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = openLoadOffFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::shtVbbCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = shortVbbFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::shtVbbCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = shortVbbFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::ilimCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = ilimFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::ilimCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = ilimFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::tsdCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = thermShutCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::tsdCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = thermShutCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::lpmWakeCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = lpmWakeFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::lpmWakeCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = lpmWakeFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::voutErrCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = voutFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::voutErrCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = voutFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tWarnCh1_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = i2tFaultCh1;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tWarnCh2_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = i2tFaultCh2;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::faultUpdateStatus_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (faultUpdateStatus->SelectedIndex == 0)
    {
        manualUpdateEnabled = false;
        manualUpdateButton->Enabled = false;
        logMessageToWindow("Fault reading changed to automatic mode.");
    }
    else if (faultUpdateStatus->SelectedIndex == 1)
    {
        manualUpdateEnabled = true;
        manualUpdateButton->Enabled = true;
        logMessageToWindow("Fault reading changed to manual mode.");
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::manualUpdateButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    HSSCommand* faultStatus1;
    tHSSReceiveTransaction^ faultTrans1;

    faultStatus1 = new HSSCommand(HSSCommand::HSS_HCS_FAULT_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(faultStatus1->getBufferData(), faultStatus1->getCommandSize());
    
    faultTrans1 = gcnew tHSSReceiveTransaction;
    faultTrans1->addr = 0;
    faultTrans1->recData = (uint8_t*)malloc(128);
    faultTrans1->reportID = HSSCommunicationManager::HSSC_FAULT_ID;
    faultTrans1->receiveSize = HSSCommand::HSS_HCS_FAULT_CONFIG_RESP_SIZE;
    commManager->getPacketData(faultTrans1);

    logMessageToWindow("Manual update initiated.");

    free(faultStatus1);
}

System::Void autohssconfigurator::AutoConfigurationGui::clearAllStatsButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    porFault->resetFaultState();
    wdtFault->resetFaultState();
    vbbUVLOFault->resetFaultState();
    vbbWarningFault->resetFaultState();
    spiFault->resetFaultState();
    vddUvloFault->resetFaultState();
    thWarnCh1->resetFaultState();
    thWarnCh2->resetFaultState();
    openLoadOnFaultCh1->resetFaultState();
    openLoadOnFaultCh2->resetFaultState();
    openLoadOffFaultCh1->resetFaultState();
    openLoadOffFaultCh2->resetFaultState();
    shortVbbFaultCh1->resetFaultState();
    shortVbbFaultCh2->resetFaultState();
    ilimFaultCh1->resetFaultState();
    ilimFaultCh2->resetFaultState();
    thermShutCh1->resetFaultState();
    thermShutCh2->resetFaultState();
    lpmWakeFaultCh1->resetFaultState();
    lpmWakeFaultCh2->resetFaultState();
    voutFaultCh1->resetFaultState();
    voutFaultCh2->resetFaultState();
    i2tFaultCh1->resetFaultState();
    i2tFaultCh2->resetFaultState();

    logMessageToWindow("Fault statistics reset.");

    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::porFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = porFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::wdFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = wdtFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::spiFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = spiFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::vbbUVFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = vbbUVLOFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::uvWarnBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = vbbWarningFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::uvFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = vddUvloFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::senseChannelBox_ItemCheck(System::Object^ sender, System::Windows::Forms::ItemCheckEventArgs^ e)
{
    uint8_t selectedItems = 0;
    UInt32 count = senseChannelBox->Items->Count;
    bool updateChannelSensing = false;

    /* If we are updating via the "All" box, do not run this event */
    if (dontRunEvents == true)
    {
        return;
    }

    /* Getting the check status */
    if (senseChannelBox->GetItemChecked(0) == true)
    {
        selectedItems |= 0b0001;
    }

    if (senseChannelBox->GetItemChecked(1) == true)
    {
        selectedItems |= 0b0010;
    }

    /* Updating with the new state */
    if (e->NewValue == CheckState::Checked)
    {
        senseChart->Series[e->Index]->Enabled = true;
        selectedItems |= (1 << e->Index);
    }
    else
    {
        senseChart->Series[e->Index]->Enabled = false;
        selectedItems &= ~(1 << e->Index);
    }

    /* If everything is disabled, disable the MCU from polling the values */
    if (selectedItems == 0)
    {
        updateChannelSensing = true;
        isCurrentSenseEnabled = false;
    }
    else if ((isCurrentSenseEnabled == false) && selectedItems != 0)
    {
        updateChannelSensing = true;
        isCurrentSenseEnabled = true;
    }

    /* If everything is checked, we need to check the all button */
    dontRunEvents = true;
    if ((senseChannelBox->CheckedItems->Count == (senseChannelBox->Items->Count - 1))
        && e->NewValue == CheckState::Checked)
    {
        senseChannelAll->CheckState = CheckState::Checked;
    }
    else
    {
        senseChannelAll->CheckState = CheckState::Unchecked;
    }
    dontRunEvents = false;

    /* If all boxes are checked or no boxes are checked, we need to enable or disable the MCU
         from polling the sense values */
    if (updateChannelSensing == true)
    {
        /* Seeing what we need to set the offset as */
        uint8_t senseSource = HSSC_CURRENT_OFFSET;

        if (senseSourceSelBox->SelectedIndex == 2)
        {
            senseSource = HSSC_TEMP_OFFSET;
        }
        else if (senseSourceSelBox->SelectedIndex == 1)
        {
            senseSource = HSSC_VOUT_OFFSET;
        }

        /* Constructing the packet to enable the device to do sampling of the
             sense values */
        uint8_t senseEnablePayload[] = { 3, senseSource, (uint8_t)(selectedItems == 0 ? 0 : 1) };
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_SNS_DATA, 0,
            (uint8_t*)senseEnablePayload, 3);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        free(cmd);

        if (selectedItems != 0)
        {
            /* Device Stat Config */
            fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() | TPS2HC10S_DEVICE_SAF_ADC_EN_MASK);
            uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] =
            { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF),
                                        (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };
            HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* ADC Config */
            fuseConfig->setADCConfig(fuseConfig->getADCConfig() & ~(TPS2HC10S_ADC_CONFIG_ADC_VBB_DIS_MASK
                                        | TPS2HC10S_ADC_CONFIG_ADC_VSNS_DIS_MASK | TPS2HC10S_ADC_CONFIG_ADC_ISNS_DIS_MASK));
            tempBuffer[1] = (uint8_t)(fuseConfig->getADCConfig() & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getADCConfig() >> 8);
            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* Channel 1 */
            fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(0) &
                ~(TPS2HC10S_DIAG_CONFIG_CH1_ISNS_DIS_CH1_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH1_TSNS_DIS_CH1_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH1_VSNS_DIS_CH1_MASK), 0);
            tempBuffer[1] = (uint8_t)(fuseConfig->getChanDiagConfig(0) & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getChanDiagConfig(0) >> 8);

            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* Channel 2 */
            fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(1) &
                ~(TPS2HC10S_DIAG_CONFIG_CH2_ISNS_DIS_CH2_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH2_TSNS_DIS_CH2_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH2_VSNS_DIS_CH2_MASK), 1);
            tempBuffer[1] = (uint8_t)(fuseConfig->getChanDiagConfig(1) & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getChanDiagConfig(1) >> 8);

            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH2_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);
        }
        else
        {
            /* Device Stat Config */
            fuseConfig->setDevSAFConfig(fuseConfig->getDevSAFConfig() & ~TPS2HC10S_DEVICE_SAF_ADC_EN_MASK);
            uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] =
            { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getDevSAFConfig() & 0xFF),
                                        (uint8_t)(fuseConfig->getDevSAFConfig() >> 8) };
            HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* ADC Config */
            fuseConfig->setADCConfig(fuseConfig->getADCConfig() | (TPS2HC10S_ADC_CONFIG_ADC_ISNS_DIS_MASK
                | TPS2HC10S_ADC_CONFIG_ADC_VBB_DIS_MASK | TPS2HC10S_ADC_CONFIG_ADC_VSNS_DIS_MASK));
            tempBuffer[1] = (uint8_t)(fuseConfig->getADCConfig() & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getADCConfig() >> 8);
            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* Channel 1 */
            fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(0) |
                (TPS2HC10S_DIAG_CONFIG_CH1_ISNS_DIS_CH1_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH1_TSNS_DIS_CH1_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH1_VSNS_DIS_CH1_MASK), 0);
            tempBuffer[1] = (uint8_t)(fuseConfig->getChanDiagConfig(0) & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getChanDiagConfig(0) >> 8);

            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);

            /* Channel 2 */
            fuseConfig->setChanDiagConfig(fuseConfig->getChanDiagConfig(1) |
                (TPS2HC10S_DIAG_CONFIG_CH2_ISNS_DIS_CH2_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH2_TSNS_DIS_CH2_MASK |
                    TPS2HC10S_DIAG_CONFIG_CH2_VSNS_DIS_CH2_MASK), 1);
            tempBuffer[1] = (uint8_t)(fuseConfig->getChanDiagConfig(1) & 0xFF);
            tempBuffer[2] = (uint8_t)(fuseConfig->getChanDiagConfig(1) >> 8);

            cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH2_REG,
                tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);
        }

        if (selectedItems == 0)
        {
            channelOneLabel->Text = "";
            channelTwoLabel->Text = "";
            globalLabel->Text = "";
            inputVoltageValueLabel->Visible = false;
            inputVoltageDescStripLabel->Visible = false;
        }
        else
        {
            inputVoltageValueLabel->Visible = true;
            inputVoltageDescStripLabel->Visible = true;
        }

        /* Updating the log*/
        logMessageToWindow("Sense enabled on channels changed to: 0b{0} ", 
            gcnew String(std::bitset<8>(selectedItems).to_string().data()));
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::senseSourceSelBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    String^ graphTitle = gcnew String("");
    String^ yTitle = gcnew String("");
    bool oldEnableStatus = isCurrentSenseEnabled;

    if (dontRunEvents == true)
    {
        return;
    }

    graphTitle = senseSourceSelBox->Items[senseSourceSelBox->SelectedIndex]->ToString();

    /* Getting the right offset */
    if (senseSourceSelBox->SelectedIndex == 0)
    {
        curDataFormat = HSSC_CURRENT_OFFSET;
        
        yTitle = graphTitle + "(A)";
        curSenseUnit = "A";
        logMessageToWindow("Sensing source changed to load current");
    }
    else if (senseSourceSelBox->SelectedIndex == 1)
    {
        curDataFormat = HSSC_VOUT_OFFSET;
        yTitle = graphTitle + "(V)";
        curSenseUnit = "V";
        logMessageToWindow("Sensing source changed to output voltage");
    }

    /* Disabling sampling before changing the sense format*/
    if (isCurrentSenseEnabled)
    {
        isCurrentSenseEnabled = false;

        /* Constructing the packet to enable the device to do sampling of the
             sense values */
        uint8_t senseEnablePayload[] = { HSSC_PAYLOAD_PACKET_SIZE, curDataFormat , 0 };
        HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_SNS_DATA, 0, (uint8_t*)senseEnablePayload, 3);
        commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

        free(cmd);
    }

    /* Cleaning up the graph and changing the titles */
    for (int ii = 0; ii < senseChart->Series->Count; ii++)
    {
        while (senseChart->Series[ii]->Points->Count != 0)
        {
            senseChart->Series[ii]->Points->RemoveAt(0);
        }
    }

    /* Updating all of the labels */
    senseChart->Titles[0]->Text = graphTitle;
    senseChart->ChartAreas[0]->AxisY->Title = yTitle;

    /* Updating the poll sense */
    uint8_t senseEnablePayload[] = {3, curDataFormat ,  oldEnableStatus ? (uint8_t)1 : (uint8_t)0 };
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_SNS_DATA, 0, (uint8_t*)senseEnablePayload, 3);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    if (oldEnableStatus)
    {
        isCurrentSenseEnabled = true;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::maxSamplesBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    return System::Void();
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tWarnText_Click(System::Object^ sender, System::EventArgs^ e)
{
    mainTabControl->SelectedIndex = 1;
}

System::Void autohssconfigurator::AutoConfigurationGui::swLimitBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setI2TConfig((swLimitBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Software current limit on channel {0} changed to {1}.", curChan + 1, swLimitBox->SelectedItem->ToString());

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::swDelayBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setI2TConfig((swDelayBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Software delay on channel {0} changed to {1}.", curChan + 1, swDelayBox->SelectedItem->ToString());

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::cooldownDelayBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    fuseConfig->setI2TConfig((cooldownDelayBox->SelectedIndex << TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_OFS) |
        (fuseConfig->getI2TConfig(curChan) & ~TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK), curChan);
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getI2TConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getI2TConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    logMessageToWindow("Cooldown timer on channel {0} changed to {1}.", curChan + 1, cooldownDelayBox->SelectedItem->ToString());

    redrawI2TGraph();

    if (cooldownDelayBox->SelectedIndex == 0)
    {
        resetLatchButton->Enabled = true;
    }
    else
    {
        resetLatchButton->Enabled = false;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::gaugeSelectionBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    int ii;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Removing all of the current data points */
    while (i2tChart->Series[3]->Points->Count != 0)
    {
        i2tChart->Series[3]->Points->RemoveAt(0);
    }

    if (gaugeSelectionBox->SelectedIndex == 0)
    {
        i2tChart->Series[3]->Enabled = false;
        i2tChart->ChartAreas[0]->AxisY->Maximum = 100;
        redrawI2TGraph();
    }
    else if (gaugeSelectionBox->SelectedIndex == 1)
    {
        for (ii = 0; ii < 15; ii++)
        {
            i2tChart->Series[3]->Points->AddXY(twentyGaugeWireValues[0,ii], twentyGaugeWireValues[1,ii]);
       }
        i2tChart->Series[3]->Enabled = true;
        i2tChart->Series[3]->Points[7]->Label  = "20 Gauge";

        customWireBox->Enabled = false;
        customWireButton->Enabled = false;

    }
    else if (gaugeSelectionBox->SelectedIndex == 2)
    {
        for (ii = 0; ii < 15; ii++)
        {
            i2tChart->Series[3]->Points->AddXY(sixteenGaugeWireValues[0, ii], sixteenGaugeWireValues[1, ii]);
        }
        i2tChart->Series[3]->Enabled = true;
        i2tChart->Series[3]->Points[7]->Label = "12 Gauge";

        customWireBox->Enabled = false;
        customWireButton->Enabled = false;
    }
    else if (gaugeSelectionBox->SelectedIndex == 3)
    {
        customWireBox->Enabled = true;
        customWireButton->Enabled = true;

        if (customWireBox->Text->Length != 0)
        {
            loadWireProfileFromBox();
        }
    }
    
}

System::Void autohssconfigurator::AutoConfigurationGui::pulseOneCheckBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (pulseOneCheckBox->CheckState == CheckState::Unchecked)
    {
        i2tChart->Series[4]->Enabled = false;
        pulse1CurrentBox->Enabled = false;
        pulse1DurationBox->Enabled = false;
        return;
    }

    pulse1CurrentBox->Enabled = true;
    pulse1DurationBox->Enabled = true;

    pulse1DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::updatePulseBox(uint8_t dataSet, double current, double duration)
{
    if (current == 0 || duration == 0)
    {
        return;
    }

    while (i2tChart->Series[dataSet]->Points->Count != 0)
    {
        i2tChart->Series[dataSet]->Points->RemoveAt(0);
    }

    i2tChart->Series[dataSet]->Points->AddXY(current, duration);
    i2tChart->Series[dataSet]->Enabled = true;
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse1CurrentBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    pulse1DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse1DurationBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    double currentValue;
    double durationValue;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Getting the new values */
    if ((Double::TryParse(pulse1CurrentBox->Text, currentValue) == false) ||
        (Double::TryParse(pulse1DurationBox->Text, durationValue) == false))
    {
        return;
    }

    updatePulseBox(4, currentValue, durationValue);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse2CurrentBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    pulse2DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse2DurationBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    double currentValue;
    double durationValue;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Getting the new values */
    if ((Double::TryParse(pulse2CurrentBox->Text, currentValue) == false) ||
        (Double::TryParse(pulse2DurationBox->Text, durationValue) == false))
    {
        return;
    }

    updatePulseBox(5, currentValue, durationValue);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulseThreeCheckBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (pulseThreeCheckBox->CheckState == CheckState::Unchecked)
    {
        i2tChart->Series[6]->Enabled = false;
        pulse3CurrentBox->Enabled = false;
        pulse3DurationBox->Enabled = false;
        return;
    }

    pulse3CurrentBox->Enabled = true;
    pulse3DurationBox->Enabled = true;

    pulse3DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulseTwoCheckBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    /* Removing the old point */
    while (i2tChart->Series[5]->Points->Count != 0)
    {
        i2tChart->Series[5]->Points->RemoveAt(0);
    }

    if (pulseTwoCheckBox->CheckState == CheckState::Unchecked)
    {
        i2tChart->Series[5]->Enabled = false;
        pulse2CurrentBox->Enabled = false;
        pulse2DurationBox->Enabled = false;
        return;
    }

    pulse2CurrentBox->Enabled = true;
    pulse2DurationBox->Enabled = true;

    pulse2DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse3DurationBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    double currentValue;
    double durationValue;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Getting the new values */
    if ((Double::TryParse(pulse3CurrentBox->Text, currentValue) == false) ||
        (Double::TryParse(pulse3DurationBox->Text, durationValue) == false))
    {
        return;
    }
    updatePulseBox(6, currentValue, durationValue);
}

System::Void autohssconfigurator::AutoConfigurationGui::pulse3CurrentBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    pulse3DurationBox_TextChanged(sender, e);
}

System::Void autohssconfigurator::AutoConfigurationGui::senseChannelAll_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    bool selectOrNot = senseChannelAll->Checked;

    for (int ii = 0; ii < senseChannelBox->Items->Count; ii++)
    {
        senseChannelBox->SetItemChecked(ii, selectOrNot);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::forceLimpCheckBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t lhiStatus;
    HSSCommand* cmd;

    if (dontRunEvents == true)
    {
        return;
    }

    if (forceLimpCheckBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setLHIForced(true);
        logMessageToWindow("Limp home input (LHI) forced high.");
        lhiStatus = 1;
    }
    else
    {
        fuseConfig->setLHIForced(false);
        logMessageToWindow("Limp home input (LHI) forced low.");
        lhiStatus = 0;
    }

    uint8_t payload[2] = {HSSCommand::HSS_IOCTL_SETLHI, lhiStatus};

    /* Telling the HSS to update the configuration data for the selected address */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_IOCTL, payload, 2);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::i2tChanNumBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        updateGuiFromConfig(true);
        return;
    }

    uint8_t curChannel = i2tChanNumBox->SelectedIndex;
    tHSSReceiveTransaction^ trans;

    /* Getting the selected channel data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, curChannel, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = curChannel;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);
}

System::Void autohssconfigurator::AutoConfigurationGui::enableLogButton_CheckStateChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    if (enableLogButton->CheckState == CheckState::Checked)
    {
        logBrowseButton->Enabled = true;
        logBrowseButton_Click(sender, e);
    }
    else
    {
        logBrowseButton->Enabled = false;
        isSnsLoggingEnabled = false;

        if (fileLogger != nullptr)
        {
            fileLogger->Close();
        }
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::logBrowseButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    SaveFileDialog^ saveFileDialog1 = gcnew SaveFileDialog;
    saveFileDialog1->Filter = "csv files (*.csv)|*.csv";
    saveFileDialog1->FilterIndex = 2;
    saveFileDialog1->RestoreDirectory = true;

    if (saveFileDialog1->ShowDialog() == System::Windows::Forms::DialogResult::OK)
    {
        if (fileLogger != nullptr)
        {
            fileLogger->Close();
        }

        fileLogger = gcnew IO::StreamWriter(saveFileDialog1->FileName);
        fileLogger->WriteLine("timestamp,input voltage,ch1,ch2");
        isSnsLoggingEnabled = true;
        filePathTextBox->Text = saveFileDialog1->FileName;
    }
    else
    {
        enableLogButton->CheckState = CheckState::Unchecked;
    }
}


System::Void autohssconfigurator::AutoConfigurationGui::enableI2TBox_CheckedChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    uint8_t curChan = (i2tChanNumBox->SelectedIndex);

    if (enableI2TBox->CheckState == CheckState::Checked)
    {
        fuseConfig->setILIMConfig(fuseConfig->getILIMConfig(curChan) | TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK, curChan);
        logMessageToWindow("I2T on channel {0} enabled.", curChan + 1);
    }
    else
    {
        fuseConfig->setILIMConfig(fuseConfig->getILIMConfig(curChan) & ~TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK, curChan);
        logMessageToWindow("I2T on channel {0} disabled.", curChan + 1);
    }

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(curChan) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(curChan) >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (8 * curChan), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);
}

System::Void autohssconfigurator::AutoConfigurationGui::enterDFUModeToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t tempBuffer[1];
    HSSCommand* cmd;

    if (dontRunEvents == true)
    {
        return;
    }

    /* Sending the DFU command. The device will disconnect after this */
    tempBuffer[0] = 0;
    cmd = new HSSCommand(HSSCommand::HSS_HCS_ENTERDFU, 0, tempBuffer, 1);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    MessageBox::Show("Device Firmware Update (DFU) entry command sent. Use dfuprog.exe to update firmware.",
        "DFU Entry Command Sent",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::senseResistorBox_SelectedIndexChanged(System::Object^ sender, System::EventArgs^ e)
{

}

System::Void autohssconfigurator::AutoConfigurationGui::exitToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    Application::Exit();
}

System::Void autohssconfigurator::AutoConfigurationGui::loadFileIntoGUI(String^ fileName)
{
    System::IO::StreamReader^ sr;
    String^ curString;
    array<String^>^ splitString;
    uint16_t curReg, curVal;

    sr = gcnew IO::StreamReader(fileName);
    curString = sr->ReadLine();
    if (curString != magicFileWord)
    {
        MessageBox::Show("File is in wrong format.", "Error Opening File!",
            MessageBoxButtons::OK, MessageBoxIcon::Error);
        sr->Close();
        return;
    }

    /* Getting the first address */
    curString = sr->ReadLine();
    if (curString == nullptr)
    {
        sr->Close();
        return;
    }

    do
    {
        if (curString->Contains("/*"))
        {
            curString = sr->ReadLine();
            continue;
        }

        splitString = curString->Split(',');
        if (splitString->Length != 2)
        {
            curString = sr->ReadLine();
            continue;
        }
        curReg = int::Parse(splitString[0]);
        curVal = int::Parse(splitString[1]);

        switch (curReg)
        {
        case TPS2HC10S_LPM_REG:
            fuseConfig->setLowPowerMode(curVal);
            break;
        case TPS2HC10S_FAULT_MASK_REG:
            fuseConfig->setDeviceFaultMask(curVal);
            break;
        case TPS2HC10S_SW_STATE_REG:
            fuseConfig->setSwitchState(curVal);
            break;
        case TPS2HC10S_DEVICE_SAF_REG:
            fuseConfig->setDeviceFaultMask(curVal);
            break;
        case TPS2HC10S_DEV_CONFIG_REG:
            fuseConfig->setDevConfig(curVal);
            break;
        case TPS2HC10S_ADC_CONFIG_REG:
            fuseConfig->setADCConfig(curVal);
            break;
        case TPS2HC10S_PWM_CH1_REG:
            fuseConfig->setPWMConfig(curVal, 0);
            break;
        case TPS2HC10S_ILIM_CONFIG_CH1_REG:
            fuseConfig->setILIMConfig(curVal, 0);
            break;
        case TPS2HC10S_DIAG_CONFIG_CH1_REG:
            fuseConfig->setChanDiagConfig(curVal, 0);
            break;
        case TPS2HC10S_I2T_CONFIG_CH1_REG:
            fuseConfig->setI2TConfig(curVal, 0);
            break;
        case TPS2HC10S_PWM_CH2_REG:
            fuseConfig->setPWMConfig(curVal, 1);
            break;
        case TPS2HC10S_ILIM_CONFIG_CH2_REG:
            fuseConfig->setILIMConfig(curVal, 1);
            break;
        case TPS2HC10S_DIAG_CONFIG_CH2_REG:
            fuseConfig->setChanDiagConfig(curVal, 1);
            break;
        case TPS2HC10S_I2T_CONFIG_CH2_REG:
            fuseConfig->setI2TConfig(curVal, 1);
            break;
        default:
            break;
        }
        curString = sr->ReadLine();
    } while (curString != nullptr);

    sr->Close();

    updateGuiFromConfig(true);
}

System::Void autohssconfigurator::AutoConfigurationGui::loaToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    OpenFileDialog^ openFileDialog1 = gcnew OpenFileDialog;

    openFileDialog1->Filter = "txt files (*.txt)|*.txt";
    openFileDialog1->FilterIndex = 2;
    openFileDialog1->RestoreDirectory = true;

    if (openFileDialog1->ShowDialog() == System::Windows::Forms::DialogResult::OK)
    {
        loadFileIntoGUI(openFileDialog1->FileName);


        if (manualPersistEnabled == true)
        {
            dirtyPersist = true;
            logMessageToWindow("Loaded {0} file configuration into GUI", openFileDialog1->FileName);
        }
        else
        {
            rawWriteFromConfig();
            logMessageToWindow("Loaded {0} file configuration into device", openFileDialog1->FileName);
        }
 
        MessageBox::Show("Configuration loaded from " + openFileDialog1->FileName, "Configuration Loaded",
            MessageBoxButtons::OK, MessageBoxIcon::Information);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::rawWriteFromConfig(void)
{
    HSSCommand* cmd;
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE];
    tempBuffer[0] = HSSC_RAW_WRITE_SIZE;

    tempBuffer[1] = (uint8_t)((fuseConfig->getLowPowerMode()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getLowPowerMode()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_LPM_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getDeviceFaultMask()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getDeviceFaultMask()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_FAULT_MASK_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getSwitchState()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getSwitchState()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_SW_STATE_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getDevSAFConfig()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getDevSAFConfig()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEVICE_SAF_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getDevConfig()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getDevConfig()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DEV_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getADCConfig()) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getADCConfig()) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ADC_CONFIG_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getPWMConfig(0)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getPWMConfig(0)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getILIMConfig(0)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getILIMConfig(0)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getChanDiagConfig(0)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getChanDiagConfig(0)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getI2TConfig(0)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getI2TConfig(0)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getPWMConfig(1)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getPWMConfig(1)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_PWM_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getILIMConfig(1)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getILIMConfig(1)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getChanDiagConfig(1)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getChanDiagConfig(1)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_DIAG_CONFIG_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    tempBuffer[1] = (uint8_t)((fuseConfig->getI2TConfig(1)) & 0xFF);
    tempBuffer[2] = (uint8_t)((fuseConfig->getI2TConfig(1)) >> 8);
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH2_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

}

System::Void autohssconfigurator::AutoConfigurationGui::saveSettingsToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    SaveFileDialog^ saveFileDialog1 = gcnew SaveFileDialog;
    System::IO::StreamWriter^ sw;

    saveFileDialog1->Filter = "txt files (*.txt)|*.txt";
    saveFileDialog1->FilterIndex = 2;
    saveFileDialog1->RestoreDirectory = true;

    if (saveFileDialog1->ShowDialog() == System::Windows::Forms::DialogResult::OK)
    {
        sw = gcnew IO::StreamWriter(saveFileDialog1->FileName);

        sw->WriteLine(magicFileWord);
        sw->WriteLine("/* TPS2HC10S_LPM */");
        sw->WriteLine("{0},{1}", TPS2HC10S_LPM_REG, fuseConfig->getLowPowerMode());
        sw->WriteLine("/* TPS2HC10S_FAULT_MASK */");
        sw->WriteLine("{0},{1}", TPS2HC10S_FAULT_MASK_REG, fuseConfig->getDeviceFaultMask());
        sw->WriteLine("/* TPS2HC10S_SW_STATE */");
        sw->WriteLine("{0},{1}", TPS2HC10S_SW_STATE_REG, fuseConfig->getSwitchState());
        sw->WriteLine("/* TPS2HC10S_DEVICE_SAF */");
        sw->WriteLine("{0},{1}", TPS2HC10S_DEVICE_SAF_REG, fuseConfig->getDevSAFConfig());
        sw->WriteLine("/* TPS2HC10S_DEV_CONFIG */");
        sw->WriteLine("{0},{1}", TPS2HC10S_DEV_CONFIG_REG, fuseConfig->getDevConfig());
        sw->WriteLine("/* TPS2HC10S_ADC_CONFIG */");
        sw->WriteLine("{0},{1}", TPS2HC10S_ADC_CONFIG_REG, fuseConfig->getADCConfig());
        sw->WriteLine("/* TPS2HC10S_PWM_CH1 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_PWM_CH1_REG, fuseConfig->getPWMConfig(0));
        sw->WriteLine("/* TPS2HC10S_ILIM_CONFIG_CH1 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_ILIM_CONFIG_CH1_REG, fuseConfig->getILIMConfig(0));
        sw->WriteLine("/* TPS2HC10S_DIAG_CONFIG_CH1 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_DIAG_CONFIG_CH1_REG, fuseConfig->getChanDiagConfig(0));
        sw->WriteLine("/* TPS2HC10S_I2T_CONFIG_CH1 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_I2T_CONFIG_CH1_REG, fuseConfig->getI2TConfig(0));
        sw->WriteLine("/* TPS2HC10S_PWM_CH2 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_PWM_CH2_REG, fuseConfig->getPWMConfig(1));
        sw->WriteLine("/* TPS2HC10S_ILIM_CONFIG_CH2 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_ILIM_CONFIG_CH2_REG, fuseConfig->getILIMConfig(1));
        sw->WriteLine("/* TPS2HC10S_DIAG_CONFIG_CH2 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_DIAG_CONFIG_CH2_REG, fuseConfig->getChanDiagConfig(1));
        sw->WriteLine("/* TPS2HC10S_I2T_CONFIG_CH2 */");
        sw->WriteLine("{0},{1}", TPS2HC10S_I2T_CONFIG_CH2_REG, fuseConfig->getI2TConfig(1));

        sw->Close();

        logMessageToWindow("Saved configuration to {0}", saveFileDialog1->FileName);

        MessageBox::Show("Configuration saved to " + saveFileDialog1->FileName, "Configuration Saved",
            MessageBoxButtons::OK, MessageBoxIcon::Information);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::sleepModeButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    if ((fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH1_ON_MASK) || 
        (fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH2_ON_MASK))
    {
        MessageBox::Show("Sleep cannot be entered with channels enabled. Disable the channels before entering sleep mode.",
            "Channels Enabled", MessageBoxButtons::OK, MessageBoxIcon::Error);
        return;
    }

     System::Windows::Forms::DialogResult res1 =  
         MessageBox::Show("Note that when the device enters sleep mode, all SPI reads/writes are ignored until sleep mode is exited (by clicking this button again). Graphing sense values will" +
                                                        " also be disabled. Proceed? ",
        "Sleep Mode",MessageBoxButtons::YesNo, MessageBoxIcon::Warning);

     if (res1 == System::Windows::Forms::DialogResult::Yes)
     {
         /* Turning off automatic polling and pausing current sense */
         bool curManualMode = manualUpdateEnabled;
         bool curDontRun = dontRunEvents;

         /* Disabing Sensing */
         uint8_t payload[] = { 3, HSSC_CURRENT_OFFSET, 0 };
         HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_SNS_DATA, 0,
             (uint8_t*)payload, 3);
         commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
         free(cmd);

         dontRunEvents = true;
         for (int ii = 0; ii < senseChannelBox->Items->Count; ii++)
         {
             senseChannelBox->SetItemChecked(ii, false);
         }
         senseChannelAll->CheckState = CheckState::Unchecked;
         senseSourceSelBox->SelectedIndex = 0;
         dontRunEvents = curDontRun;
         isCurrentSenseEnabled = false;
         manualUpdateEnabled = true;

         /* Waiting a second to make sure that polling has stopped*/
         Sleep(500);

         payload[0] = HSSC_RAW_WRITE_SIZE;
         payload[1] = TPS2HC10S_SLEEP_SLEEP_MASK;
         payload[2] = 0;
         cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_SLEEP_REG, payload, HSSC_PAYLOAD_PACKET_SIZE);
         commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
         free(cmd);

         MessageBox::Show("Sleep mode entered. Click OK to exit sleep mode.",
             "Sleep Mode Entered", MessageBoxButtons::OK, MessageBoxIcon::Information);

         /* Writing a dummy write command to take the device out of sleep mode if it is in there */
         payload[1] = 0xFF;
         payload[2] = 0xFF;
         cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, 0xFF, payload, HSSC_PAYLOAD_PACKET_SIZE);
         commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
         free(cmd);

         /* Waitng one ms to make sure the device gets out of sleep mode */
         Sleep(1);

         manualUpdateEnabled = curManualMode;

         /* Restoring the configuration values */
         rawWriteFromConfig();

         MessageBox::Show("Sleep mode exited. Configuration values restored. Have a wonderful day!",
             "Configuration Restored", MessageBoxButtons::OK, MessageBoxIcon::Information);
     }
}

System::Void autohssconfigurator::AutoConfigurationGui::enabledToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    if (manualPersistEnableCheckBox->CheckState == CheckState::Checked)
    {
        manualPersistEnabled = false;
        dirtyPersist = false;
        rawWriteFromConfig();
        logMessageToWindow("Manual persist disabled. Settings synced.");
        manualPersistToolStripMenuItem->Enabled = false;
    }
    else
    {
        logMessageToWindow("Manual persist enabled.");
        manualPersistEnabled = true;
        manualPersistToolStripMenuItem->Enabled = true;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::manualPersistToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    dirtyPersist = false;
    rawWriteFromConfig();
    logMessageToWindow("Settings persisted to device.");
}

System::Void autohssconfigurator::AutoConfigurationGui::manualRefreshToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    tHSSReceiveTransaction^ trans;
    HSSCommand* cmd;

    /* Telling the HSS to update the configuration data for the selected address */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

    /* Getting the initial device data */
    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the initial channel 1 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the initial channel 2 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 1, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 1;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    logMessageToWindow("Manual refresh completed.");
}

System::Void autohssconfigurator::AutoConfigurationGui::manualRefreshToolCheck_CheckStateChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (manualRefreshToolCheck->CheckState == CheckState::Checked)
    {
        manualRefreshEnabled = false;
        mainTabControl_SelectedIndexChanged(sender, e);
        logMessageToWindow("Manual refresh disabled");
        faultUpdateStatus->SelectedIndex = 0;
        manualRefreshToolStripMenuItem->Enabled = false;
    }
    else
    {
        logMessageToWindow("Manual persist enabled.");
        manualRefreshEnabled = true;
        faultUpdateStatus->SelectedIndex = 1;
        manualRefreshToolStripMenuItem->Enabled = true;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::cStructToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    SaveFileDialog^ saveFileDialog1 = gcnew SaveFileDialog;
    System::IO::StreamWriter^ sw;
    msclr::interop::marshal_context context;

    saveFileDialog1->Filter = "Header files (*.h)|*.h";
    saveFileDialog1->FilterIndex = 2;
    saveFileDialog1->RestoreDirectory = true;

    if (saveFileDialog1->ShowDialog() == System::Windows::Forms::DialogResult::OK)
    {
        sw = gcnew IO::StreamWriter(saveFileDialog1->FileName);

        String^ guardName = "__" + System::IO::Path::GetFileName(saveFileDialog1->FileName)->Replace(".", "_")->ToUpper() + "__";

        sw->WriteLine("#ifndef " + guardName);
        sw->WriteLine("#define " + guardName + "\n");

        sw->WriteLine("#include <stdint.h>");
        sw->WriteLine("#include <tps_hcsdev.h>\n");
        

        sw->WriteLine("typedef struct TPS2HCS10Q1_CONFIG");
        sw->WriteLine("{");
        sw->WriteLine("    TPS2HC10S_CRC_CONFIG_OBJ        crcConfig;");
        sw->WriteLine("    TPS2HC10S_LPM_OBJ               lpmConfig;");
        sw->WriteLine("    TPS2HC10S_FAULT_MASK_OBJ        faultMaskConfig;");
        sw->WriteLine("    TPS2HC10S_SW_STATE_OBJ          swState;");
        sw->WriteLine("    TPS2HC10S_DEVICE_SAF_OBJ        devSAFConfig;");
        sw->WriteLine("    TPS2HC10S_DEV_CONFIG_OBJ        devConfig;");
        sw->WriteLine("    TPS2HC10S_ADC_CONFIG_OBJ        adcConfig;");
        sw->WriteLine("    TPS2HC10S_PWM_CH1_OBJ           pwmCh1Config;");
        sw->WriteLine("    TPS2HC10S_ILIM_CONFIG_CH1_OBJ   ilimCh1Config;");
        sw->WriteLine("    TPS2HC10S_DIAG_CONFIG_CH1_OBJ   diagConfigCh1;");
        sw->WriteLine("    TPS2HC10S_I2T_CONFIG_CH1_OBJ    i2tConfigCh1;");
        sw->WriteLine("    TPS2HC10S_PWM_CH2_OBJ           pwmCh2Config;");
        sw->WriteLine("    TPS2HC10S_ILIM_CONFIG_CH2_OBJ   ilimCh2Config;");
        sw->WriteLine("    TPS2HC10S_DIAG_CONFIG_CH2_OBJ   diagConfigCh2;");
        sw->WriteLine("    TPS2HC10S_I2T_CONFIG_CH2_OBJ    i2tConfigCh2;");
        sw->WriteLine(" } TPS2HCS10Q1_CONFIG;\n");

        sw->WriteLine("extern TPS2HCS10Q1_CONFIG exportConfig;\n");

        sw->WriteLine("#endif");

        sw->Close();

        /* Making the definition */
        sw = gcnew IO::StreamWriter(saveFileDialog1->FileName->Replace(".h", ".c"));

        sw->WriteLine("#include \"" + System::IO::Path::GetFileName(saveFileDialog1->FileName) + "\"\n");

        sw->WriteLine("TPS2HCS10Q1_CONFIG exportConfig =");
        sw->WriteLine("{");
        sw->WriteLine("    .crcConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_CRC_CONFIG_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getCRCConfig().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .lpmConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_LPM_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getLowPowerMode().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .faultMaskConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_FAULT_MASK_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getDeviceFaultMask().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .swState =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_SW_STATE_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getSwitchState().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .devSAFConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_DEVICE_SAF_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getDevSAFConfig().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .devConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_DEV_CONFIG_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getDevConfig().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .adcConfig =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_ADC_CONFIG_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getADCConfig().ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .pwmCh1Config =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_PWM_CH1_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getPWMConfig(0).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .ilimCh1Config =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_ILIM_CONFIG_CH1_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getILIMConfig(0).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .diagConfigCh1 =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_DIAG_CONFIG_CH1_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getChanDiagConfig(0).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .i2tConfigCh1 =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_I2T_CONFIG_CH1_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getI2TConfig(0).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .pwmCh2Config =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_PWM_CH2_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getPWMConfig(1).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .ilimCh2Config =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_ILIM_CONFIG_CH2_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getILIMConfig(1).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .diagConfigCh2 =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_DIAG_CONFIG_CH2_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getChanDiagConfig(1).ToString("X4"));
        sw->WriteLine("    },");
        sw->WriteLine("    .i2tConfigCh2 =");
        sw->WriteLine("    {");
        sw->WriteLine("        .reg = TPS2HC10S_I2T_CONFIG_CH2_REG,");
        sw->WriteLine("        .value.word = 0x" + fuseConfig->getI2TConfig(1).ToString("X4"));
        sw->WriteLine("    }");
        sw->WriteLine("};\n");

        sw->Close();

        LPCTSTR convertedString = context.marshal_as<const TCHAR*>(System::IO::Path::GetDirectoryName(saveFileDialog1->FileName) + "/tps2hcs10.h");
        CopyFile(L"headers/tps2hcs10.h", convertedString, FALSE);
        convertedString = context.marshal_as<const TCHAR*>(System::IO::Path::GetDirectoryName(saveFileDialog1->FileName) + "/tps_hcsdev.h");
        CopyFile(L"headers/tps_hcsdev.h", convertedString, FALSE);

        logMessageToWindow("Saved configuration to {0}", saveFileDialog1->FileName);

        MessageBox::Show("Configuration saved to " + saveFileDialog1->FileName, "Configuration Saved",
            MessageBoxButtons::OK, MessageBoxIcon::Information);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::loadWireProfileFromBox()
{
    float curCurrent, curTime;
    System::IO::StreamReader^ sr;
    String^ curString;
    array<String^>^ splitString;
    IntPtr ptr = System::Runtime::InteropServices::Marshal::StringToHGlobalUni(customWireBox->Text);
    LPCWSTR filePtr = (LPCWSTR)ptr.ToPointer();

    GetFileAttributes(filePtr); // from winbase.h
    if (INVALID_FILE_ATTRIBUTES == GetFileAttributes(filePtr))
    {
        MessageBox::Show("File not found!",
            "File not found", MessageBoxButtons::OK, MessageBoxIcon::Error);
        System::Runtime::InteropServices::Marshal::FreeHGlobal(ptr);
        customWireBox->Text = "";
        return;
    }

    System::Runtime::InteropServices::Marshal::FreeHGlobal(ptr);

    sr = gcnew IO::StreamReader(customWireBox->Text);

    /* Discarding the first line as it's the header*/
    curString = sr->ReadLine();

    /* Getting the first value */
    curString = sr->ReadLine();
    if (curString == nullptr)
    {
        sr->Close();
        return;
    }
    do
    {
        splitString = curString->Split(',');
        if (splitString->Length != 2)
        {
            curString = sr->ReadLine();
            continue;
        }
        curCurrent = float::Parse(splitString[0]);
        curTime = float::Parse(splitString[1]);

        i2tChart->Series[3]->Points->AddXY(curCurrent, curTime);

        curString = sr->ReadLine();
    } while (curString != nullptr);

    sr->Close();

    i2tChart->Series[3]->Enabled = true;
}

System::Void autohssconfigurator::AutoConfigurationGui::customWireButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    OpenFileDialog^ openFileDialog1 = gcnew OpenFileDialog;

    openFileDialog1->Filter = "comma-separated value files (*.csv)|*.csv";
    openFileDialog1->FilterIndex = 2;
    openFileDialog1->RestoreDirectory = true;

    if (openFileDialog1->ShowDialog() == System::Windows::Forms::DialogResult::OK)
    {
        customWireBox->Text = openFileDialog1->FileName;
        loadWireProfileFromBox();

        if (customWireBox->Text->Length != 0)
        {
            MessageBox::Show("Wire gauge loaded from " + openFileDialog1->FileName, "Wire Model Loaded",
                MessageBoxButtons::OK, MessageBoxIcon::Information);
        }
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::gaugeSelectionBox_KeyDown(System::Object^ sender, System::Windows::Forms::KeyEventArgs^ e)
{
    if (e->KeyCode == Keys::Enter)
    {
        loadWireProfileFromBox();
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::senseResistorValBox_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "Not a register";
    helpLabel->Text = "Sets the sense resistor value for the soldered down SNS resistor on R6. This directly correlates to the value of I2T " +
        "parameters being used. If a different resistor has been soldered down, update the value here.";
}

System::Void autohssconfigurator::AutoConfigurationGui::senseResistorValBox_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::resToleranceBox_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "Not a register";
    helpLabel->Text = "The tolerance of the resistor connected to the SNS pin.";
}

System::Void autohssconfigurator::AutoConfigurationGui::resToleranceBox_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::resetLatchButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t curChanNum = i2tChanNumBox->SelectedIndex;

    /* Disabling the current channel */
    fuseConfig->setSwitchState(fuseConfig->getSwitchState() & ~(curChanNum + 1));

    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] =
    { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getSwitchState() & 0xFF), (uint8_t)(fuseConfig->getSwitchState() >> 8) };

    /* Sending the updated data */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_SW_STATE_REG, tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    /* Setting the cooldown timer */
    fuseConfig->setI2TConfig(fuseConfig->getI2TConfig(curChanNum) | tcldn_ch1_en_2_0x1, curChanNum);

    /* Sending the updated data */
    tempBuffer[1] = (uint8_t)(fuseConfig->getI2TConfig(curChanNum) & 0xFF);
    tempBuffer[2] = (uint8_t)(fuseConfig->getI2TConfig(curChanNum) >> 8);

    /* Sending the updated data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChanNum), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    /* Sleeping for 1s */
    Sleep(1000);

    /* Setting the latch mode */
    fuseConfig->setI2TConfig(fuseConfig->getI2TConfig(curChanNum) & ~TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK, curChanNum);

    /* Sending the updated data */
    tempBuffer[1] = (uint8_t)(fuseConfig->getI2TConfig(curChanNum) & 0xFF);
    tempBuffer[2] = (uint8_t)(fuseConfig->getI2TConfig(curChanNum) >> 8);

    /* Sending the updated data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_I2T_CONFIG_CH1_REG + (8 * curChanNum), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    sendRawWriteGUIResponse(cmd);
    free(cmd);

    MessageBox::Show("Latch state reset. The channel is now ready to be re-enabled", "Latch Reset",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::resetLatchButton_MouseEnter(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "Not a register";
    helpLabel->Text = "Clicking this button will perform the series of steps required to reset the latch state of the high-side switch.";
}

System::Void autohssconfigurator::AutoConfigurationGui::resetLatchButton_MouseLeave(System::Object^ sender, System::EventArgs^ e)
{
    registerLabel->Text = "";
    helpLabel->Text = "";
}

System::Void autohssconfigurator::AutoConfigurationGui::resToleranceBox_KeyDown(System::Object^ sender, System::Windows::Forms::KeyEventArgs^ e)
{
    if (e->KeyCode == Keys::Enter)
    {
        resToleranceBox_Leave(sender, e);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::resToleranceBox_Leave(System::Object^ sender, System::EventArgs^ e)
{
    if (!float::TryParse(resToleranceBox->Text, curResTolerance))
    {
        resToleranceBox->Text = "0.0";
        curResTolerance = 0.0f;
    }

    redrawI2TGraph();
}

System::Void autohssconfigurator::AutoConfigurationGui::limpHomeFaultBox_Click(System::Object^ sender, System::EventArgs^ e)
{
    curFault = limpFault;
    handleFaultClick(curFault);
}

System::Void autohssconfigurator::AutoConfigurationGui::aboutToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    String^ target = "https://www.ti.com/lit/pdf/slvucr9";
    System::Diagnostics::Process::Start(target);
}

System::Void autohssconfigurator::AutoConfigurationGui::senseResistorValBox_TextChanged(System::Object^ sender, System::EventArgs^ e)
{
    float resTolerance;

    if (!float::TryParse(senseResistorValBox->Text, resTolerance))
    {
        senseResistorValBox->Text = "698";
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::clearLimpButton_Click(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_LIMPHOME_STAT_MASK, 0x00};

    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_REG, 
                                        tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    MessageBox::Show("Limp-Home fault reset.", "Limp Home Reset",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::demoModeToolStripMenuItem_CheckStateChanged(System::Object^ sender, System::EventArgs^ e)
{
    if(demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        MessageBox::Show("Demo/Config mode is used when no EVM is available, however the user wants to still access the interface. Note that " +
                           "in this mode all USB reads/writes are ignored. This mode can be used to configure the device and export the C configuration structure.",
                            "Demo/Config Mode", MessageBoxButtons::OK, MessageBoxIcon::Information);
        
        mainTabControl->Enabled = true;
        dontRunEvents = true;
        nomCurrentBox->SelectedIndex = 0;
        swLimitBox->SelectedIndex = 1;
        swDelayBox->SelectedIndex = 2;
        cooldownDelayBox->SelectedIndex = 0;
        hwLimitBox->SelectedIndex = 6;
        connectionLabel->Text = curDemoWord;
        connectionLabel->ForeColor = Color::Orange;
        capChargeModeBox->SelectedIndex = 0;
        chargeLevelBox->SelectedIndex = 0;
        capChargeDurationBox->SelectedIndex = 0;
        dvdtPeriodBox->SelectedIndex = 0;
        pulsePeriodBox->SelectedIndex = 0;
        capTurnOffBox->SelectedIndex = 0;
        chanSettingCurrentLimitBox->SelectedIndex = 0;
        inRushValueBox->SelectedIndex = 0;
        chanSettingInrushDurationBox->SelectedIndex = 0;
        threshReductionBox->SelectedIndex = 0;
        exportToolStripMenuItem->Enabled = true;
        saveSettingsToolStripMenuItem->Enabled = true;
        loaToolStripMenuItem->Enabled = true;
        dontRunEvents = false;
        loadFileIntoGUI(demoFile);
        redrawI2TGraph();
    }
    else
    {
        MessageBox::Show("Demo mode exited.", "Demo Mode", MessageBoxButtons::OK, MessageBoxIcon::Information);
        mainTabControl->Enabled = false;
        connectionLabel->Text = curdisConnectWord;
        connectionLabel->ForeColor = Color::Red;
        saveSettingsToolStripMenuItem->Enabled = false;
        loaToolStripMenuItem->Enabled = false;
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::persistToDeviceFlashToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    tHSSReceiveTransaction^ trans;
    uint8_t payload[3];

    /* Updating the buffered configuration data before  persisting */
    HSSCommand*  cmd = new HSSCommand(HSSCommand::HSS_HCS_DEV_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    /* Getting the device data */
    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_DEV_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_DEV_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the channel 1 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 0, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 0;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Getting the channel 2 data */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_CH_CONFIG, 1, NULL, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    trans = gcnew tHSSReceiveTransaction;
    trans->addr = 1;
    trans->recData = (uint8_t*)malloc(128);
    trans->reportID = HSSCommunicationManager::HSSC_CH_CONFIG_ID;
    trans->receiveSize = HSSCommand::HSS_HCS_CH_CONFIG_RESP_SIZE;
    this->commManager->getPacketData(trans);

    /* Persisting the flash */
    cmd = new HSSCommand(HSSCommand::HSS_HCS_PERSISTFLASH, payload, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    MessageBox::Show("Configuration saved to MCU flash memory", "Configuration Saved",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::loadFromDeviceFlashToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t payload[3];

    /* Pinging for the firmware version */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_LOADFLASH, payload, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    updateMainGUIFromSelection();

    MessageBox::Show("Configuration loaded from flash (if it has been peviously programmed)", "Configuration Loaded",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::eraseMCUFlashToolStripMenuItem_Click(System::Object^ sender, System::EventArgs^ e)
{
    uint8_t payload[3];

    /* Pinging for the firmware version */
    HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_ERASEFLASH, payload, 0);
    commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
    free(cmd);

    MessageBox::Show("Configuration memory space on MCU flash erased.", "Configuration Erased",
        MessageBoxButtons::OK, MessageBoxIcon::Information);
}

System::Void autohssconfigurator::AutoConfigurationGui::japaneseLanguageBox_CheckStateChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    curConnectWord = lang_connected_jp;
    curdisConnectWord = lang_disconnected_jp;
    curFaultWord = lang_fault_jp;
    curNoFaultWord = lang_nfault_jp;
    curDemoWord = lang_demomode_jp;

    if (demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        connectionLabel->Text = curDemoWord;
    }
    else if (commManager->isHidDeviceConnected())
    {
        connectionLabel->Text = curConnectWord;
    }
    else
    {
        connectionLabel->Text = curdisConnectWord;
    }

    dontRunEvents = true;
    englishToolStripMenuItem->CheckState = CheckState::Unchecked;
    japaneseLanguageBox->CheckState = CheckState::Checked;
    dontRunEvents = false;

    channelEnGroupBox->Text = lang_chen_jp;
    enChanBoxCC->Items[0] = lang_chan1_jp;
    enChanBoxCC->Items[1] = lang_chan2_jp;

    detDeviceGroupBox->Text = lang_detdev_jp;
    partnumLab->Text = lang_partnum_jp;
    firmwareRevLab->Text = lang_firmware_jp;
    partDesLab->Text = lang_devdes_jp;

    pwmGroupBox->Text = lang_pwm_jp;
    freqCCLabel->Text = lang_pwmfreq_jp;
    dutyCCLabel->Text = lang_pwmduty_jp;
    enPWMccLab->Text = lang_en_jp;

    graphSettingBox->Text = lang_graphsetting_jp;
    graphSourceLab->Text = lang_cursensesrc_jp;
    graphCountLab->Text = lang_cursensecnt_jp;
    graphAvgLab->Text = lang_cursenseavg_jp;
    enGraphLab->Text = lang_cursenseen_jp;
    senseChannelAll->Text = lang_cursenseall_en;
    senseChannelBox->Items[0] = lang_chan1_jp;
    senseChannelBox->Items[1] = lang_chan2_jp;
    senseSourceSelBox->Items[0] = lang_loadcurrent_jp;
    senseSourceSelBox->Items[1] = lang_outputvoltage_jp;
    senseChannelAll->Text = lang_cursenseall_jp;
    senseChart->Titles[0]->Text = senseSourceSelBox->Text;


    faultDiagBox->Text = lang_faultdiag_jp;
    chan1Label->Text = lang_chan1_jp;
    chan2Label->Text = lang_chan2_jp;
    senseChart->Series[0]->Name = lang_chan1_jp;
    senseChart->Series[1]->Name = lang_chan2_jp;
    globalFaultLangLabel->Text = lang_globalchan_jp;
  
    quickCommandBox->Text = lang_quickcommand_jp;
    lpmCurrentLab->Text = lang_lpmcurrent_jp;
    ccLPMLab->Text = lang_lpmcurrent_jp;
    sleepModeButton->Text = lang_sleepmode_jp;
    forceLimpLab->Text = lang_forcelimp_jp;

    mainTabControl->TabPages[0]->Text = lang_commandcenter_jp;
    mainTabControl->TabPages[1]->Text = lang_i2ttuner_jp;
    mainTabControl->TabPages[2]->Text = lang_devsetting_jp;
    mainTabControl->TabPages[3]->Text = lang_chansetting_jp;
    mainTabControl->TabPages[4]->Text = lang_console_jp;
    mainTabControl->TabPages[5]->Text = lang_log_jp;

    connectionStatusLabel->Text = lang_constatus_jp;
    inputVoltageDescStripLabel->Text = lang_involtage_jp;
    toolStatusLabel->Text = lang_status_jp;

    fileToolStripMenuItem->Text = lang_file_jp;
    saveSettingsToolStripMenuItem->Text = lang_savesetting_jp;
    loaToolStripMenuItem->Text = lang_loadsetting_jp;
    exitToolStripMenuItem->Text = lang_exit_jp;
    settingsToolStripMenuItem->Text = lang_settings_jp;
    autoPersistSettingsToolStripMenuItem->Text = lang_persistrefresh_jp;
    manualPersistEnableCheckBox->Text = lang_persist_jp;
    manualRefreshToolCheck->Text = lang_manualrefresh_jp;
    toolsToolStripMenuItem->Text = lang_tools_jp;
    enterDFUModeToolStripMenuItem->Text = lang_dfumode_jp;
    languageToolStripMenuItem->Text = lang_lang_jp;
    helpToolStripMenuItem->Text = lang_help_jp;
    demoModeToolStripMenuItem->Text = lang_demomode_jp;
    aboutToolStripMenuItem->Text = lang_about_jp;
    flashPersistToolStripMenuItem->Text = lang_mcuflash_jp;
    persistToDeviceFlashToolStripMenuItem->Text = lang_saveflash_jp;
    loadFromDeviceFlashToolStripMenuItem->Text = lang_loadflash_jp;
    eraseMCUFlashToolStripMenuItem->Text = lang_eraseflash_jp;

    chaNumi2tLab->Text = lang_channum_jp;
    wirepropertyBox->Text = lang_wireprop_jp;
    wiregaugeLabel->Text = lang_gauge_jp;
    customWireLab->Text = lang_customwire_jp;
    pulseBox->Text = lang_pulsecheck_jp;
    rmsCurrentLab->Text = lang_rmscurrent_jp;
    durLab->Text = lang_duration_jp;
    tuningParametersLab->Text = lang_tuningparam_jp;
    nomCurrentLabel->Text = lang_nomcurrent_jp;
    i2tTripLabel->Text = lang_i2ttripcc_jp;
    swCurrentLabel->Text = lang_delshut_jp;
    cooldownLabel->Text = lang_cooldel_jp;
    resetLatchButton->Text = lang_reslatch_jp;
    swDelayLabel->Text = lang_softdel_jp;
    hwLimitLabel->Text = lang_immshut_jp;
    parameterDesBox->Text = lang_paramdesc_jp;
    registerLab->Text = lang_regcc_jp;
    regDescLab->Text = lang_pdes_jp;
    generalBox->Text = lang_general_jp;
    parallelChanLab->Text = lang_parallel_jp;
    parallelBox->Text = lang_oneandtwo_jp;
    watchdogLab->Text = lang_watchdog_jp;
    wdEnableBox->Text = lang_enable_jp;
    durationLab->Text = lang_watchdogdur_jp;
    pwmDelayLab->Text = lang_chanpwmdelay_jp;
    pwmDelayBox->Text = lang_enable_jp;
    abistLab->Text = lang_abistenable_jp;
    abistSNSEnable->Text = lang_abistenablesns_jp;
    abistBGEnable->Text = lang_abistenablebg_jp;
    abistVDDEnable->Text = lang_abistenablevdd_jp;
    abistADCEnable->Text = lang_abistenableadc_jp;
    lowPowerStateLab->Text = lang_lpmstate_jp;
    lpmRadioButton->Text = lang_lowpower_jp;
    activeRadioButton->Text = lang_active_jp;
    globalFaultBox->Text = lang_globalchan_jp;
    porGlobLab->Text = lang_por_jp;
    limpGlobLab->Text = lang_limphomeerr_jp;
    wdtGlobLab->Text = lang_wdterr_jp;
    spiGlobLab->Text = lang_spierr_jp;
    vbbGlobLab->Text = lang_vbbwarn_jp;
    vbbUVGlobLab->Text = lang_vbbuvlo_jp;
    vddUvGlobLab->Text = lang_vdduvlo_jp;
    faultPinLab->Text = lang_faultpin_jp;
    statsGroupBox->Text = lang_stats_jp;
    falDescLab->Text = lang_pdes_en;
    currentStatLab->Text = lang_curstat_jp;
    numOfFaultLab->Text = lang_numoffault_jp;
    lastFaultLab->Text = lang_lastfault_jp;
    settingLab->Text = lang_setting_jp;
    manualUpdateButton->Text = lang_manup_jp;
    clearAllStatsButton->Text = lang_clearall_jp;
    adcEnLab->Text = lang_adcen_jp;
    adcSrcEnLab->Text = lang_adcsrcen_jp;
    adcSourceOutVBox->Text = lang_outputvoltage_jp;
    adcSouceInVBox->Text = lang_involtage_jp;
    adcSourceJunctionBox->Text = lang_juntemp_jp;
    adcSourceCurrentBox->Text = lang_loadcurrent_jp;
    faultMaskLab->Text = lang_faultmask_jp;
    faultMaskCLBox->Text = lang_curlimit_jp;
    faultMaskShortVBBBox->Text = lang_shortvbb_jp;
    faultMaskOpenLoadBox->Text = lang_openloadoff_jp;
    faultMaskOpenLoadOnBox->Text = lang_openloadon_jp;
    faultMaskWDTBox->Text = lang_watchdogerr_jp;
    faultMaskVBBUVLO->Text = lang_vbbuvlo_jp;
    faultMaskSPIBox->Text = lang_spierr_jp;
    latchFaultLab->Text = lang_latchfault_jp;
    senseResistorLabel->Text = lang_senseres_jp;
    toleranceLab->Text = lang_sensetol_jp;
    faulChan1Lab->Text = lang_chan1_jp;
    faulChan2Lab->Text = lang_chan2_jp;
    thermalWarnLab->Text = lang_thermalwarn_jp;
    openLoadOnLab->Text = lang_openloadon_jp;
    openLoadOffLab->Text = lang_openloadoff_jp;
    shortToVBBLab->Text = lang_shortvbb_jp;
    ilimLab->Text = lang_curlimit_jp;
    thermalShutdowLab->Text = lang_thermalshut_jp;
    voutErrorLab->Text = lang_vouterr_jp;
    i2tFaultLab->Text = lang_i2t_jp;
    enableI2TBox->Text = lang_eni2t_jp;
    lpmIndicatorBox->Text = lang_lpmwake_jp;
    lpmWakeChan1Lab->Text = lang_chan1_jp;
    lpmWakeChan2Lab->Text = lang_chan2_jp;
    lpmLabel1->Text = lang_nottriggered_jp;
    lpmLabel2->Text = lang_nottriggered_jp;
    ccChanLab->Text = lang_channum_jp;
    operationalBox->Text = lang_operational_jp;
    chanEnCCLab->Text = lang_chen_jp;
    pwmCCLab->Text = lang_pwmen_jp;
    freqCCLab->Text = lang_pwmfreq_jp;
    dutyCCLab->Text = lang_pwmduty_jp;
    configurationBox->Text = lang_configuration_jp;
    lhiBehaveLab->Text = lang_lhiinput_jp;
    lpmExitCCLab->Text = lang_lpmextcurrent_jp;
    adcSourceEnLab->Text = lang_adcsrcen_jp;
    chanVBBADCBox->Text = lang_ccadcvol_jp;
    chanJunctionADCBox->Text = lang_ccadctemp_jp;
    chanCurrentADCBox->Text = lang_ccadccur_jp;
    adcInputScaleLab->Text = lang_adcenscale_jp;
    openLoadOnThreshLab->Text = lang_openonthresh_jp;
    openLoadOnScalLab->Text = lang_openonscale_jp;
    openLoadOffPullLab->Text = lang_openoffpull_jp;
    openLoadOffStrLab->Text = lang_openoffstren_jp;
    openLoadBlankLab->Text = lang_openloadblank_jp;
    vbbShortLab->Text = lang_vbbshortpull_jp;
    olScalingBox->Text = lang_enable_jp;
    olOffPullBox->Text = lang_enable_jp;
    shutdownRetLab->Text = lang_shutretry_jp;
    shutdownLatchBox->Text = lang_latched_jp;
    shutdownRetryBox->Text = lang_retry_jp;
    slewLab->Text = lang_slew_jp;
    capChargeBox->Text = lang_capcharge_jp;
    capChargeLab->Text = lang_capcharge_jp + ":";
    capChargeEnableBox->Text = lang_enable_jp;
    capChargeModeLab->Text = lang_capmode_jp;
    chargeRateLab->Text = lang_chargerate_jp;
    capChargeDurLab->Text = lang_chargerate_jp;
    dvdtLab->Text = lang_dvdtrate_jp;
    pulsePerLab->Text = lang_pulseper_jp;
    ocOffLab->Text = lang_overcuroff_jp;
    overCurBox->Text = lang_overcurprot_jp;
    immOverCurLab->Text = lang_delshut_jp;
    inrushValLab->Text = lang_inrushval_jp;
    overCurDurLab->Text = lang_chargedur_jp;
    threshReductLab->Text = lang_threshred_jp;
    i2tParamBox->Text = lang_i2tparam_jp;
    chanSettingI2TEnBox->Text = lang_eni2t_jp;
    i2tWarnText->Text = lang_i2tnote_jp;
    nomCurCCLab->Text = lang_nomcur_jp;
    tripValCCLab->Text = lang_i2ttripcc_jp;
    cooldownCCLab->Text = lang_cooldel_jp;
    delShutCCLab->Text = lang_delshut_jp;
    softDelCCLab->Text = lang_softdel_jp;
    commmandsLab->Text = lang_coms_jp;
    histLab->Text = lang_hist_jp;
    conComLab->Text = lang_com_jp;
    enableLogButton->Text = lang_logen_jp;
    logLabel->Text = lang_logfile_jp;
    exportToolStripMenuItem->Text = lang_exportf_jp;
    cStructToolStripMenuItem->Text = lang_cheader_jp;

    this->Text = lang_sfconfig_jp;
}

System::Void autohssconfigurator::AutoConfigurationGui::englishToolStripMenuItem_CheckStateChanged(System::Object^ sender, System::EventArgs^ e)
{
    if (dontRunEvents == true)
    {
        return;
    }

    curConnectWord = lang_connected_en;
    curdisConnectWord = lang_disconnected_en;
    curFaultWord = lang_fault_en;
    curNoFaultWord = lang_nfault_en;
    curDemoWord = lang_demomode_en;

    if (demoModeToolStripMenuItem->CheckState == CheckState::Checked)
    {
        connectionLabel->Text = curDemoWord;
    }
    else if (commManager->isHidDeviceConnected())
    {
        connectionLabel->Text = curConnectWord;
    }
    else
    {
        connectionLabel->Text = curdisConnectWord;
    }

    dontRunEvents = true;
    englishToolStripMenuItem->CheckState = CheckState::Checked;
    japaneseLanguageBox->CheckState = CheckState::Unchecked;
    dontRunEvents = false;

    channelEnGroupBox->Text = lang_chen_en;
    enChanBoxCC->Items[0] = lang_chan1_en;
    enChanBoxCC->Items[1] = lang_chan2_en;

    detDeviceGroupBox->Text = lang_detdev_en;
    partnumLab->Text = lang_partnum_en;
    firmwareRevLab->Text = lang_firmware_en;
    partDesLab->Text = lang_devdes_en;

    pwmGroupBox->Text = lang_pwm_en;
    freqCCLabel->Text = lang_pwmfreq_en;
    dutyCCLabel->Text = lang_pwmduty_en;
    enPWMccLab->Text = lang_en_en;

    graphSettingBox->Text = lang_graphsetting_en;
    graphSourceLab->Text = lang_cursensesrc_en;
    graphCountLab->Text = lang_cursensecnt_en;
    graphAvgLab->Text = lang_cursenseavg_en;
    enGraphLab->Text = lang_cursenseen_en;
    senseChannelAll->Text = lang_cursenseall_en;
    senseChannelBox->Items[0] = lang_chan1_en;
    senseChannelBox->Items[1] = lang_chan2_en;
    senseSourceSelBox->Items[0] = lang_loadcurrent_en;
    senseSourceSelBox->Items[1] = lang_outputvoltage_en;
    senseChannelAll->Text = lang_cursenseall_en;
    senseChart->Titles[0]->Text = senseSourceSelBox->Text;


    faultDiagBox->Text = lang_faultdiag_en;
    chan1Label->Text = lang_chan1_en;
    chan2Label->Text = lang_chan2_en;
    senseChart->Series[0]->Name = lang_chan1_en;
    senseChart->Series[1]->Name = lang_chan2_en;
    globalFaultLangLabel->Text = lang_globalchan_en;

    quickCommandBox->Text = lang_quickcommand_en;
    lpmCurrentLab->Text = lang_lpmcurrent_en;
    ccLPMLab->Text = lang_lpmcurrent_en;
    sleepModeButton->Text = lang_sleepmode_en;
    forceLimpLab->Text = lang_forcelimp_en;

    mainTabControl->TabPages[0]->Text = lang_commandcenter_en;
    mainTabControl->TabPages[1]->Text = lang_i2ttuner_en;
    mainTabControl->TabPages[2]->Text = lang_devsetting_en;
    mainTabControl->TabPages[3]->Text = lang_chansetting_en;
    mainTabControl->TabPages[4]->Text = lang_console_en;
    mainTabControl->TabPages[5]->Text = lang_log_en;

    connectionStatusLabel->Text = lang_constatus_en;
    inputVoltageDescStripLabel->Text = lang_involtage_en;
    toolStatusLabel->Text = lang_status_en;

    fileToolStripMenuItem->Text = lang_file_en;
    saveSettingsToolStripMenuItem->Text = lang_savesetting_en;
    loaToolStripMenuItem->Text = lang_loadsetting_en;
    exitToolStripMenuItem->Text = lang_exit_en;
    settingsToolStripMenuItem->Text = lang_settings_en;
    autoPersistSettingsToolStripMenuItem->Text = lang_persistrefresh_en;
    manualPersistEnableCheckBox->Text = lang_persist_en;
    manualRefreshToolCheck->Text = lang_manualrefresh_en;
    toolsToolStripMenuItem->Text = lang_tools_en;
    enterDFUModeToolStripMenuItem->Text = lang_dfumode_en;
    languageToolStripMenuItem->Text = lang_lang_en;
    helpToolStripMenuItem->Text = lang_help_en;
    demoModeToolStripMenuItem->Text = lang_demomode_en;
    aboutToolStripMenuItem->Text = lang_about_en;
    flashPersistToolStripMenuItem->Text = lang_mcuflash_en;
    persistToDeviceFlashToolStripMenuItem->Text = lang_saveflash_en;
    loadFromDeviceFlashToolStripMenuItem->Text = lang_loadflash_en;
    eraseMCUFlashToolStripMenuItem->Text = lang_eraseflash_en;

    chaNumi2tLab->Text = lang_channum_en;
    wirepropertyBox->Text = lang_wireprop_en;
    wiregaugeLabel->Text = lang_gauge_en;
    customWireLab->Text = lang_customwire_en;
    pulseBox->Text = lang_pulsecheck_en;
    rmsCurrentLab->Text = lang_rmscurrent_en;
    durLab->Text = lang_duration_en;
    tuningParametersLab->Text = lang_tuningparam_en;
    nomCurrentLabel->Text = lang_nomcurrent_en;
    i2tTripLabel->Text = lang_i2ttripcc_en;
    swCurrentLabel->Text = lang_delshut_en;
    cooldownLabel->Text = lang_cooldel_en;
    resetLatchButton->Text = lang_reslatch_en;
    swDelayLabel->Text = lang_softdel_en;
    hwLimitLabel->Text = lang_immshut_en;
    parameterDesBox->Text = lang_paramdesc_en;
    registerLab->Text = lang_regcc_en;
    regDescLab->Text = lang_pdes_en;
    generalBox->Text = lang_general_en;
    parallelChanLab->Text = lang_parallel_en;
    parallelBox->Text = lang_oneandtwo_en;
    watchdogLab->Text = lang_watchdog_en;
    wdEnableBox->Text = lang_enable_en;
    durationLab->Text = lang_watchdogdur_en;
    pwmDelayLab->Text = lang_chanpwmdelay_en;
    pwmDelayBox->Text = lang_enable_en;
    abistLab->Text = lang_abistenable_en;
    abistSNSEnable->Text = lang_abistenablesns_en;
    abistBGEnable->Text = lang_abistenablebg_en;
    abistVDDEnable->Text = lang_abistenablevdd_en;
    abistADCEnable->Text = lang_abistenableadc_en;
    lowPowerStateLab->Text = lang_lpmstate_en;
    lpmRadioButton->Text = lang_lowpower_en;
    activeRadioButton->Text = lang_active_en;
    globalFaultBox->Text = lang_globalchan_en;
    porGlobLab->Text = lang_por_en;
    limpGlobLab->Text = lang_limphomeerr_en;
    wdtGlobLab->Text = lang_wdterr_en;
    spiGlobLab->Text = lang_spierr_en;
    vbbGlobLab->Text = lang_vbbwarn_en;
    vbbUVGlobLab->Text = lang_vbbuvlo_en;
    vddUvGlobLab->Text = lang_vdduvlo_en;
    faultPinLab->Text = lang_faultpin_en;
    statsGroupBox->Text = lang_stats_en;
    falDescLab->Text = lang_pdes_en;
    currentStatLab->Text = lang_curstat_en;
    numOfFaultLab->Text = lang_numoffault_en;
    lastFaultLab->Text = lang_lastfault_en;
    settingLab->Text = lang_setting_en;
    manualUpdateButton->Text = lang_manup_en;
    clearAllStatsButton->Text = lang_clearall_en;
    adcEnLab->Text = lang_adcen_en;
    adcSrcEnLab->Text = lang_adcsrcen_en;
    adcSourceOutVBox->Text = lang_outputvoltage_en;
    adcSouceInVBox->Text = lang_involtage_en;
    adcSourceJunctionBox->Text = lang_juntemp_en;
    adcSourceCurrentBox->Text = lang_loadcurrent_en;
    faultMaskLab->Text = lang_faultmask_en;
    faultMaskCLBox->Text = lang_curlimit_en;
    faultMaskShortVBBBox->Text = lang_shortvbb_en;
    faultMaskOpenLoadBox->Text = lang_openloadoff_en;
    faultMaskOpenLoadOnBox->Text = lang_openloadon_en;
    faultMaskWDTBox->Text = lang_watchdogerr_en;
    faultMaskVBBUVLO->Text = lang_vbbuvlo_en;
    faultMaskSPIBox->Text = lang_spierr_en;
    latchFaultLab->Text = lang_latchfault_en;
    senseResistorLabel->Text = lang_senseres_en;
    toleranceLab->Text = lang_sensetol_en;
    faulChan1Lab->Text = lang_chan1_en;
    faulChan2Lab->Text = lang_chan2_en;
    thermalWarnLab->Text = lang_thermalwarn_en;
    openLoadOnLab->Text = lang_openloadon_en;
    openLoadOffLab->Text = lang_openloadoff_en;
    shortToVBBLab->Text = lang_shortvbb_en;
    ilimLab->Text = lang_curlimit_en;
    thermalShutdowLab->Text = lang_thermalshut_en;
    voutErrorLab->Text = lang_vouterr_en;
    i2tFaultLab->Text = lang_i2t_en;
    enableI2TBox->Text = lang_eni2t_en;
    lpmIndicatorBox->Text = lang_lpmwake_en;
    lpmWakeChan1Lab->Text = lang_chan1_en;
    lpmWakeChan2Lab->Text = lang_chan2_en;
    lpmLabel1->Text = lang_nottriggered_en;
    lpmLabel2->Text = lang_nottriggered_en;
    ccChanLab->Text = lang_channum_en;
    operationalBox->Text = lang_operational_en;
    chanEnCCLab->Text = lang_chen_en;
    pwmCCLab->Text = lang_pwmen_en;
    freqCCLab->Text = lang_pwmfreq_en;
    dutyCCLab->Text = lang_pwmduty_en;
    configurationBox->Text = lang_configuration_en;
    lhiBehaveLab->Text = lang_lhiinput_en;
    lpmExitCCLab->Text = lang_lpmextcurrent_en;
    adcSourceEnLab->Text = lang_adcsrcen_en;
    chanVBBADCBox->Text = lang_ccadcvol_en;
    chanJunctionADCBox->Text = lang_ccadctemp_en;
    chanCurrentADCBox->Text = lang_ccadccur_en;
    adcInputScaleLab->Text = lang_adcenscale_en;
    openLoadOnThreshLab->Text = lang_openonthresh_en;
    openLoadOnScalLab->Text = lang_openonscale_en;
    openLoadOffPullLab->Text = lang_openoffpull_en;
    openLoadOffStrLab->Text = lang_openoffstren_en;
    openLoadBlankLab->Text = lang_openloadblank_en;
    vbbShortLab->Text = lang_vbbshortpull_en;
    olScalingBox->Text = lang_enable_en;
    olOffPullBox->Text = lang_enable_en;
    shutdownRetLab->Text = lang_shutretry_en;
    shutdownLatchBox->Text = lang_latched_en;
    shutdownRetryBox->Text = lang_retry_en;
    slewLab->Text = lang_slew_en;
    capChargeBox->Text = lang_capcharge_en;
    capChargeLab->Text = lang_capcharge_en + ":";
    capChargeEnableBox->Text = lang_enable_en;
    capChargeModeLab->Text = lang_capmode_en;
    chargeRateLab->Text = lang_chargerate_en;
    capChargeDurLab->Text = lang_chargerate_en;
    dvdtLab->Text = lang_dvdtrate_en;
    pulsePerLab->Text = lang_pulseper_en;
    ocOffLab->Text = lang_overcuroff_en;
    overCurBox->Text = lang_overcurprot_en;
    immOverCurLab->Text = lang_delshut_en;
    inrushValLab->Text = lang_inrushval_en;
    overCurDurLab->Text = lang_chargedur_en;
    threshReductLab->Text = lang_threshred_en;
    i2tParamBox->Text = lang_i2tparam_en;
    chanSettingI2TEnBox->Text = lang_eni2t_en;
    i2tWarnText->Text = lang_i2tnote_en;
    nomCurCCLab->Text = lang_nomcur_en;
    tripValCCLab->Text = lang_i2ttripcc_en;
    cooldownCCLab->Text = lang_cooldel_en;
    delShutCCLab->Text = lang_delshut_en;
    softDelCCLab->Text = lang_softdel_en;
    commmandsLab->Text = lang_coms_en;
    histLab->Text = lang_hist_en;
    conComLab->Text = lang_com_en;
    enableLogButton->Text = lang_logen_en;
    logLabel->Text = lang_logfile_en;
    exportToolStripMenuItem->Text = lang_exportf_en;
    cStructToolStripMenuItem->Text = lang_cheader_en;

    this->Text = lang_sfconfig_en;
}

/* Command handler for raw value reads */
void autohssconfigurator::AutoConfigurationGui::handleRawReadResponse(sHSSReceiveTransaction^ trans)
{
    if (this->InvokeRequired)
    {
        UpdateSenseCallback^ updateCallback = gcnew UpdateSenseCallback(this,
            &autohssconfigurator::AutoConfigurationGui::handleRawReadResponse);
        this->Invoke(updateCallback, trans);
    }
    else
    {
        String^ dateString = DateTime::Now.ToString("[hh:mm:ss] ");

        /* Reporting the value of the raw read to the log */
        char szNum[32];
        sprintf_s(szNum, "%02x", trans->recData[1]);
        String^ statusString = gcnew String(szNum);
        sprintf_s(szNum, "%02x", trans->recData[2]);
        String^ tempString = gcnew String(szNum);
        sprintf_s(szNum, "%02x", trans->recData[3]);
        String^ tempString2 = gcnew String(szNum);
        this->commandLogBox->AppendText(dateString + "Register read result: 0x" + tempString->ToUpper() + tempString2->ToUpper() +
                                        " Status: 0x" + statusString->ToUpper() + Environment::NewLine);
    }
}

void autohssconfigurator::AutoConfigurationGui::updateCurrentLimitSettings(uint8_t chanNum)
{
    /* Current Limit Configuration */
    if (((fuseConfig->getILIMConfig(chanNum) & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK)) == cap_chrg_ch1_en_1_0x0)
    {
        chargeLevelBox->Enabled = false;
        capChargeDurationBox->Enabled = false;
        capChargeModeBox->Enabled = false;
        dvdtPeriodBox->Enabled = false;
        capTurnOffBox->Enabled = false;
        chanSettingCurrentLimitBox->Enabled = true;
        inRushValueBox->Enabled = true;
        chanSettingInrushDurationBox->Enabled = true;
        threshReductionBox->Enabled = true;
        pulsePeriodBox->Enabled = false;
    }
    else if (((fuseConfig->getILIMConfig(chanNum) & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK)) == cap_chrg_ch1_en_2_0x1)
    {
        capChargeModeBox->SelectedIndex = 0;
        capChargeModeBox->Enabled = true;
        chargeLevelBox->Enabled = false;
        capChargeDurationBox->Enabled = true;
        dvdtPeriodBox->Enabled = true;
        capTurnOffBox->Enabled = false;
        chanSettingCurrentLimitBox->Enabled = true;
        inRushValueBox->Enabled = false;
        chanSettingInrushDurationBox->Enabled = false;
        threshReductionBox->Enabled = false;
        pulsePeriodBox->Enabled = false;

        uint16_t curdvdtValue = (fuseConfig->getILIMConfig(chanNum) & TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK) >> 4;

        switch (curdvdtValue)
        {
        case DVDT_P33:
            dvdtPeriodBox->SelectedIndex = 0;
            break;
        case DVDT_P50:
            dvdtPeriodBox->SelectedIndex = 1;
            break;
        case DVDT_P67:
            dvdtPeriodBox->SelectedIndex = 2;
            break;
        case DVDT_P90:
            dvdtPeriodBox->SelectedIndex = 3;
            break;
        case DVDT_1P0:
            dvdtPeriodBox->SelectedIndex = 4;
            break;
        case DVDT_1P1:
            dvdtPeriodBox->SelectedIndex = 5;
            break;
        case DVDT_1P33:
            dvdtPeriodBox->SelectedIndex = 6;
            break;
        case DVDT_1P66:
            dvdtPeriodBox->SelectedIndex = 7;
            break;
        case DVDT_2P22:
            dvdtPeriodBox->SelectedIndex = 8;
            break;
        default:
            fuseConfig->setILIMConfig((DVDT_P33 << 4) |
                (fuseConfig->getILIMConfig(chanNum) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), chanNum);
            uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(chanNum) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(chanNum) >> 8) };

            /* Sending the updated data */
            HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (HSSC_CHAN_NUM_OFFSET * chanNum), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            commManager->sendPacketData(cmd->getBufferData(), cmd->getCommandSize());
            dvdtPeriodBox->SelectedIndex = 0;
            break;
        }
    }
    else if (((fuseConfig->getILIMConfig(chanNum) & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK)) == cap_chrg_ch1_en_3_0x2)
    {
        capChargeModeBox->SelectedIndex = 1;
        capChargeModeBox->Enabled = true;
        chargeLevelBox->Enabled = true;
        capChargeDurationBox->Enabled = true;
        dvdtPeriodBox->Enabled = false;
        capTurnOffBox->Enabled = false;
        chanSettingCurrentLimitBox->Enabled = true;
        inRushValueBox->Enabled = false;
        chanSettingInrushDurationBox->Enabled = false;
        threshReductionBox->Enabled = false;
        pulsePeriodBox->Enabled = false;
    }
    else if (((fuseConfig->getILIMConfig(chanNum) & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK)) == cap_chrg_ch1_en_4_0x3)
    {
        capChargeModeBox->SelectedIndex = 2;
        capChargeModeBox->Enabled = false;
        chargeLevelBox->Enabled = false;
        capChargeDurationBox->Enabled = false;
        dvdtPeriodBox->Enabled = false;
        capTurnOffBox->Enabled = true;
        chanSettingCurrentLimitBox->Enabled = true;
        inRushValueBox->Enabled = false;
        chanSettingInrushDurationBox->Enabled = false;
        threshReductionBox->Enabled = false;
        pulsePeriodBox->Enabled = true;

        uint8_t curPulsePeriod = ((fuseConfig->getILIMConfig(chanNum)
            & TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK) >> 4);

        switch (curPulsePeriod)
        {
        case PUL_PER150:
            pulsePeriodBox->SelectedIndex = 0;
            break;
        case PUL_PER200:
            pulsePeriodBox->SelectedIndex = 1;
            break;
        case PUL_PER250:
            pulsePeriodBox->SelectedIndex = 2;
            break;
        case PUL_PER300:
            pulsePeriodBox->SelectedIndex = 3;
            break;
        default:
            fuseConfig->setILIMConfig((PUL_PER150 << 4) |
                (fuseConfig->getILIMConfig(chanNum) & ~TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK), chanNum);
            uint8_t tempBuffer[HSSC_PAYLOAD_PACKET_SIZE] = { HSSC_RAW_WRITE_SIZE, (uint8_t)(fuseConfig->getILIMConfig(chanNum) & 0xFF), (uint8_t)(fuseConfig->getILIMConfig(chanNum) >> 8) };

            /* Sending the updated data */
            HSSCommand* cmd = new HSSCommand(HSSCommand::HSS_HCS_RAW_WRITE, TPS2HC10S_ILIM_CONFIG_CH1_REG + (HSSC_CHAN_NUM_OFFSET * chanNum), tempBuffer, HSSC_PAYLOAD_PACKET_SIZE);
            sendRawWriteGUIResponse(cmd);
            free(cmd);
            pulsePeriodBox->SelectedIndex = 0;
            break;
        }
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::updateGuiFromConfig(bool dontRunEventUpdate)
{
    uint8_t tempVal1;
    bool oldConfigValue;

    if (this->enChanBoxCC->InvokeRequired)
    {
        UpdateConfigCallback^ updateCallback = gcnew UpdateConfigCallback(this, 
            &autohssconfigurator::AutoConfigurationGui::updateGuiFromConfig);
        this->Invoke(updateCallback, dontRunEventUpdate);
    }
    else
    {
        oldConfigValue = dontRunEvents;
        dontRunEvents = dontRunEventUpdate;

        /* Enable boxes */
        if (fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH1_ON_MASK)
        {
            enChanBoxCC->SetItemCheckState(0, CheckState::Checked);
        }
        else
        {
            enChanBoxCC->SetItemCheckState(0, CheckState::Unchecked);
        }
        
        if (fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH2_ON_MASK)
        {
            enChanBoxCC->SetItemCheckState(1, CheckState::Checked);
        }
        else
        {
            enChanBoxCC->SetItemCheckState(1, CheckState::Unchecked);
        }

        /* Fault Mask */
        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_VBB_UVLO_MASK)
        {
            faultMaskVBBUVLO->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskVBBUVLO->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_WD_ERR_MASK)
        {
            faultMaskWDTBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskWDTBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_SPI_ERR_MASK)
        {
            faultMaskSPIBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskSPIBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_OL_ON_MASK)
        {
            faultMaskOpenLoadOnBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskOpenLoadOnBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_OL_OFF_MASK)
        {
            faultMaskOpenLoadBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskOpenLoadBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_SHRT_VBB_MASK)
        {
            faultMaskShortVBBBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskShortVBBBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDeviceFaultMask() & TPS2HC10S_FAULT_MASK_MASK_ILIMIT_MASK)
        {
            faultMaskCLBox->CheckState = CheckState::Checked;
        }
        else
        {
            faultMaskCLBox->CheckState = CheckState::Unchecked;
        }

        /* Device Stat Config */
        if (fuseConfig->getDevSAFConfig() & TPS2HC10S_DEVICE_SAF_ADC_EN_MASK)
        {
            adcEnableBox->CheckState = CheckState::Checked;
        }
        else
        {
            adcEnableBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDevSAFConfig() & TPS2HC10S_DEVICE_SAF_ISNS_ABIST_RUN_MASK)
        {
            abistSNSEnable->CheckState = CheckState::Checked;
        }
        else
        {
            abistSNSEnable->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDevSAFConfig() & TPS2HC10S_DEVICE_SAF_BG_ABIST_RUN_MASK)
        {
            abistBGEnable->CheckState = CheckState::Checked;
        }
        else
        {
            abistBGEnable->CheckState = CheckState::Unchecked;
        }

       /* if (fuseConfig->getDevSAFConfig() & TPS2HC10S_DEVICE_SAF_VDDUVP_ABIST_RUN_MASK)
        {
            abistVDDEnable->CheckState = CheckState::Checked;
        }
        else
        {
            abistVDDEnable->CheckState = CheckState::Unchecked;
        }*/

        if (fuseConfig->getDevSAFConfig() & TPS2HC10S_DEVICE_SAF_ADC_ABIST_RUN_MASK)
        {
            abistADCEnable->CheckState = CheckState::Checked;
        }
        else
        {
            abistADCEnable->CheckState = CheckState::Unchecked;
        }

        /* Device Config */
        if (fuseConfig->isLHIForced() == true)
        {
            forceLimpCheckBox->CheckState = CheckState::Checked;
        }
        else
        {
            forceLimpCheckBox->CheckState = CheckState::Unchecked;
        }


        if (fuseConfig->getDevConfig() & TPS2HC10S_DEV_CONFIG_FLT_LTCH_DIS_MASK)
        {
            latchFaultBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            latchFaultBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getDevConfig() & TPS2HC10S_DEV_CONFIG_WD_EN_MASK)
        {
            wdEnableBox->CheckState = CheckState::Checked;
        }
        else
        {
            wdEnableBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDevConfig() & TPS2HC10S_DEV_CONFIG_PARALLEL_12_MASK)
        {
            parallelBox->CheckState = CheckState::Checked;
        }
        else
        {
            parallelBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getDevConfig() & TPS2HC10S_DEV_CONFIG_PWM_SHIFT_DIS_MASK)
        {
            pwmDelayBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            pwmDelayBox->CheckState = CheckState::Checked;
        }

        wdtDurationBox->SelectedIndex =
            (fuseConfig->getDevConfig() & TPS2HC10S_DEV_CONFIG_WD_TO_MASK) >> 1;

        
        if (fuseConfig->getADCConfig() & TPS2HC10S_ADC_CONFIG_ADC_ISNS_DIS_MASK)
        {
            adcSourceCurrentBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            adcSourceCurrentBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getADCConfig() & TPS2HC10S_ADC_CONFIG_ADC_TSNS_DIS_MASK)
        {
            adcSourceJunctionBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            adcSourceJunctionBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getADCConfig() & TPS2HC10S_ADC_CONFIG_ADC_VSNS_DIS_MASK)
        {
            adcSourceOutVBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            adcSourceOutVBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getADCConfig() & TPS2HC10S_ADC_CONFIG_ADC_VBB_DIS_MASK)
        {
            adcSouceInVBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            adcSouceInVBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getLowPowerMode() & TPS2HC10S_LPM_LPM_MASK)
        {
            lpmRadioButton->Checked = true;
            activeRadioButton->Checked = false;
        }
        else
        {
            lpmRadioButton->Checked = false;
            activeRadioButton->Checked = true;
        }

        lpmCh1ThreshBox->SelectedIndex =
            (fuseConfig->getLowPowerMode() & TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_MASK) >> TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_OFS;

        lpmCh2ThreshBox->SelectedIndex =
            (fuseConfig->getLowPowerMode() & TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_MASK) >> TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_OFS;

        cmdPWMFreqBox1->SelectedIndex = (fuseConfig->getPWMConfig(0) &
            TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_OFS;
        cmdPWMFreqBox2->SelectedIndex = (fuseConfig->getPWMConfig(1) &
            TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_OFS;
        cmdPWMDutyBox1->SelectedIndex = (fuseConfig->getPWMConfig(0) &
            TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_DTY_CH1_OFS;
        cmdPWMDutyBox2->SelectedIndex = (fuseConfig->getPWMConfig(1) &
            TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_DTY_CH1_OFS;

        if (fuseConfig->getPWMConfig(0) & TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK)
        {
            cmdPWMEn1->CheckState = CheckState::Checked;
        }
        else
        {
            cmdPWMEn1->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getPWMConfig(1) & TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK)
        {
            cmdPWMEn2->CheckState = CheckState::Checked;
        }
        else
        {
            cmdPWMEn2->CheckState = CheckState::Unchecked;
        }

        /* I2T Tab */
        tempVal1 = i2tChanNumBox->SelectedIndex;
        nomCurrentBox->SelectedIndex = fuseConfig->getI2TConfig(tempVal1) &
                                            TPS2HC10S_I2T_CONFIG_CH1_NOM_CUR_CH1_MASK;
        i2tTrack->Value = (fuseConfig->getI2TConfig(tempVal1) &
                             TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_OFS;
        i2tTripValueLabel->Text = i2tArrayValues[i2tTrack->Value].ToString() + " A2s";
        swLimitBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
                             TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_OFS;
        swDelayBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
                             TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_OFS;
        cooldownDelayBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
                             TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_OFS;
        if (cooldownDelayBox->SelectedIndex == 0)
        {
            resetLatchButton->Enabled = true;
        }
        else
        {
            resetLatchButton->Enabled = false;
        }


        hwLimitBox->SelectedIndex = fuseConfig->getILIMConfig(tempVal1) &
                             TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK;

        if (fuseConfig->getILIMConfig(tempVal1) & TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK)
        {
            enableI2TBox->CheckState = CheckState::Checked;
        }
        else
        {
            enableI2TBox->CheckState = CheckState::Unchecked;
        }

        /* Channel Setting Tab */
        tempVal1 = chanSettingBox->SelectedIndex;

        /* Channel configuration */
        if ((fuseConfig->getILIMConfig(tempVal1) & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK) == 0)
        {
            capChargeEnableBox->CheckState = CheckState::Unchecked;
            capChargeModeBox->SelectedIndex = 0;
        }
        else
        {
            capChargeEnableBox->CheckState = CheckState::Checked;
            capChargeModeBox->SelectedIndex = ((fuseConfig->getILIMConfig(tempVal1)
                & TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_CAP_CHRG_CH1_OFS) - 1;
        }
        updateCurrentLimitSettings(tempVal1);

        chanSettingNomBox->SelectedIndex = fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_NOM_CUR_CH1_MASK;
        chanSettingI2TBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_I2T_TRIP_CH1_OFS;
        chanSettingSWCLBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_ISWCL_CH1_OFS;
        chanSettingSWDelayBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_SWCL_DLY_TMR_CH1_OFS;
        cooldownDelayBox->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_OFS;
        chanSettingCLTimer->SelectedIndex = (fuseConfig->getI2TConfig(tempVal1) &
            TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_MASK) >> TPS2HC10S_I2T_CONFIG_CH1_TCLDN_CH1_OFS;

        if (fuseConfig->getILIMConfig(tempVal1) & TPS2HC10S_ILIM_CONFIG_CH1_I2T_EN_CH1_MASK)
        {
            chanSettingI2TEnBox->CheckState = CheckState::Checked;
        }
        else
        {
            chanSettingI2TEnBox->CheckState = CheckState::Unchecked;
        }

        chanSettingInrushDurationBox->SelectedIndex = (fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_OFS;
        chanSettingCurrentLimitBox->SelectedIndex = fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK;
        chargeLevelBox->SelectedIndex = fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_ILIMIT_SET_CH1_MASK;
        capChargeDurationBox->SelectedIndex = (fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_DURATION_CH1_OFS;
        capTurnOffBox->SelectedIndex = (fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_OFS;
        threshReductionBox->SelectedIndex = (fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_OCP_INRUSH_CH1_OFS;
        inRushValueBox->SelectedIndex = (fuseConfig->getILIMConfig(tempVal1) &
            TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_MASK) >> TPS2HC10S_ILIM_CONFIG_CH1_INRUSH_LIMIT_CH1_OR_CAP_CHRG_DVDT_OFS;

        if (tempVal1 == 0)
        {
            chanSettingLPMCurrentBox->SelectedIndex =
                (fuseConfig->getLowPowerMode() & TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_MASK) >> TPS2HC10S_LPM_LPM_EXIT_CURR_CH1_OFS;

            if (fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH1_ON_MASK)
            {
                chanSettingENBox->CheckState = CheckState::Checked;
            }
            else
            {
                chanSettingENBox->CheckState = CheckState::Unchecked;
            }

            lhiBehaviorBox->SelectedIndex = (fuseConfig->getDevSAFConfig()
                & TPS2HC10S_DEVICE_SAF_CH1_LH_IN_MASK) >> TPS2HC10S_DEVICE_SAF_CH1_LH_IN_OFS;
        }
        else
        {
            chanSettingLPMCurrentBox->SelectedIndex =
                (fuseConfig->getLowPowerMode() & TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_MASK) >> TPS2HC10S_LPM_LPM_EXIT_CURR_CH2_OFS;

            if (fuseConfig->getSwitchState() & TPS2HC10S_SW_STATE_CH2_ON_MASK)
            {
                chanSettingENBox->CheckState = CheckState::Checked;
            }
            else
            {
                chanSettingENBox->CheckState = CheckState::Unchecked;
            }

            lhiBehaviorBox->SelectedIndex = (fuseConfig->getDevSAFConfig()
                & TPS2HC10S_DEVICE_SAF_CH2_LH_IN_MASK) >> TPS2HC10S_DEVICE_SAF_CH2_LH_IN_OFS;
        }

        if (fuseConfig->getPWMConfig(tempVal1) & TPS2HC10S_PWM_CH1_PWM_EN_CH1_MASK)
        {
            chanSettingPwmEnBox->CheckState = CheckState::Checked;
        }
        else
        {
            chanSettingPwmEnBox->CheckState = CheckState::Unchecked;
        }

        chanSettingPWMFreqBox->SelectedIndex = (fuseConfig->getPWMConfig(tempVal1) &
            TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_FREQ_CH1_OFS;
        chanSettingPWMDutyBox->SelectedIndex = (fuseConfig->getPWMConfig(tempVal1) &
            TPS2HC10S_PWM_CH1_PWM_DTY_CH1_MASK) >> TPS2HC10S_PWM_CH1_PWM_DTY_CH1_OFS;
        
        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_VSNS_DIS_CH1_MASK)
        {
            chanVBBADCBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            chanVBBADCBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_TSNS_DIS_CH1_MASK)
        {
            chanJunctionADCBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            chanJunctionADCBox->CheckState = CheckState::Checked;
        }

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_ISNS_DIS_CH1_MASK)
        {
            chanCurrentADCBox->CheckState = CheckState::Unchecked;
        }
        else
        {
            chanCurrentADCBox->CheckState = CheckState::Checked;
        }

        wbOnThreshBox->SelectedIndex = (fuseConfig->getChanDiagConfig(tempVal1) &
            TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_THLD_CH1_MASK) >> TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_THLD_CH1_OFS;

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_ISNS_SCALE_CH1_MASK)
        {
            adcScaleBox->SelectedIndex = 1;
        }
        else
        {
            adcScaleBox->SelectedIndex = 0;
        }

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_EN_CH1_MASK)
        {
            olScalingBox->CheckState = CheckState::Checked;
        }
        else
        {
            olScalingBox->CheckState = CheckState::Unchecked;
        }

        wbBlankBox->SelectedIndex = (fuseConfig->getChanDiagConfig(tempVal1) &
            TPS2HC10S_DIAG_CONFIG_CH1_OL_SVBB_BLANK_CH1_MASK) >> TPS2HC10S_DIAG_CONFIG_CH1_OL_SVBB_BLANK_CH1_OFS;

        wbOffCurrrentBox->SelectedIndex = (fuseConfig->getChanDiagConfig(tempVal1) &
            TPS2HC10S_DIAG_CONFIG_CH1_OL_PU_STR_CH1_MASK) >> TPS2HC10S_DIAG_CONFIG_CH1_OL_PU_STR_CH1_OFS;
        
        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_LATCH_CH1_MASK)
        {
            shutdownLatchBox->Checked = true;
            shutdownRetryBox->Checked = false;
        }
        else
        {
            shutdownLatchBox->Checked = false;
            shutdownRetryBox->Checked = true;
        }

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_OL_OFF_EN_CH1_MASK)
        {
            olOffPullBox->CheckState = CheckState::Checked;
        }
        else
        {
            olOffPullBox->CheckState = CheckState::Unchecked;
        }

        if (fuseConfig->getChanDiagConfig(tempVal1) & TPS2HC10S_DIAG_CONFIG_CH1_SVBB_EN_CH1_MASK)
        {
            vbbPulldownBox->CheckState = CheckState::Checked;
        }
        else
        {
            vbbPulldownBox->CheckState = CheckState::Unchecked;
        }

        slewRateBox->SelectedIndex = (fuseConfig->getChanDiagConfig(tempVal1) &
            TPS2HC10S_DIAG_CONFIG_CH1_SLRT_CH1_MASK);

        dontRunEvents = oldConfigValue;

        redrawI2TGraph();
    }
}

void autohssconfigurator::AutoConfigurationGui::updateFaultGuiFromConfig(void)
{
    uint16_t curFaultStatus = 0;

    if (this->enChanBoxCC->InvokeRequired)
    {
        UpdateFaultCallback^ updateCallback = gcnew UpdateFaultCallback(this,
            &autohssconfigurator::AutoConfigurationGui::updateFaultGuiFromConfig);
        this->Invoke(updateCallback);
    }
    else
    {
        /* Updating the device fault status */
        curFaultStatus = fuseConfig->getDevFault();

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_VBB_UVLO_MASK)
        {
            vbbUVFaultBox->Image = redButton;

            if (vbbUVLOFault->isFaulted() == false)
            {
                logMessageToWindow("VBB UVLO fault detected.");
            }

            porFault->setFaultState(true);
        }
        else
        {
            vbbUVFaultBox->Image = greenButton;

            if (vbbUVLOFault->isFaulted() == true)
            {
                logMessageToWindow("VBB UVLO Fault recovered.");
            }

            vbbUVLOFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_VBB_UV_WRN_MASK)
        {
            uvWarnBox->Image = redButton;

            if (vbbWarningFault->isFaulted() == false)
            {
                logMessageToWindow("VBB UVLO warning detected.");
            }

            vbbWarningFault->setFaultState(true);
        }
        else
        {
            uvWarnBox->Image = greenButton;

            if (vbbWarningFault->isFaulted() == true)
            {
                logMessageToWindow("VBB UVLO warning detected.");
            }

            vbbWarningFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_VDD_UVLO_MASK)
        {
            uvFaultBox->Image = redButton;

            if (vddUvloFault->isFaulted() == false)
            {
                logMessageToWindow("VDD UVLO fault detected.");
            }

            vddUvloFault->setFaultState(true);
        }
        else
        {
            uvFaultBox->Image = greenButton;

            if (vddUvloFault->isFaulted() == true)
            {
                logMessageToWindow("VDD UVLO fault recovered.");
            }

            vddUvloFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_WD_ERR_MASK)
        {
            wdFaultBox->Image = redButton;

            if (wdtFault->isFaulted() == false)
            {
                logMessageToWindow("Watchdog fault detected.");
            }

            wdtFault->setFaultState(true);
        }
        else
        {
            wdFaultBox->Image = greenButton;

            if (wdtFault->isFaulted() == true)
            {
                logMessageToWindow("Watchdog fault recovered.");
            }

            wdtFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_SPI_ERR_MASK)
        {
            spiFaultBox->Image = redButton;

            if (spiFault->isFaulted() == false)
            {
                logMessageToWindow("SPI fault detected.");
            }

            spiFault->setFaultState(true);
        }
        else
        {
            spiFaultBox->Image = greenButton;

            if (spiFault->isFaulted() == true)
            {
                logMessageToWindow("SPI fault recovered.");
            }

            spiFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_POR_MASK)
        {
            porFaultBox->Image = redButton;

            if (porFault->isFaulted() == false)
            {
                logMessageToWindow("Power on Reset (POR) detected.");
            }

            porFault->setFaultState(true);

        }
        else
        {
            porFaultBox->Image = greenButton;

            if (porFault->isFaulted() == true)
            {
                logMessageToWindow("Power on Reset (POR) recovered.");
            }

            porFault->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_LIMPHOME_STAT_MASK)
        {
            limpHomeFaultBox->Image = redButton;

            if (limpFault->isFaulted() == false)
            {
                logMessageToWindow("Limp home fault detected.");
            }

            limpFault->setFaultState(true);
            clearLimpButton->Enabled = true;

        }
        else
        {
            limpHomeFaultBox->Image = greenButton;

            if (limpFault->isFaulted() == true)
            {
                logMessageToWindow("Limp home fault recovered.");
            }

            limpFault->setFaultState(false);
            clearLimpButton->Enabled = false;
        }

        if (curFaultStatus & TPS2HC10S_CH_FLT_TYPE_FAULT_GLOBAL_TYPE_GLOBAL_ERR_WRN_MASK)
        {
            faultBoxGlobal->Image = redButton;
            globalFaultLabel->ForeColor = Color::Red;
            globalFaultLabel->Text = curFaultWord;
        }
        else
        {
            faultBoxGlobal->Image = greenButton;
            globalFaultLabel->ForeColor = Color::Black;
            globalFaultLabel->Text = curNoFaultWord;
        }



        /* Channel Faults */
        curFaultStatus = fuseConfig->getChanFault(0);

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_THERMAL_WRN_CH1_MASK)
        {
            thermalWarnCh1->Image = redButton;

            if (thWarnCh1->isFaulted() == false)
            {
                logMessageToWindow("Thermal warning (CH1) detected.");
            }

            thWarnCh1->setFaultState(true);
        }
        else
        {
            thermalWarnCh1->Image = greenButton;

            if (thWarnCh1->isFaulted() == true)
            {
                logMessageToWindow("Thermal warning (CH1) recovered.");
            }

            thWarnCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_OL_ON_CH1_MASK)
        {
            olOnCh1->Image = redButton;

            if (openLoadOnFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("Open load (ON) (CH1) detected.");
            }

            openLoadOnFaultCh1->setFaultState(true);
        }
        else
        {
            olOnCh1->Image = greenButton;

            if (openLoadOnFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("Open load (ON) (CH1) recovered.");
            }

            openLoadOnFaultCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_OL_OFF_CH1_MASK)
        {
            olOffCh1->Image = redButton;

            if (openLoadOffFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("Open load (OFF) (CH1) detected.");
            }

            openLoadOffFaultCh1->setFaultState(true);
        }
        else
        {
            olOffCh1->Image = greenButton;

            if (openLoadOffFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("Open load (OFF) (CH1) recovered.");
            }

            openLoadOffFaultCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_SHRT_VBB_CH1_MASK)
        {
            shtVbbCh1->Image = redButton;

            if (shortVbbFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("Short-to-VBB (CH1) detected.");
            }

            shortVbbFaultCh1->setFaultState(true);
        }
        else
        {
            shtVbbCh1->Image = greenButton;

            if (shortVbbFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("Short-to-VBB (CH1) recovered.");
            }

            shortVbbFaultCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_ILIMIT_CH1_MASK)
        {
            ilimCh1->Image = redButton;

            if (ilimFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("Current limit (CH1) detected.");
            }

            ilimFaultCh1->setFaultState(true);
        }
        else
        {
            ilimCh1->Image = greenButton;

            if (ilimFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("Current limit (CH1) recovered.");
            }

            ilimFaultCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_THERMAL_SD_CH1_MASK)
        {
            tsdCh1->Image = redButton;

            if (thermShutCh1->isFaulted() == false)
            {
                logMessageToWindow("Thermal shutdown (CH1) detected.");
            }

            thermShutCh1->setFaultState(true);
        }
        else
        {
            tsdCh1->Image = greenButton;

            if (thermShutCh1->isFaulted() == true)
            {
                logMessageToWindow("Thermal shutdown (CH1) recovered.");
            }

            thermShutCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_LPM_WAKE_CH1_MASK)
        {
            lpmWakeCh1->Image = redButton;
            lpmLabel1->Text = "Triggered";
            lpmLabel1->ForeColor = Color::Red;

            if (lpmWakeFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("LPM Wakeup (CH1) detected.");
            }

            lpmWakeFaultCh1->setFaultState(true);
        }
        else
        {
            lpmWakeCh1->Image = greenButton;
            lpmLabel1->Text = "Not Triggered";
            lpmLabel1->ForeColor = Color::Green;

            if (lpmWakeFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("LPM Wakeup (CH1) recovered.");
            }

            lpmWakeFaultCh1->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_I2T_FLT_CH1_MASK)
        {
            i2tWarnCh1->Image = redButton;

            if (i2tFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("I2T Fault (CH1) detected.");
            }

            i2tFaultCh1->setFaultState(true);
        }
        else
        {
            i2tWarnCh1->Image = greenButton;

            if (i2tFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("I2T Fault (CH1) recovered.");
            }

            i2tFaultCh1->setFaultState(false);
        }

       /* if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_VOUT_ERR_CH1_MASK)
        {
            voutErrCh1->Image = redButton;

            if (voutFaultCh1->isFaulted() == false)
            {
                logMessageToWindow("VOUT Fault (CH1) detected.");
            }

            voutFaultCh1->setFaultState(true);
        }
        else
        {
            voutErrCh1->Image = greenButton;

            if (voutFaultCh1->isFaulted() == true)
            {
                logMessageToWindow("VOUT Fault (CH1) recovered.");
            }

            voutFaultCh1->setFaultState(false);
        } */

        if (curFaultStatus & HSS_CHANNEL_FAULT_MASK)
        {
            faultBox1->Image = redButton;
            channelOneFaultLabel->ForeColor = Color::Red;
            channelOneFaultLabel->Text = curFaultWord;
        }
        else
        {
            faultBox1->Image = greenButton;
            channelOneFaultLabel->ForeColor = Color::Black;
            channelOneFaultLabel->Text = curNoFaultWord;
        }

        curFaultStatus = fuseConfig->getChanFault(1);

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_THERMAL_WRN_CH1_MASK)
        {
            thermalWarnCh2->Image = redButton;

            if (thWarnCh2->isFaulted() == false)
            {
                logMessageToWindow("Thermal warning (CH2) detected.");
            }

            thWarnCh2->setFaultState(true);
        }
        else
        {
            thermalWarnCh2->Image = greenButton;

            if (thWarnCh2->isFaulted() == true)
            {
                logMessageToWindow("Thermal warning (CH2) recovered.");
            }

            thWarnCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_OL_ON_CH1_MASK)
        {
            olOnCh2->Image = redButton;

            if (openLoadOnFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("Open load ON (CH2) detected.");
            }

            openLoadOnFaultCh2->setFaultState(true);
        }
        else
        {
            olOnCh2->Image = greenButton;

            if (openLoadOnFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("Open load ON (CH2) recovered.");
            }

            openLoadOnFaultCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_OL_OFF_CH1_MASK)
        {
            olOffCh2->Image = redButton;

            if (openLoadOffFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("Open load OFF (CH2) detected.");
            }

            openLoadOffFaultCh2->setFaultState(true);
        }
        else
        {
            olOffCh2->Image = greenButton;

            if (openLoadOffFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("Open load OFF (CH2) recovered.");
            }

            openLoadOffFaultCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_SHRT_VBB_CH1_MASK)
        {
            shtVbbCh2->Image = redButton;

            if (shortVbbFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("Short-to-VBB fault (CH2) detected.");
            }

            shortVbbFaultCh2->setFaultState(true);
        }
        else
        {
            shtVbbCh2->Image = greenButton;


            if (shortVbbFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("Short-to-VBB fault (CH2) recovered.");
            }

            shortVbbFaultCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_ILIMIT_CH1_MASK)
        {
            ilimCh2->Image = redButton;

            if (ilimFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("Current Limit Fault (CH2) detected.");
            }

            ilimFaultCh2->setFaultState(true);
        }
        else
        {
            ilimCh2->Image = greenButton;

            if (ilimFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("Current Limit Fault (CH2) recovered.");
            }

            ilimFaultCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_THERMAL_SD_CH1_MASK)
        {
            tsdCh2->Image = redButton;

            if (thermShutCh2->isFaulted() == false)
            {
                logMessageToWindow("Thermal Shutdown (CH2) detected.");
            }

            thermShutCh2->setFaultState(true);
        }
        else
        {
            tsdCh2->Image = greenButton;

            if (thermShutCh2->isFaulted() == true)
            {
                logMessageToWindow("Thermal Shutdown (CH2) recovered.");
            }

            thermShutCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_LPM_WAKE_CH1_MASK)
        {
            lpmWakeCh2->Image = redButton;
            lpmLabel2->Text = "Triggered";
            lpmLabel2->ForeColor = Color::Red;

            if (lpmWakeFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("LPM Wake (CH2) detected.");
            }

            lpmWakeFaultCh2->setFaultState(true);
        }
        else
        {
            lpmWakeCh2->Image = greenButton;
            lpmLabel2->Text = "Not Triggered";
            lpmLabel2->ForeColor = Color::Green;

            if (lpmWakeFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("LPM Wake (CH2) recovered.");
            }

            lpmWakeFaultCh2->setFaultState(false);
        }

        if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_I2T_FLT_CH1_MASK)
        {
            i2tWarnCh2->Image = redButton;

            if (i2tFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("I2T Fault (CH2) detected.");
            }

            i2tFaultCh2->setFaultState(true);
        }
        else
        {
            i2tWarnCh2->Image = greenButton;

            if (i2tFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("I2T Fault (CH2) recovered.");
            }

            i2tFaultCh2->setFaultState(false);
        }

       /* if (curFaultStatus & TPS2HC10S_FLT_STAT_CH1_VOUT_ERR_CH1_MASK)
        {
            voutErrCh2->Image = redButton;

            if (voutFaultCh2->isFaulted() == false)
            {
                logMessageToWindow("VOUT fault (CH2) detected.");
            }

            voutFaultCh2->setFaultState(true);
        }
        else
        {
            voutErrCh2->Image = greenButton;

            if (voutFaultCh2->isFaulted() == true)
            {
                logMessageToWindow("VOUT fault (CH2) recovered.");
            }

            voutFaultCh2->setFaultState(false);
        } */

        if (curFaultStatus & HSS_CHANNEL_FAULT_MASK)
        {
            faultBox2->Image = redButton;
            channelTwoFaultLabel->ForeColor = Color::Red;
            channelTwoFaultLabel->Text = curFaultWord;
        }
        else
        {
            faultBox2->Image = greenButton;
            channelTwoFaultLabel->ForeColor = Color::Black;
            channelTwoFaultLabel->Text = curNoFaultWord;
        }

        if (fuseConfig->isFaultTriggered())
        {
            faultPinLabel->Text = "Triggered";
            faultPinLabel->ForeColor = Color::Red;
        }
        else
        {
            faultPinLabel->Text = "Not Triggered";
            faultPinLabel->ForeColor = Color::Green;
        }

        handleFaultClick(curFault);
    }
}

System::Void autohssconfigurator::AutoConfigurationGui::handleFaultClick(HSSFault^ fault)
{
    this->faultDescriptionLabel->Text = fault->getFaultDescription();
    this->statsGroupBox->Text = "Statistics - " + fault->getFaultName();

    if (fault->isFaulted())
    {
        currentFaultLabelStatus->Text = "Faulted";
        currentFaultLabelStatus->ForeColor = Color::Red;
    }
    else
    {
        currentFaultLabelStatus->Text = "Not Faulted";
        currentFaultLabelStatus->ForeColor = Color::Green;
    }

    numberFaultLabel->Text = fault->getFaultCount().ToString();
    dateFaultLabel->Text = fault->getFaultDate();
}

float autohssconfigurator::AutoConfigurationGui::generateAverage(System::Collections::Generic::List<float>^ arr)
{
    int ii;
    double aggVal = 0;

    for (ii = 0; ii < arr->Count; ii++)
    {
        aggVal += arr[ii];
    }

    aggVal = aggVal / arr->Count;

    return (float)(aggVal);
}

/* Updates the GUI with new current sense data received from the current sense transaction */
void autohssconfigurator::AutoConfigurationGui::updateCurrentSenseGUI(tHSSReceiveTransaction^ cmd)
{
    float normalizedValue1 = 0.0;
    float normalizedValue2 = 0.0;
    float avgValue1 = 0.0;
    float avgValue2 = 0.0;
    uint8_t* bufferData = cmd->recData + 1;
    float maxValue = 0.01f;
    int dataCount;
    int avgCount;
    float totalValue;
    float ksnsCh1, ksnsCh2;


    if (enChanBoxCC->InvokeRequired)
    {
        UpdateSenseCallback^ updateCallback = gcnew UpdateSenseCallback(this,
            &autohssconfigurator::AutoConfigurationGui::updateCurrentSenseGUI);
        this->Invoke(updateCallback, cmd);
    }
    else
    {

        if (!int::TryParse(maxSamplesBox->Text, dataCount) || dataCount < 10)
        {
            maxSamplesBox->Text = "10";
            dataCount = 50;
        }

        if (!int::TryParse(averageCountBox->Text, avgCount))
        {
            averageCountBox->Text = "10";
            avgCount = 10;
        }

        if (dataCount < avgCount)
        {
            avgCount = dataCount;
            averageCountBox->Text = dataCount.ToString();
        }

        if (dataCount > 1000)
        {
            maxSamplesBox->Text = "1000";
            dataCount = 1000;
        }

        if (avgCount > 1000)
        {
            averageCountBox->Text = "1000";
            avgCount = 1000;
        }

        /* Converting the samples */
        ksnsCh1 = fuseConfig->getChanDiagConfig(0) & TPS2HC10S_DIAG_CONFIG_CH1_OL_ON_EN_CH1_MASK ? HSSC_KSNS_SCALE : HSSC_KSNS;
        ksnsCh2 = fuseConfig->getChanDiagConfig(1) & TPS2HC10S_DIAG_CONFIG_CH2_OL_ON_EN_CH2_MASK ? HSSC_KSNS_SCALE : HSSC_KSNS;
        normalizedValue1 = convertRawADCValue((bufferData[0] << 8) | bufferData[1], curDataFormat, ksnsCh1);
        normalizedValue2 = convertRawADCValue((bufferData[2] << 8) | bufferData[3], curDataFormat, ksnsCh2);

        while (avgValsCh1->Count >= avgCount)
        {
            avgValsCh1->RemoveAt(0);
            avgValsCh2->RemoveAt(0);
        }
        avgValsCh1->Add(normalizedValue1);
        avgValsCh2->Add(normalizedValue2);

        avgValue1 = ceil(generateAverage(avgValsCh1) * 100.0f) / 100.0f;
        avgValue2 = ceil(generateAverage(avgValsCh2) * 100.0f) / 100.0f;
        totalValue = avgValue1;

        if (avgValue1 > maxValue && senseChart->Series[0]->Enabled)
        {
            maxValue = avgValue1;
        }
        senseChart->Series[0]->Points->Add(avgValue1);
        if (avgValue1 < 1 && curDataFormat == HSSC_CURRENT_OFFSET)
        {
            avgValue1 = avgValue1 * 1000;
            channelOneLabel->Text = avgValue1.ToString() + "m" + curSenseUnit;
        }
        else
        {
            channelOneLabel->Text = avgValue1.ToString() + curSenseUnit;
        }

        while (senseChart->Series[0]->Points->Count > dataCount)
        {
            senseChart->Series[0]->Points->RemoveAt(0);
        }

        /* Channel 2 */
        totalValue += avgValue2;

        if (avgValue2 > maxValue && senseChart->Series[1]->Enabled)
        {
            maxValue = avgValue2;
        }
        senseChart->Series[1]->Points->Add(avgValue2);
        if (avgValue2 < 1 && curDataFormat == HSSC_CURRENT_OFFSET)
        {
            avgValue2 = avgValue2 * 1000;
            channelTwoLabel->Text = avgValue2.ToString() + "m" + curSenseUnit;
        }
        else
        {
            channelTwoLabel->Text = avgValue2.ToString() + curSenseUnit;
        }
        while (senseChart->Series[1]->Points->Count > dataCount)
        {
            senseChart->Series[1]->Points->RemoveAt(0);
        }

        /* Total value */
        if (curDataFormat == HSSC_CURRENT_OFFSET)
        {
            if (totalValue < 1)
            {
                totalValue = totalValue * 1000;
                    globalLabel->Text = totalValue.ToString() + "m" + curSenseUnit;
            }
            else
            {
                globalLabel->Text = totalValue.ToString() + curSenseUnit;
            }
        }
        else
        {
            globalLabel->Text = "";
        }

        /* Normalizing the chart's scale */
        maxValue = ceil(maxValue * 1.1f);
        senseChart->ChartAreas[0]->AxisY->Minimum = 0;
        senseChart->ChartAreas[0]->AxisY->Maximum = maxValue;

        /* Input voltage */
        avgValue1 = convertRawADCValue((bufferData[4] << 8) | bufferData[5], HSSC_VBB_OFFSET, 0);
        inputVoltageValueLabel->Text = avgValue1.ToString() + "V";

        if (isSnsLoggingEnabled == true)
        {
            String^ dateString = DateTime::Now.ToString("hh:mm:ss");
            fileLogger->WriteLine(dateString + "," + avgValue1.ToString() + ","
                + normalizedValue1.ToString() + "," + normalizedValue2.ToString());
        }
    }
}

/* This function will take a raw ADC value and convert it into the "real" (temperature, current, etc.) */
float autohssconfigurator::AutoConfigurationGui::convertRawADCValue(uint16_t rawValue, uint8_t format, float ksnsVal)
{
    float senseResistor = float::Parse(senseResistorValBox->Text);
    rawValue = rawValue & TPS2HC10S_ADC_RESULT_CH1_I_ADC_RESULT_CH1_I_MASK;

    if (format == HSSC_CURRENT_OFFSET)
    {
        return ceilf((((float)rawValue) * (ksnsVal * HSSC_ADCREF)) /
            (1023.0f * senseResistor) * 100.0f) / 100.0f;
    }
    else if (format == HSSC_TEMP_OFFSET)
    {
        return ceilf(((-1.415f + (((float)rawValue * HSSC_ADCREF) / 1705.341f)) * -300.0f) * 100.0f) / 100.0f;
    }
    else if (format == HSSC_VOUT_OFFSET)
    {
        return ceilf((((float)rawValue) * (17.89f * HSSC_ADCREF)) /
            (1023.0f * 1.667f) * 100.0f) / 100.0f;
    }
    else if (format == HSSC_VBB_OFFSET)
    {
        return ceilf((((float)rawValue) * (18.18f * HSSC_ADCREF)) /
            (1023.0f * 1.667f) * 100.0f) / 100.0f;
    }
    else
    {
        return 0.0f;
    }
}

void autohssconfigurator::AutoConfigurationGui::updateVerString(void)
{
    if (enChanBoxCC->InvokeRequired)
    {
        UpdateConnnectionCallback^ updateCallback = gcnew UpdateConnnectionCallback(this,
            &AutoConfigurationGui::AutoConfigurationGui::updateVerString);
        this->Invoke(updateCallback);
    }
    else
    {
        this->firmwareRevisionLabel->Text = "v" + fuseConfig->getFWVersion().ToString("X4");

        if (fuseConfig->getFWVersion() < FW_VER_PURPLE)
        {
            flashPersistToolStripMenuItem->Enabled = false;
        }
        else
        {
            flashPersistToolStripMenuItem->Enabled = true;
        }
    }
}

/* This function handles IOCTL read responses from the device */
void autohssconfigurator::AutoConfigurationGui::handleIOCTL(uint8_t ioctl, uint8_t* payload)
{
    uint16_t fwVer;

    switch (ioctl)
    {
        case HSSCommand::HSS_IOCTL_GETVERSION:
            fwVer = (payload[1] << 8) | payload[0];
            this->fuseConfig->setFWVersion(fwVer);
            this->updateVerString();
            break;
        default:
            break;
    }
}

/* Main event handler from received commands from the device */
void autohssconfigurator::AutoConfigurationGui::bufferReceiveEventHandler(tHSSReceiveTransaction^ cmd)
{
    /* Making sure the buffer is valid */
    if (cmd->recData == NULL)
    {
        return;
    }

    /* Handles the receive event */
    switch (cmd->reportID)
    {
    case HSSCommunicationManager::HSSC_DEV_CONFIG_ID:
        this->fuseConfig->updateDevConfigFromBuffer(cmd->recData + 1);
        this->updateGuiFromConfig(true);
        break;
    case HSSCommunicationManager::HSSC_CH_CONFIG_ID:
        this->fuseConfig->updateChConfigFromBuffer(cmd->recData + 2, cmd->recData[1]);
        this->updateGuiFromConfig(true);
        break;
    case HSSCommunicationManager::HSSC_FAULT_ID:
        this->fuseConfig->setFaultStatus(cmd->recData + 1);
        this->updateFaultGuiFromConfig();
        break;
    case HSSCommunicationManager::HSSC_RAWREAD_ID:
        this->handleRawReadResponse(cmd);
        break;
    case HSSCommunicationManager::HSSC_SNS_ID:
        this->updateCurrentSenseGUI(cmd);
        break;
    case HSSCommunicationManager::HSSC_IOCTL_ID:
        this->handleIOCTL(cmd->recData[1], cmd->recData + 2);
        break;
    default:
        return;
    }

    free(cmd->recData);
}